#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>


int main(void)
{
	char	inbuf[_POSIX_MAX_INPUT];
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	pid_t 	pid;

	setup(E_INTERFACE);
	funct(E_INTERFACE);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(10, BASE)
		GTI_NOT_PROVIDED_RESULT(11, BASE)
		GTI_NOT_PROVIDED_RESULT(12, BASE)
		GTI_NOT_PROVIDED_RESULT(13, BASE)
		done();
	}

	continue_on(ERROR);

	assertion(10, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
		finis

		FORK
			prep("Set TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag |= TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis
	
			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTIN, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child of background process receives SIGTTIN.");
				start
					sleep(WAIT_FOR_TIME);
					if(sig_rcvd == 0)
						error("SIGTTIN not received.");
				finis

				done();
			}

			testing("A background process, not ignoring or blocking");
			testing("SIGTTIN, attempting to read from the session group's");
			testing("controlling terminal shall cause its process group");
			testing("to be sent a SIGTTIN signal.");
			start
				SET_TIMEOUT
					read(tty_fildes, inbuf, sizeof(inbuf));
				CLEAR_ALARM
				if(sig_rcvd == 0)
					error("SIGTTIN not received.");
			finis

			WAIT_FOR_PID(pid);
			done();
		KROF

							/* Repeat test with TOSTOP clear. */
		FORK
			prep("Clear TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag &= ~TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis

			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTIN, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child of background process receives SIGTTIN.");
				start
					sleep(WAIT_FOR_TIME);
					if(sig_rcvd == 0)
						error("SIGTTIN not received.");
				finis

				done();
			}

			testing("A background process, not ignoring or blocking");
			testing("SIGTTIN, attempting to read from the session group's");
			testing("controlling terminal shall cause its process group");
			testing("to be sent a SIGTTIN signal.");
			start
				SET_TIMEOUT
					read(tty_fildes, inbuf, sizeof(inbuf));
				CLEAR_ALARM
				if(sig_rcvd == 0)
					error("SIGTTIN not received.");
			finis

			WAIT_FOR_PID(pid);
			done();
		KROF
	finis

	assertion(11, BASE);
	start
		FORK
			prep("Set TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag |= TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis
	
			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTIN, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child of background process receives no SIGTTIN.");
				start
					sleep(WAIT_FOR_TIME);
					if(sig_rcvd != 0)
						error("SIGTTIN was received.");
				finis

				done();
			}

			testing("A background process, ignoring SIGTTIN,");
			testing("attempting to read from its controlling");
			testing("terminal returns errno set to [EIO] and no");
			testing("SIGTTIN signal is sent to any process in process group.");
			start
				SIGNAL(SIGTTIN, SIG_IGN);
				SET_TIMEOUT
					expecting((ssize_t)-1);
					expecting(EIO);
					Zread(tty_fildes, inbuf, sizeof(inbuf));
				CLEAR_ALARM
				if(sig_rcvd != 0)
					error("SIGTTIN was received.");
			finis

			WAIT_FOR_PID(pid);
			done();
		KROF

							/* Repeat test with TOSTOP clear. */
		FORK
			prep("Clear TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag &= ~TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis

			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTIN, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child background process does not receive SIGTTIN.");
				start
					sleep(WAIT_FOR_TIME);
					if(sig_rcvd != 0)
						error("SIGTTIN was received.");
				finis

				done();
			}

			testing("A background process, ignoring SIGTTIN,");
			testing("attempting to read from its controlling");
			testing("terminal returns errno set to [EIO] and no");
			testing("SIGTTIN signal is sent to any process in process group.");
			start
				SIGNAL(SIGTTIN, SIG_IGN);
				SET_TIMEOUT
					expecting((ssize_t)-1);
					expecting(EIO);
					Zread(tty_fildes, inbuf, sizeof(inbuf));
				CLEAR_ALARM
				if(sig_rcvd != 0)
					error("SIGTTIN was received.");
			finis

			WAIT_FOR_PID(pid);
			done();
		KROF
	finis


	assertion(12, BASE);
	start
		sigset_t	set;

		FORK
			prep("Set TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag |= TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis
	
			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTIN, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child of background process receives no SIGTTIN.");
				start
					sleep(WAIT_FOR_TIME);
					if(sig_rcvd != 0)
						error("SIGTTIN was received.");
				finis

				done();
			}

			testing("A background process, blocking SIGTTIN,");
			testing("attempting to read from its controlling");
			testing("terminal returns errno set to [EIO] and no");
			testing("SIGTTIN signal is sent to any process in process group.");
			start
				SIGBLOCK(set, SIGTTIN);
				SET_TIMEOUT
					expecting((ssize_t)-1);
					expecting(EIO);
					Zread(tty_fildes, inbuf, sizeof(inbuf));
				CLEAR_ALARM

				Zsigpending(&set);
				if(sigismember(&set, SIGTTIN))
					error("SIGTTIN was received.");
			finis

			WAIT_FOR_PID(pid);
			done();
		KROF

							/* Repeat test with TOSTOP clear. */
		FORK
			prep("Clear TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag &= ~TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis

			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTIN, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child background process does not receive SIGTTIN.");
				start
					sleep(WAIT_FOR_TIME);
					if(sig_rcvd != 0)
						error("SIGTTIN was received.");
				finis

				done();
			}

			testing("A background process, blocking SIGTTIN,");
			testing("attempting to read from its controlling");
			testing("terminal returns errno set to [EIO] and no");
			testing("SIGTTIN signal is sent to any process in process group.");
			start
				SIGBLOCK(set, SIGTTIN);
				SET_TIMEOUT
					expecting((ssize_t)-1);
					expecting(EIO);
					Zread(tty_fildes, inbuf, sizeof(inbuf));
				CLEAR_ALARM

				Zsigpending(&set);
				if(sigismember(&set, SIGTTIN))
					error("SIGTTIN was received.");
			finis

			WAIT_FOR_PID(pid);
			done();
		KROF
	finis


	assertion(13, BASE);
	start
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Set TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag |= TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis
	
			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTIN, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Fork a child process.");
				start
					sv_flush();
					pid = Zfork();
				finis

				if(pid == (pid_t)0) {
					testing("Child of orphaned background process group");
					testing("receives no SIGTTIN.");
					start
						sleep(2*WAIT_FOR_TIME);
						if(sig_rcvd != 0)
							error("SIGTTIN was received.");
					finis

					done();
				}

				testing("An orphaned background process,");
				testing("attempting to read from its controlling");
				testing("terminal returns errno set to [EIO] and no SIGTTIN");
				testing("signal is sent to any process in process group.");
				start
					SLEEP(WAIT_FOR_TIME);	/* Ensure process group orphaned. */
					SET_ALARM(WAIT_FOR_TIME)
						expecting((ssize_t)-1);
						expecting(EIO);
						Zread(tty_fildes, inbuf, sizeof(inbuf));
					CLEAR_ALARM

					if(sig_rcvd != 0)
						error("SIGTTIN was received.");
				finis
	
				WAIT_FOR_PID(pid);
				done();
			}
			done();					/* Exit process group leader. */
		}
		WAIT_FOR_PID(pid);			/* Establish orphaned process group. */
						/* Allow time for orphaned processses to complete.	*/
		SLEEP(3*WAIT_FOR_TIME);
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis
	done();
}
