#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

char String[] = "With a capital T that rhymes with P that stands for pool!\n";
#define Strlen	(sizeof(String) -1)


int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	*inbuf, *outbuf;
	long	max_canon;

	setup(E_INTERFACE);
	funct(E_INTERFACE);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(24, BASE)
		GTI_NOT_PROVIDED_RESULT(25, BASE)
		GTI_NOT_PROVIDED_RESULT(26, BASE)
		done();
	}

								/* At most one line */
	assertion(24, BASE);
	start
		prep("Establish read/write buffers.");
		start
			exit_on(ERROR);
			inbuf = (char *)Zmalloc(PCTS_MAX_CANON);
			outbuf = (char *)Zmalloc(PCTS_MAX_CANON);
		finis

		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
		finis

		prep("Write two lines of data.");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);
			CLEAR_ALARM
		finis

		testing("Call to read() returns at most one line.");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zread(tty_fildes, inbuf, 2*Strlen);
			CLEAR_ALARM
		finis
	finis


					/*	Doing read()s from a terminal in less than line
						increments will lose no data. */
	assertion(25, BASE);
	start
		prep("Clear PORT's queues.");
		start
			Ztcflush(loop_fildes, TCIOFLUSH);
			dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			Ztcflush(tty_fildes, TCIOFLUSH);
		finis

		testing("If a read() requests fewer characters than available");
		testing("in the input line, even a single character, that number");
		testing("of characters will be returned and the next read()'s");
		testing("first character read will be the first unread character");
		testing("in the line(thereby losing no information).");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);

				expecting(1);
				Zread(tty_fildes, inbuf, 1);
				expecting(2);
				Zread(tty_fildes, inbuf +1, 2);
				expecting(Strlen -3);
				Zread(tty_fildes, inbuf +3, PCTS_MAX_CANON);
			CLEAR_ALARM

			if(strncmp(inbuf, String, Strlen) != 0)
				error("Read string != written string.");
		finis
	finis

	assertion(26, BASE);
	start
		prep("Determine number of bytes to test canonical queue with.");
			prep("Clear PORT's queues.");
		start
			errno = 0;
			max_canon = pathconf(PORT_TTY, _PC_MAX_CANON);
			exit_on(ERROR);
			if(errno != 0)
				error("_PC_MAX_CANON not supported.");
			continue_on(ERROR);
			if((int)max_canon > PCTS_MAX_CANON)
				max_canon = PCTS_MAX_CANON;

			Ztcflush(loop_fildes, TCIOFLUSH);
			Ztcflush(tty_fildes, TCIOFLUSH);
		finis

		testing("When ICANON is set, a terminal's canonical input queue");
		if((int)max_canon <= PCTS_MAX_CANON)
			testing("is able to contain at least {MAX_CANON} bytes.");
		else
			testing("is able to contain at least {PCTS_MAX_CANON} bytes.");
		start
			memset(outbuf, (int)'A', max_canon);	/* Clear out NLs. */
			outbuf[max_canon -1] = '\n';			/* Terminate line. */

			SET_TIMEOUT
				expecting((ssize_t)max_canon);
				Zwrite(loop_fildes, outbuf, (unsigned) max_canon);
				Ztcdrain(loop_fildes);
			CLEAR_ALARM

			SET_TIMEOUT
				expecting((ssize_t)max_canon);
				Zread(tty_fildes, inbuf, (unsigned) max_canon);
			CLEAR_ALARM
		finis
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
