#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>


int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	inbuf[_POSIX_MAX_INPUT];
	char	outbuf[_POSIX_MAX_INPUT];
	struct	tms	tms_buf;
	clock_t	t_beg, t_end;
	size_t	nb_xmit, n_read;
	long	clk_tcks;
	clock_t	t_xmit;
	pid_t	pid;
	int		i;
	sigset_t	set;

	setup(E_TCDRAIN);
	funct(E_TCDRAIN);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(6, BASE)
		GTI_NOT_PROVIDED_RESULT(7, BASE)
		done();
	}

	SIGNAL(SIGHUP, SIG_IGN);		/* Ignore hangups on PORTs. */

	assertion(6, BASE);
	start
		ENABLE_PSYNC
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		prep("Set TOSTOP.");
		start
			exit_on(ERROR);

			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

			Zcfsetospeed(loop_termios_p, baud_min());
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			Zcfsetispeed(tty_termios_p, baud_min());
			memset(outbuf, 'A', _POSIX_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
			nb_xmit = (baud_value(baud_min())/10)*TIMEOUT/2;
			if(nb_xmit > _POSIX_MAX_INPUT)
				nb_xmit = _POSIX_MAX_INPUT;
			clk_tcks = Zsysconf(_SC_CLK_TCK);
		finis

		for(i=0; i<2; i++) {
			if(i==0)
				prep("Set TOSTOP flag.");
			else
				prep("Clear TOSTOP flag.");
			start
				if(i==0)
					tty_termios_p->c_lflag |= TOSTOP;
				else
					tty_termios_p->c_lflag &= ~TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Establish background process group with two processes.");
				start
					Zsetpgid((pid_t)0, (pid_t)0);
					SIGNAL(SIGTTOU, sig_int);
					sig_rcvd = 0;
				finis
	
				prep("Fork a child process.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {
					testing("Child process in background process group");
					testing("did not receive SIGTTOU signal.");
					start
						SLEEP(WAIT_FOR_TIME);
						if(sig_rcvd != 0)
							error("Child process received SIGTTOU.");
					finis
	
					done();
				}
		
				testing("Call to tcdrain(), on the controlling terminal of");
				testing("the process, from a member of background");
				testing("process group that is blocking the SIGTTOU signal");
				if(i==0)
					testing("when TOSTOP is set is successful.");
				else
					testing("when TOSTOP is clear is successful.");
				start
					SIGBLOCK(set, SIGTTOU);
					if(PCTS_GTI_BUFFERS_OUTPUT) {
						clk_tcks = Zsysconf(_SC_CLK_TCK);
						t_xmit = clk_tcks*nb_xmit*10/baud_value(baud_min());
						SET_TIMEOUT
								/*	Record the starting time before the write()
									is called.  In case write() blocks, the time
									to drain the data is monitored properly.
								*/
							t_beg = Ztimes(&tms_buf);
							expecting(nb_xmit);
							Zwrite(tty_fildes, outbuf, nb_xmit);	
							Ztcdrain(tty_fildes);
							t_end = Ztimes(&tms_buf);
						CLEAR_ALARM
					}
					else
						t_xmit = 0;
					SEND_PSYNC
				finis
	
				testing("Call to tcdrain() waits until all output written to");
				testing("the terminal has been transmitted before returning.");
				start
					if(PCTS_GTI_BUFFERS_OUTPUT) {
						if(t_end -t_beg < t_xmit -t_xmit/10)
							error("%s%lu clock-ticks.\n%s%lu clock-ticks.",
									"Expected tcdrain() to take ",
									(unsigned long)t_xmit,
									"Actually time taken was ",
									(unsigned long) (t_end -t_beg) );
					}
				finis
	
				testing("SIGTTOU signal not sent to calling process.");
				start
					dt_secs(TIME_ACQUIRE_XMITTED_DATA);
					if(sig_rcvd != 0)
						error("Calling process received SIGTTOU.");
				finis
	
				done();
			}
	
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
	
			testing("Ensure all data was transmitted.");
			start
				break_on(ERROR);
				SET_TIMEOUT
					for(n_read=0; n_read != nb_xmit;)
						n_read += Zread(loop_fildes, inbuf, _POSIX_MAX_INPUT);
				CLEAR_ALARM
			finis
	
			WAIT_FOR_PID(pid);
		}
		DISABLE_PSYNC
	finis
	
	
	assertion(7, BASE);
	start
		ENABLE_PSYNC
		for(i=0; i<2; i++) {
			if(i==0)
				prep("Set TOSTOP flag.");
			else
				prep("Clear TOSTOP flag.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				if(i==0)
					tty_termios_p->c_lflag |= TOSTOP;
				else
					tty_termios_p->c_lflag &= ~TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis
	
			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Establish background process group with two processes.");
				start
					Zsetpgid((pid_t)0, (pid_t)0);
					SIGNAL(SIGTTOU, sig_int);
					sig_rcvd = 0;
				finis
	
				prep("Fork a child process.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {
					testing("Child process in background process group");
					testing("did not receive SIGTTOU signal.");
					start
						SLEEP(WAIT_FOR_TIME);
						if(sig_rcvd != 0)
							error("Child process received SIGTTOU.");
					finis
	
					done();
				}
		
				testing("Call to tcdrain(), on the controlling terminal of");
				testing("the process, from a member of background");
				testing("process group that is ignoring the SIGTTOU signal");
				if(i==0)
					testing("when TOSTOP is set is successful.");
				else
					testing("when TOSTOP is clear is successful.");
				start
					SIGNAL(SIGTTOU, SIG_IGN);
					if(PCTS_GTI_BUFFERS_OUTPUT) {
						clk_tcks = Zsysconf(_SC_CLK_TCK);
						t_xmit = clk_tcks*nb_xmit*10/baud_value(baud_min());
						SET_TIMEOUT
							t_beg = Ztimes(&tms_buf);
							expecting(nb_xmit);
							Zwrite(tty_fildes, outbuf, nb_xmit);	
							expecting(SUCCESS);
							Ztcdrain(tty_fildes);
							t_end = Ztimes(&tms_buf);
						CLEAR_ALARM
					}
					else
						t_xmit = 0;
					SEND_PSYNC
				finis
	
				testing("Call to tcdrain() waits until all output written to");
				testing("the terminal has been transmitted before returning.");
				start
					if(PCTS_GTI_BUFFERS_OUTPUT) {
						if(t_end -t_beg < t_xmit -t_xmit/10)
							error("%s%lu clock-ticks.\n%s%lu clock-ticks.",
									"Expected tcdrain() to take ",
									(unsigned long)t_xmit,
									"Actually time taken was ",
									(unsigned long) (t_end -t_beg) );
					}
				finis
	
				testing("SIGTTOU signal not sent to calling process.");
				start
					dt_secs(TIME_ACQUIRE_XMITTED_DATA);
					if(sig_rcvd != 0)
						error("Calling process received SIGTTOU.");
				finis
	
				WAIT_FOR_PID(pid);
	
				done();
			}
	
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
	
			testing("Ensure all data was transmitted.");
			start
				break_on(ERROR);
				SET_TIMEOUT
					for(n_read=0; n_read != nb_xmit;)
						n_read += Zread(loop_fildes, inbuf, _POSIX_MAX_INPUT);
				CLEAR_ALARM
			finis
	
			WAIT_FOR_PID(pid);
		}
		DISABLE_PSYNC
	finis
	
	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(loop_fildes);
		Zclose(tty_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis
	
	done();
}
