#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>


int main(void)
{
	int		notty_fildes;
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	outbuf[_POSIX_MAX_INPUT];
	size_t	nb_xmit;
	pid_t	pid;

	setup(E_TCDRAIN);
	funct(E_TCDRAIN);

	SIGNAL(SIGHUP, SIG_IGN);		/* Ignore hangups on PORTs. */

	assertion(9, BASE);
	start
		testing("Call tcdrain() with an invalid file descriptor.");
		testing("Expect return value of -1 and errno set to [EBADF].");
		start
			expecting(-1);
			expecting(EBADF);
			Ztcdrain(-1);
		finis
	finis


	assertion(10, BASE);
	start
		testing("Call to tcdrain() when fildes is not a tty.");
		testing("Expect a return value of -1 and errno set to [ENOTTY].");
		start
			notty_fildes = temp_open("notspec", O_RDWR | O_CREAT, PERM_CTTY);
			expecting(-1);
			expecting(ENOTTY);
			Ztcdrain(notty_fildes);
		finis
	finis

	if(GTI_NOT_PROVIDED) {
		assertion(11, BASE);
		start
			GTI_NOT_PROVIDED_EXIT
		finis
	}

	if(PCTS_GTI_BUFFERS_OUTPUT)
		assertion(11, BASE);
	else {
		assertion(11, NOT_TESTABLE);
		done();
	}
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		start
			exit_on(ERROR);

			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VMIN] = 0;
			tty_termios_p->c_cc[VTIME] = 0;
			Zcfsetospeed(loop_termios_p, baud_min());
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			Zcfsetispeed(tty_termios_p, baud_min());

			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
			memset(outbuf, 'A', _POSIX_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
			nb_xmit = (baud_value(baud_min())/10)*TIMEOUT/2;
			if(nb_xmit > _POSIX_MAX_INPUT)
				nb_xmit = _POSIX_MAX_INPUT;
		finis

		ENABLE_PSYNC
			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Write data to controlling terminal.");
				start
					expecting(nb_xmit);
					Zwrite(loop_fildes, outbuf, nb_xmit);
					Ztcflow(loop_fildes, TCOOFF);
				finis

				testing("Call to tcdrain() interrupted by signal returns a");
				testing("value of -1 and sets errno to [EINTR].");
				start
					SIGNAL(SIGINT, sig_int);
					SEND_PSYNC
					expecting(-1);
					expecting(EINTR);
					Ztcdrain(loop_fildes);

								/*	Ensure drain of output queue continues.	*/
					Ztcflow(loop_fildes, TCOON);
				finis
	
				done();
			}
	
			prep("Wait for child to write out data, then");
			prep("interrupt child processing tcdrain().");
			start
				sv_flush();
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM

				dt_secs(1);			/* Ensure tcdrain() is called. */
				Zkill(pid, SIGINT);
			finis

			testing("Ensure validity of test result code reported.");
			start
			finis

			WAIT_FOR_PID(pid);
		DISABLE_PSYNC
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		SET_TIMEOUT
			Ztcflush(loop_fildes, TCIOFLUSH);
			dt_secs(1);
			Ztcflush(tty_fildes, TCIOFLUSH);
			Zclose(loop_fildes);
			Zclose(tty_fildes);
		CLEAR_ALARM
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
