#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<Ztermios.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

char String[] = "Still crazy after all these years...\n";
#define	 Strlen	(sizeof(String)-1)

char inbuf[_POSIX_MAX_INPUT];			/* I/O buffers						*/
char stop_char, start_char;				/* START/STOP values				*/

void	com_code(void);

static	int		tty_fildes, loop_fildes;
static	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
static	sigset_t set;

int main(void)
{
	pid_t	pid;

	setup(E_TCFLOW);
	funct(E_TCFLOW);

	assertion(13, BASE);
	start
		GTI_NOT_PROVIDED_EXIT

		prep("Establish PORT_TTY and PORT_LOOP as non-controlling");
		prep("terminals.");
		start
			exit_on(ERROR);
	
			termios_nct(&tty_fildes, tty_termios_p, &loop_fildes,
															loop_termios_p);
								/* Enable flow control on input and output. */
			Ztcgetattr(tty_fildes, tty_termios_p);
			tty_termios_p->c_iflag |=(IXOFF|IXON);
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);

								/* Flow control disabled on PORT_LOOP. */
			Ztcgetattr(loop_fildes, loop_termios_p);
			loop_termios_p->c_cc[VMIN] = 1;
			loop_termios_p->c_cc[VTIME] = 0;
			Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
		finis

		ENABLE_PSYNC
									/* Foreground process test. */
		FORK
			com_code();
		KROF

									/* Background process tests. */
		FORK
			prep("Catch SIGTTOU and establish a background process group.");
			start
				SIGNAL(SIGTTOU, sig_int);
				Zsetpgid((pid_t)0, (pid_t)0);
			finis

			com_code();

			prep("Block SIGTTOU in background process group.");
			start
				SIGBLOCK(set, SIGTTOU);
			finis

			com_code();

			prep("Unblock and ignore SIGTTOU in background process group.");
			start
				SIGNAL(SIGTTOU, SIG_IGN);
				SIGUNBLOCK(set, SIGTTOU);
			finis

			com_code();
		KROF

									/* Orphaned background process test. */
		FORK
			prep("Catch SIGTTOU and establish an orphaned background");
			prep("process group.");
			start
				SIGNAL(SIGTTOU, sig_int);
				Zsetpgid((pid_t)0, (pid_t)0);
			finis

			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				SLEEP(WAIT_FOR_TIME)

				com_code();
				SEND_PSYNCA

				done();
			}
		KROF
										/* Wait for test to complete. */
		SET_TIMEOUT
			WAIT_PSYNCA
		CLEAR_ALARM
		DISABLE_PSYNC
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(loop_fildes);
		Zclose(tty_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}

void com_code(void)
{
	int	cnt =0;

	testing("A call to tcflow(fildes, TCOOFF), on a non-controlling");
	testing("terminal of the process, suspends output from that terminal.");
	start
		expecting(0);
		Ztcflow(tty_fildes, TCOOFF);

		if(PCTS_GTI_BUFFERS_OUTPUT) {
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(tty_fildes, String, Strlen);
			CLEAR_ALARM
		}
		else {
				/*	Implementations that do not provide a buffering
					mechanism on write() will timeout when terminal
					output is suspended.
				*/
			SET_ALARM(WAIT_FOR_TIME)
				write(tty_fildes, "A", 2);
			EXPECT_ALARM
		}

		SLEEP(WAIT_FOR_TIME);

		SET_ALARM(WAIT_FOR_TIME)
			cnt = read(loop_fildes, inbuf, Strlen);
		EXPECT_ALARM

		if(cnt > 0)
			error("TCOOFF failed, output was not stopped.");
		if(memchr(inbuf, stop_char, cnt))
			error("TCOOFF failed, STOP was sent.");
		if(memchr(inbuf, start_char, cnt))
			error("TCOOFF failed, START was sent.");
	finis

	cleanup("Flush queues and restart suspended output.");
	start
		Ztcflush(tty_fildes, TCIOFLUSH);
		Ztcflush(loop_fildes, TCIOFLUSH);
		Ztcflow(tty_fildes, TCOON);
	finis
}
