#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Zstat.h>

#include	<sv_env.h>
#include	<sv_errno.h>

/* This module has the S_ISUID mode set to ROOT. */

#define	FILENAME	"5.6.5_35"


int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	struct	stat st_buf;
	pid_t	pid;
	char	*pathm;

	setup(E_CHOWN);
	funct(E_CHOWN);

	assertion(35, BASE);
	start
		if(PCD_READ_ONLY_FILE_SYSTEM) {
			pathm = fs_mount(MOUNT_RW, ROFS_REG_FILE);
	
			abstract("Ensure mounted file system gets unmounted.");
			start
				break_on(ERROR);
				prep("Establish path to implementation provided");
				prep("ROFS_TESTFILE.");
				start
					Zstrcpy(path, pathm);
					Zstrcat(path, "/");
					Zstrcat(path, ROFS_TESTFILE);
				finis
	
				prep("Set file owner to root's UID and GID.");
				start
					Zchown(path, ROOT_UID, ROOT_GID);
				finis
	
				fs_umount(NO_FILE);
	
				fs_mount(MOUNT_RO, NO_FILE);
	
				testing("Call to chown(), by process with appropriate");
				testing("privileges, to change file's owner on a read-only");
				testing("file system.");
				testing("Path name is \"%s\".", path);
				start
					expecting(EROFS);
					Zchown(path, BIN_UID, BIN_GID);
				finis
		
				testing("No change to owner or group of file.");
				start
					Zstat(path, &st_buf);
					if(st_buf.st_gid != ROOT_GID)
						error("Call to %s\n\t\"%s\"\n%s%lu\n%s%lu",
							"stat() reported unexpected group ID for path ",
							path,
							"Expected:	",
							(unsigned long) ROOT_GID,
							"Actual:	",
							(unsigned long) st_buf.st_gid);
					if(st_buf.st_uid != ROOT_UID)
						error("Call to %s\n\t\"%s\"\n%s%lu\n%s%lu",
							"stat() reported unexpected user ID for path ",
							path,
							"expected:	",
							(unsigned long) ROOT_UID,
							"actual:	",
							(unsigned long) st_buf.st_uid);
				finis
	
				prep("Fork a child.");
				start
					sv_flush();
					pid = Zfork();
				finis

				if(pid == (pid_t)0) {
					exit_on(ERROR);
	
					fs_umount(ROFS_REG_FILE);
	
					fs_mount(MOUNT_RW, ROFS_REG_FILE);
	
					prep("Set file owner to real user ID.");
					start
						Zchown(path, getuid(), getgid());
					finis
	
					fs_umount(NO_FILE);
	
					fs_mount(MOUNT_RO, NO_FILE);
	
					prep("Set effective user ID to real user ID.");
					start
						Zsetuid(getuid());
					finis
		
					testing("Call to chown(), by process without appropriate");
					testing("privileges, to change file's owner on read-only");
					testing("file system.");
					start
						expecting(EROFS);
						Zchown(path, getuid(), getgid());
					finis
		
					testing("No change to owner or group of file.");
					start
						Zstat(path, &st_buf);
						if(st_buf.st_gid != GID)
							error("Call to %s\n\t\"%s\"\n%s%lu\n%s%lu",
								"stat() reported unexpected group ID for path ",
								path,
								"Expected:	",
								(unsigned long) GID,
								"Actual:	",
								(unsigned long) st_buf.st_gid);
						if(st_buf.st_uid != UID)
							error("Call to %s\n\t\"%s\"\n%s%lu\n%s%lu",
								"stat() reported unexpected user ID for path ",
								path,
								"expected:	",
								(unsigned long) UID,
								"actual:	",
								(unsigned long) st_buf.st_uid);
					finis

					done();
				}
				wait_for(pid, TIME_PCTS_FS_TESTING);
			finis
		
			fs_umount(ROFS_REG_FILE);
		}
		else
			not_implemented("read-only file systems");
	finis
	
	if(PCTS_INVALID_OWNER) {
		assertion(36, BASE);
		start
			break_on(ERROR);
			if(APP_CHOWN) {
				if(!(INVALID_UID_VALUE || INVALID_GID_VALUE))
					error("%s\n%s",
					"Value for INVALID_UID_VALUE and/or || INVALID_GID_VALUE"
					"not provided in <config> file.");

				prep("Create test file.");
				start
					Zstrcpy(path, temp_path());
					Zstrcat(path, FILENAME);
					Zclose(Zcreat(path, PROT_ALL));
					Zchown(path, ROOT_UID, ROOT_GID);
				finis

				if(INVALID_UID_VALUE) {
					testing("Call to chown() with invalid user ID value.");
					start
						expecting(-1);
						expecting(EINVAL);
						Zchown(path, INVALID_UID_VALUE, ROOT_GID);
					finis

					testing("No change made to owner of the file.");
					start
						Zstat(path, &st_buf);
						if(st_buf.st_uid != ROOT_UID)
							error("%s%lu\n%s%lu",
									"Expected:	",
									(unsigned long) ROOT_UID,
									"Actual:	",
									(unsigned long) st_buf.st_uid);
					finis
				}

				if(INVALID_GID_VALUE) {
					testing("Call to chown() with invalid group ID value.");
					start
						expecting(-1);
						expecting(EINVAL);
						Zchown(path, ROOT_UID, INVALID_GID_VALUE);
					finis

					testing("No change made to group owner of the file.");
					start
						Zstat(path, &st_buf);
						if(st_buf.st_gid != ROOT_GID)
							error("%s%lu\n%s%lu",
									"Expected:	",
									(unsigned long) ROOT_GID,
									"Actual:	",
									(unsigned long) st_buf.st_gid);
					finis
				}
			}
			else
				not_implemented(
						"mechanism to create process to change owner of file");
		finis
	}
	else
		assertion(36, NOT_TESTABLE);

	assertion(37, EXTENDED);

	done();
}
