#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	TIME_TO_DELAY	(4)
#define	TIMING_SLOP		(1)
#define	FILENAME		"5.3.2_27"
#define	FIFONAME		"5.3.2_27_fifo"
#define	String			"DaTa"
#define	Strlen			(sizeof(String) -1)


int main(void)
{
	struct	stat st_buf, st_buf2;
	char	path[_POSIX_PATH_MAX +1];
	int		fd, fd_rd, fd_wr;
	pid_t	pid;
	char	inbuf[Strlen +1];
	int		trunc;

	setup(E_CREAT);
	funct(E_CREAT);

	assertion(27, BASE);
	start
		exit_on(ERROR);
		prep("Establish path to test data file and ensure file doesn't exist.");
		start
			Zstrcpy(path, data_path());
							/*	Have file created in a directory whose group
								owner is different than the process's
								effective group ID.
							*/
			Zstrcat(path, "group_bin.d");
			Zstrcat(path, "/test_grp");
			unlink(path);
		finis

		testing("When path does not refer to an existing file, then a call");
		testing("to creat() creates the file.");
		start
			expecting(FAILURE);
			Zstat(path, &st_buf);
			expecting(SUCCESS);
			fd = Zcreat(path, PROT_ALL);
			expecting(SUCCESS);
			Zstat(path, &st_buf);
		finis
	finis

	assertion(28, BASE);
	start
		testing("User ID of newly created file is set to process's effective");
		testing("user ID.");
		start
			if(st_buf.st_uid != getuid())
				error("%s\n%s%lu\n%s%lu.\n",
					"The new file user ID is incorrect.",
					"Expected:	",
					(unsigned long) getuid(),
					"Actual:	",
					(unsigned long) st_buf.st_uid);
		finis
	finis

	assertion(30, BASE);
	start
		prep("Set creation mask to allow all file permission bits.");
		prep("Create a new file with all permission bits set.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			unlink(path);
			(void) umask((mode_t)0);
			fd = Zcreat(path, PROT_ALL);
			Zstat(path, &st_buf);
		finis

		testing("New file has proper mode.");
		start
			if ((st_buf.st_mode & PROT_ALL) != PROT_ALL)
				error("%s\n%s%lu\n%s%lu\n",
				  		"Unexpected mode.",
					  	"Expected:	",
					  	PROT_ALL,
					  	"Actual:	",
					  	(unsigned long) (st_buf.st_mode & PROT_ALL));
		finis

		cleanup("Close and unlink test file.");
		start
			Zclose(fd);
			Zunlink(path);
		finis

		prep("Set creation mask to inhibit all file permission bits.");
		prep("Creat a new file with all permission bits set.");
		start
			(void) umask(PROT_ALL);
			fd = Zcreat(path, PROT_ALL);
			Zstat(path, &st_buf);
		finis

		testing("New file has proper mode.");
		start
			if ((st_buf.st_mode & PROT_ALL) != (mode_t) 0)
				error("%s\n%s%d\n%s%lu\n",
				  		"Unexpected mode.",
					  	"Expected:	",
					  	0,
					  	"Actual:	",
					  	(unsigned long) (st_buf.st_mode & PROT_ALL));
		finis

		cleanup("Reset creation mask to allow all file permission bits.");
		cleanup("Close and unlink test file.");
		start
			(void) umask(0);
			Zclose(fd);
			Zunlink(path);
		finis
	finis

	assertion(31, BASE);
	start
	  ENABLE_PSYNC
		prep("Establish FIFO test file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "FIFONAME");
			Zmkfifo(path, PROT_ALL);
		finis

		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis
	
		if(pid == (pid_t)0) {
			prep("Child opens FIFO with O_RDONLY and O_NONBLOCK set.");
			start
				sv_flush();
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM

				SLEEP(TIME_TO_DELAY);
				Zopen(path, O_RDONLY | O_NONBLOCK);
				SLEEP(WAIT_FOR_TIME);
			finis

			done();
		}

		prep("Call creat() on FIFO's write channel.");
		start
			SEND_PSYNC
			SET_ALARM(TIMEOUT)
				fd_wr = Zcreat(path, PROT_ALL);
			CLEAR_ALARM
		finis

		testing("When the named file is a FIFO, then a call to creat() blocks");
		testing("until a process opens the file for reading.");
		start
			if(unused_time > (MAX_TIME_TO_RUN -TIME_TO_DELAY +TIMING_SLOP))
					error("The call to creat() did not block.");
		finis

		WAIT_FOR_PID(pid);

		cleanup("Close open FIFO file.");
		start
			Zclose(fd_wr);
		finis
	  DISABLE_PSYNC
	finis

	assertion(32, EXTENDED);
	assertion(33, EXTENDED);

	assertion(34, BASE);
	start
		prep("Create a test file with data.  Close it, then");
		prep("call creat() to open the test file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zopen(path, O_WRONLY | O_CREAT | O_TRUNC, PROT_ALL);
			Zwrite(fd, String, Strlen);
			Zclose(fd);
			Zstat(path, &st_buf);

			fd = Zcreat(path, PROT_ALL);
			Zstat(path, &st_buf2);
		finis
	
		testing("Length of test file truncated to zero.");
		start
			if(st_buf2.st_size != 0)
				error("%s\n%s\n%s%lu.\n",
					"The test file has incorrect length.",
					"Expected file length of 0.",
					"Actual file length is ",
					(unsigned long) st_buf2.st_size);
		finis

		testing("File permission bits of test file not changed.");
		start
			if(st_buf2.st_mode != st_buf.st_mode)
				error("%s\n%s%lu\n%s%lu.\n",
					"The test file mode has changed.",
					"Expected file mode:	",
					(unsigned long) st_buf.st_mode,
					"Actual file mode:	",
					(unsigned long) st_buf2.st_mode);
		finis

		testing("Owner of test file not changed.");
		start
			if( st_buf2.st_uid != st_buf.st_uid )
				error("%s\n%s%lu\n%s%lu.\n",
					"The test file uid has changed.",
					"Expected file uid:	",
					(unsigned long) st_buf.st_uid,
					"Actual file uid:	",
					(unsigned long) st_buf2.st_uid);
			if( st_buf2.st_gid != st_buf.st_gid )
				error("%s\n%s%lu\n%s%lu.\n",
					"The test file gid has changed.",
					"Expected file gid:	",
					(unsigned long) st_buf.st_gid,
					"Actual file gid:	",
					(unsigned long) st_buf2.st_gid);
		finis

		cleanup("Close and unlink test file.");
		start
			Zclose(fd);
			Zunlink(path);
		finis
	finis

	assertion(35, BASE);
	start
		prep("Place data in FIFO file.");
		prep("Call creat() on FIFO file, then attempt to read FIFO.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "FIFONAME");

			SET_ALARM(TIMEOUT)
				fd_rd = Zopen(path, O_RDONLY | O_NONBLOCK);
				fd_wr = Zopen(path, O_WRONLY);
				Zwrite(fd_wr, String, Strlen);

				fd = Zcreat(path, PROT_ALL);
				trunc = TRUE;
				if(read(fd_rd, inbuf, Strlen) == Strlen)
					trunc = FALSE;

				Zclose(fd);
			CLEAR_ALARM
		finis

		testing("Call to creat() on FIFO file produces same result");
		testing("as call to open() with the O_TRUNC flag clear.");
		start
			SET_ALARM(TIMEOUT)
				Zwrite(fd_wr, String, Strlen);

				fd = Zopen(path, O_WRONLY);
				if(trunc == FALSE) {
					expecting(Strlen);
					Zread(fd_rd, inbuf, Strlen);
				}
				else {
					expecting(EAGAIN);
					Zread(fd_rd, inbuf, Strlen);
				}
				Zclose(fd);
			CLEAR_ALARM

		finis

		testing("Call to creat() on FIFO file produces same result");
		testing("as call to open() with the O_TRUNC flag set.");
		start
			SET_ALARM(TIMEOUT)
				Zwrite(fd_wr, String, Strlen);

				fd = Zopen(path, O_WRONLY | O_TRUNC);
				if(trunc == FALSE) {
					expecting(Strlen);
					Zread(fd_rd, inbuf, Strlen);
				}
				else {
					expecting(EAGAIN);
					Zread(fd_rd, inbuf, Strlen);
				}
				Zclose(fd);
				Zclose(fd_wr);
				Zclose(fd_rd);
			CLEAR_ALARM
		finis
	finis

	done();
}
