#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define	FILENAME	"5.3.1_30"
#define	String		"DaTa"

int main(void)
{
	struct	stat st_buf;
	char	path[_POSIX_PATH_MAX +1];
	int		fd;

	setup(E_OPEN);
	funct(E_OPEN);

	assertion(30, BASE);
	start
		exit_on(ERROR);
		prep("Establish path to test data file and ensure file doesn't exist.");
		start
			Zstrcpy(path, data_path());
							/*	Create a new file in directory whose group
								owner is different than the process's
								effective group ID.
							*/
			Zstrcat(path, "group_bin.d");
			Zstrcat(path, "/test_grp");
			unlink(path);
		finis

		testing("When path does not refer to an existing file and O_CREAT");
		testing("flag is set, call to open() creates a new file.");
		start
			expecting(FAILURE);
			Zstat(path, &st_buf);
			expecting(SUCCESS);
			fd = Zopen(path, O_WRONLY | O_CREAT, PROT_ALL);
			expecting(SUCCESS);
			Zstat(path, &st_buf);
		finis
	finis

	assertion(31, BASE);
	start
		testing("User ID of newly created file is set to process's effective");
		testing("user ID.");
		start
			if(st_buf.st_uid != getuid())
				error("%s\n%s%lu\n%s%lu.\n",
					"The new file user ID is incorrect.",
					"Expected:	",
					(unsigned long) getuid(),
					"Actual:	",
					(unsigned long) st_buf.st_uid);
		finis
	finis

	assertion(33, BASE);
	start
		prep("Set creation mask to allow all file permission bits.");
		prep("Open a new file with all permission bits set.");
		start
			unlink(path);
			(void) umask((mode_t)0);
			fd = Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
			Zstat(path, &st_buf);
		finis

		testing("New file has proper mode.");
		start
			if ((st_buf.st_mode & PROT_ALL) != PROT_ALL)
				error("%s\n%s%lu\n%s%lu\n",
				  		"Unexpected mode.",
					  	"Expected:	",
					  	PROT_ALL,
					  	"Actual:	",
					  	(unsigned long) (st_buf.st_mode & PROT_ALL));
		finis

		cleanup("Close and unlink test file.");
		start
			Zclose(fd);
			Zunlink(path);
		finis

		prep("Set creation mask to inhibit all file permission bits.");
		prep("Open a new file with all permission bits set.");
		start
			(void) umask(PROT_ALL);
			fd = Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
			Zstat(path, &st_buf);
		finis

		testing("New file has proper mode.");
		start
			if ((st_buf.st_mode & PROT_ALL) != (mode_t) 0)
				error("%s\n%s%d\n%s%lu\n",
				  		"Unexpected mode.",
					  	"Expected:	",
					  	0,
					  	"Actual:	",
					  	(unsigned long) (st_buf.st_mode & PROT_ALL));
		finis

		cleanup("Reset creation mask to allow all file permission bits.");
		cleanup("Close and unlink test file.");
		start
			(void) umask(0);
			Zclose(fd);
			Zunlink(path);
		finis
	finis

	assertion(34, EXTENDED);

	assertion(42, BASE);
	start
		int		loop;
		struct	stat st_buf2;

		for (loop=0; loop<2; loop++) {
			prep("Create a test file and write data to it.");
			start
				fd = Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
				Zwrite(fd, String, 4);
				Zclose(fd);
				Zstat(path, &st_buf);
			finis
				
			if(loop) {
				prep("Open test file with O_RDWR and O_TRUNC flag set.");
				start
					fd = Zopen(path, O_RDWR | O_TRUNC);
				finis
			} else {
				prep("Open test file with O_WRONLY and O_TRUNC flag set.");
				start
					fd = Zopen(path, O_WRONLY | O_TRUNC);
				finis
			}
	
			Zstat(path, &st_buf2);
	
			testing("Length of test file truncated to zero.");
			start
				if ( st_buf2.st_size != 0 )
							error("%s\n%s\n%s%lu.\n",
						"The test file has incorrect length.",
						"Expected file length of 0.",
						"Actual file length is ",
						(unsigned long) st_buf2.st_size);
			finis
	
			testing("File permission bits of test file not changed.");
			start
				if( st_buf2.st_mode != st_buf.st_mode )
					error("%s\n%s%lu\n%s%lu.\n",
						"The test file mode has changed.",
						"Expected file mode:	",
						(unsigned long) st_buf.st_mode,
						"Actual file mode:	",
						(unsigned long) st_buf2.st_mode);
			finis

			testing("Owner of test file not changed.");
			start
				if( st_buf2.st_uid != st_buf.st_uid )
					error("%s\n%s%lu\n%s%lu.\n",
						"The test file uid has changed.",
						"Expected file uid:	",
						(unsigned long) st_buf.st_uid,
						"Actual file uid:	",
						(unsigned long) st_buf2.st_uid);
				if( st_buf2.st_gid != st_buf.st_gid )
					error("%s\n%s%lu\n%s%lu.\n",
						"The test file gid has changed.",
						"Expected file gid:	",
						(unsigned long) st_buf.st_gid,
						"Actual file gid:	",
						(unsigned long) st_buf2.st_gid);
			finis
	
			cleanup("Close and unlink test file.");
			start
				Zclose(fd);
				Zunlink(path);
			finis
		}
	finis
	
	done();
}
