#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Ztermios.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	FIFONAME	"5.7.1a_20_f"
#define	FILENAME	"5.7.1a_20"
#define	MORE	5


int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	long	ret_val;
	pid_t	pid;
	size_t	pipe_buf_sz, fifo_buf_sz, max_buf_sz, dir_buf_sz;

	setup(E_PATHCONF);
	funct(E_PATHCONF);

	assertion(20, BASE);
	start
		testing("Call to pathconf(path, _PC_LINK_MAX) either returns a");
		testing("value >= {_POSIX_LINK_MAX} or returns -1.");
		start
			Zstrcpy(path, temp_path());		/* Path points to a directory. */

			errno = 0;
			ret_val = pathconf(path, _PC_LINK_MAX);
			if(ret_val == -1) {
				if(errno != 0)
					error("%s\n%s[%s].",
						"Wrong return value from pathconf(path, _PC_LINK_MAX).",
						"Returned \"-1\" with errno of ",
						err_lookup(errno));
			}
			else
				if(_POSIX_LINK_MAX > ret_val)
					error("%s[%ld].\n%s%ld.",
						"Expected minimum value to be {_POSIX_LINK_MAX} ",
						(long) _POSIX_LINK_MAX,
						"Actual return value was ",
						ret_val);
		finis
	finis

	assertion(21, BASE);
	start
#ifdef	LINK_MAX
		testing("Call to pathconf(path, _PC_LINK_MAX) returns a");
		testing("value >= LINK_MAX defined in <limits.h>.");
		start
			if(LINK_MAX > (ret_val = pathconf(path, _PC_LINK_MAX)) )
				error("%s[%ld].\n%s%ld.",
					"Expected minimum value of {LINK_MAX} ",
					(long) LINK_MAX,
					"Actual return value was ",
					ret_val);
		finis
#else
		not_implemented("LINK_MAX defined in <limits.h>");
#endif
	finis

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(22, BASE);
		GTI_NOT_PROVIDED_RESULT(23, BASE);
		GTI_NOT_PROVIDED_RESULT(24, BASE);
		GTI_NOT_PROVIDED_RESULT(25, BASE);
	}
	else {
		assertion(22, BASE);
		start
			testing("Call to pathconf(path, _PC_MAX_CANON) either returns a");
			testing("value >= {_POSIX_MAX_CANON} or returns -1.");
			start
				Zstrcpy(path, PORT_TTY);

				errno = 0;
				ret_val = pathconf(path, _PC_MAX_CANON);
				if(ret_val == -1) {
					if(errno != 0)
						error("Returned -1 with errno of %s.",
											err_lookup(errno));
				}
				else
					if(_POSIX_MAX_CANON > ret_val)
						error("%s[%ld].\n%s%ld.",
							"Expected minimum value was {_POSIX_MAX_CANON} ",
							(long) _POSIX_MAX_CANON,
							"Actual return value was ",
							ret_val);
			finis
		finis

		assertion(23, BASE);
		start
#ifdef	MAX_CANON
			testing("Call to pathconf(path, _PC_MAX_CANON) returns a");
			testing("value >= MAX_CANON defined in <limits.h>.");
			start
				Zstrcpy(path, PORT_TTY);

				if(MAX_CANON > (ret_val = pathconf(path, _PC_MAX_CANON)) )
					error("%s[%ld].\n%s%ld.",
						"Expected minimum value to be {MAX_CANON} ",
						(long) MAX_CANON,
						"Actual return value was ",
						ret_val);
			finis
#else
			not_implemented("MAX_CANON defined in <limits.h>");
#endif
		finis

		assertion(24, BASE);
		start
			testing("Call to pathconf(path, _PC_MAX_INPUT) either returns a");
			testing("value >= {_POSIX_MAX_INPUT} or returns -1.");
			start
				Zstrcpy(path, PORT_TTY);

				errno = 0;
				ret_val = pathconf(path, _PC_MAX_INPUT);
				if(ret_val == -1) {
					if(errno != 0)
						error("Returned -1 with errno of %s.",
										err_lookup(errno));
				}
				else
					if(_POSIX_MAX_INPUT > ret_val)
						error("%s[%ld].\n%s%ld.",
							"Expected minimum value to be {_POSIX_MAX_INPUT} ",
							(long) _POSIX_MAX_INPUT,
							"Actual return value was ",
							ret_val);
			finis
		finis

		assertion(25, BASE);
		start
#ifdef	MAX_INPUT
			testing("Call to pathconf(path, _PC_MAX_INPUT) returns a");
			testing("value >= MAX_INPUT defined in <limits.h>.");
			start
				Zstrcpy(path, PORT_TTY);

				if(MAX_INPUT > (ret_val = pathconf(path, _PC_MAX_INPUT)) )
					error("%s[%ld].\n%s%ld.",
						"Expected minimum value is {MAX_INPUT} ",
						(long) MAX_INPUT,
						"Actual return value was ",
						ret_val);
			finis
#else
		not_implemented("MAX_INPUT defined in <limits.h>");
#endif
		finis
	}

	assertion(26, BASE);
	start
		testing("Call to pathconf(path, _PC_NAME_MAX) either returns a");
		testing("value >= {_POSIX_NAME_MAX} or returns -1.");
		start
			Zstrcpy(path, temp_path());

			errno = 0;
			ret_val = pathconf(path, _PC_NAME_MAX);
			if(ret_val == -1) {
				if(errno != 0)
					error("Returned -1 with errno of %s.",
										err_lookup(errno));
			}
			else
				if(_POSIX_NAME_MAX > ret_val)
					error("%s[%ld].\n%s%ld.",
						"Expected minimum value to be {_POSIX_NAME_MAX} ",
						(long) _POSIX_NAME_MAX,
						"Actual return value was ",
						ret_val);
		finis
	finis

	assertion(27, BASE);
	start
#ifdef	NAME_MAX
		testing("Call to pathconf(path, _PC_NAME_MAX) returns a");
		testing("value >= NAME_MAX defined in <limits.h>.");
		start
			Zstrcpy(path, temp_path());

			if(NAME_MAX > (ret_val = pathconf(path, _PC_NAME_MAX)) )
				error("%s[%ld].\n%s%ld.",
					"Expected minimum value to be {NAME_MAX} ",
					(long) NAME_MAX,
					"Actual return value was ",
					ret_val);
		finis
#else
		not_implemented("NAME_MAX defined in <limits.h>");
#endif
	finis

	assertion(28, BASE);
	start
		FORK
			char	*pathm_u, *ptr, *ptr_end;
			long	name_lng;
		
			exit_on(ERROR);

			prep("Obtain value for {NAME_MAX} from pathconf().");
			start
				Zstrcpy(path, temp_path());
				Zchdir(path);
		
				name_lng = pathconf("./", _PC_NAME_MAX);
				if((name_lng == (long) -1) || (name_lng >= PCTS_NAME_MAX))
					name_lng = PCTS_NAME_MAX;
			finis
		
			prep("Create a filename of %ld bytes.", name_lng);
			start
				pathm_u = (char *) Zmalloc((size_t)(name_lng +1));
				ptr = pathm_u;
				ptr_end = ptr +name_lng;
				for ( ; ptr < ptr_end; ptr++ )
		        		*ptr = 'E';
				*ptr = '\000';
			finis

			testing("Value returned by pathconf(path, _PC_NAME_MAX) applies");
			testing("to filenames within the directory.");
			start
				Zclose(Zcreat(pathm_u, PROT_ALL));
				Zaccess(pathm_u, F_OK);
			finis
		KROF
	finis

	assertion(29, BASE);
	start
		testing("Call to pathconf(path, _PC_PATH_MAX) either returns a");
		testing("value >= {_POSIX_PATH_MAX} or returns -1.");
		start
			Zstrcpy(path, temp_path());

			errno = 0;
			ret_val = pathconf(path, _PC_PATH_MAX);
			if(ret_val == -1) {
				if(errno != 0)
					error("Returned -1 with errno of %s.",
										err_lookup(errno));
			}
			else
				if(_POSIX_PATH_MAX > ret_val)
					error("%s[%ld].\n%s%ld.",
						"Expected minimum value to be {_POSIX_PATH_MAX} ",
						(long) _POSIX_PATH_MAX,
						"Actual return value was ",
						ret_val);
		finis
	finis

	assertion(30, BASE);
	start
#ifdef	PATH_MAX
		testing("Call to pathconf(path, _PC_PATH_MAX) returns a");
		testing("value >= PATH_MAX defined in <limits.h>.");
		start
			if(PATH_MAX > (ret_val = pathconf(path, _PC_PATH_MAX)) )
				error("%s[%ld].\n%s%ld.",
					"Expected minimum value to be {PATH_MAX} ",
					(long) PATH_MAX,
					"Actual return value was ",
					ret_val);
		finis
#else
		not_implemented("PATH_MAX defined in <limits.h>");
#endif
	finis

	assertion(31, BASE);
	start
		char	*pathm_u;

		FORK
			pathm_u =  g_path_max();		/* Generate maximum pathlength. */

			testing("Value returned by pathconf(path, _PC_PATH_MAX) applies");
			testing("to pathnames within the directory.");
			start
				Zclose(Zcreat(pathm_u, PROT_ALL));

				Zunlink(pathm_u);
			finis
		KROF

		FORK
			pathm_u =  g_lpath();		/* Generate greater pathlength. */

			testing("Value returned by pathconf(path, _PC_PATH_MAX) applies");
			testing("to pathnames within the directory.");
			start
				expecting(ENAMETOOLONG);
				Zcreat(pathm_u, PROT_ALL);
			finis
		KROF
	finis

	assertion(32, BASE);
	start
		testing("Call to pathconf(path, _PC_PIPE_BUF) either returns a");
		testing("value >= {_POSIX_PIPE_BUF} or returns -1.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FIFONAME);
			Zmkfifo(path, PROT_ALL);

			errno = 0;
			ret_val = pathconf(path, _PC_PIPE_BUF);
			if(ret_val == -1) {
				if(errno != 0)
					error("Returned -1 with errno of %s.", err_lookup(errno));
			}
			else
				if(_POSIX_PIPE_BUF > ret_val)
					error("%s[%ld].\n%s%ld.",
						"Expected minimum value to be {_POSIX_PIPE_BUF} ",
						(long) _POSIX_PIPE_BUF,
						"Actual return value was ",
						ret_val);
		finis
	finis

	assertion(33, BASE);
	start
#ifdef	PIPE_BUF
		testing("Call to pathconf(path, _PC_PIPE_BUF) returns a");
		testing("value >= PIPE_BUF defined in <limits.h>.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FIFONAME);

			if(PIPE_BUF > (ret_val = pathconf(path, _PC_PIPE_BUF)))
				error("%s[%ld].\n%s%ld.",
					"Expected minimum value is {PIPE_BUF} ",
					(long) PIPE_BUF,
					"Actual return value was ",
					ret_val);
		finis
#else
		not_implemented("PIPE_BUF defined in <limits.h>");
#endif
	finis

	assertion(34, BASE);
	start
		char	fifop[_POSIX_PATH_MAX +1];
		char	*buf_in, *buf_out;
		int 	fiford, fifowr;

		Zstrcpy(fifop, temp_path());
		Zstrcat(fifop, FIFONAME);

		max_buf_sz = g_pipe_buf(fifop, &pipe_buf_sz, &fifo_buf_sz);

		prep("Allocate and initialize I/O buffers.");
		start
			buf_in = (char *) Zmalloc(fifo_buf_sz +MORE);
			buf_out = (char *) Zmalloc(fifo_buf_sz +MORE);
		finis

		prep("Initialize input and output data buffers.");
		start
			int	i;

			for(i = 0; i < fifo_buf_sz +MORE; ++i)
				buf_out[i] = i % CHAR_MAX;
		finis

		FORK
			prep("Open FIFO file.");
			start
				fiford = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifowr =  Zopen(fifop, O_WRONLY | O_NONBLOCK);
				Zclose(fiford);
				fiford = Zopen(fifop, O_RDONLY);
			finis

			testing("Write [PIPE_BUF] bytes to FIFO.");
			start
				SET_ALARM(TIMEOUT)
					expecting(fifo_buf_sz);
					Zwrite(fifowr, buf_out, fifo_buf_sz);
					expecting(fifo_buf_sz);
					Zread(fiford, buf_in, fifo_buf_sz);
				CLEAR_ALARM
			finis

			testing("Write greater than [PIPE_BUF] bytes to empty FIFO.");
			start
				int	cnt;
	
				SET_ALARM(TIMEOUT)
					cnt = Zwrite(fifowr, buf_out, fifo_buf_sz +MORE);
					if(cnt < fifo_buf_sz)
						error("%s\n%s\n%s%d bytes.",
							"Write more than [PIPE_BUF] bytes to empty pipe.",
							"Expected at least [PIPE_BUF] bytes written.",
							"Actual wrote ",
							cnt);
				CLEAR_ALARM
			finis
		KROF
	finis

	assertion(35, BASE);
	start
		max_buf_sz = g_pipe_buf(temp_path(), &pipe_buf_sz, &dir_buf_sz);

		testing("Value returned by a call to pathconf(path, _PC_PIPE_BUF)");
		testing("applies to any FIFOs that exist or can be created ");
		testing("within the directory.");
		start
			if(dir_buf_sz != fifo_buf_sz)
				error("%s\n%s\n%s%ld\n%s%ld\n",
					"Value from directory and FIFO within directory not equal.",
					"Directory value was ",
					(long) dir_buf_sz,
					"FIFO value was ",
					(long) fifo_buf_sz);
		finis
	finis

	assertion(36, BASE);
	start
#ifdef	_POSIX_CHOWN_RESTRICTED
		testing("Call to pathconf(path, _PC_CHOWN_RESTRICTED) returns a");
		testing("value of {_POSIX_CHOWN_RESTRICTED}.");
		start
			if(_POSIX_CHOWN_RESTRICTED !=
							(ret_val = pathconf(path, _PC_CHOWN_RESTRICTED)) )
				error("%s[%ld].\n%s%ld.",
					"Expected return value to be {_POSIX_CHOWN_RESTRICTED} ",
					(long) _POSIX_CHOWN_RESTRICTED,
					"Actual return value was ",
					ret_val);
		finis
#else
		not_implemented("_POSIX_CHOWN_RESTRICTED defined in <unistd.h>");
#endif
	finis

						/*	Value returned by a call to
							pathconf(path, _PC_CHOWN_RESTRICTED) applies
							to any files that exist or can be created within
							the directory.
						*/
	assertion(37, BASE);
	start
		struct	stat st_buf;
		int			i;
		union		u {
						char	*p_beg;
						gid_t	*g_beg;
					} u;
		gid_t		*s_gids;
		long		n_grps;
		int			n_getgrp;

		prep("Get path of directory.");
		start
			Zstrcpy(path, temp_path());
		finis
		
		if(pathconf(path, _PC_CHOWN_RESTRICTED) != (long) -1) {
			prep("Obtain effective group ID and supplementary group IDs."); 
			start
				Zstrcat(path, FILENAME);
				Zclose(Zcreat(path, PROT_ALL));

				/* Number of supplementary groups plus possibly the effective
				   group ID if not included in the count
				*/
				n_grps = Zsysconf(_SC_NGROUPS_MAX) +1;
				/* Add +1 for adding effective group ID to array in case
				   the implementation does not provide it from getgroups()
				*/
				u.p_beg = (char *) Zmalloc((size_t)(sizeof(gid_t)*(n_grps +1)));
				n_getgrp = Zgetgroups((int)n_grps, u.g_beg);
										/* Add process group ID to list */
				s_gids = u.g_beg;
				s_gids[n_getgrp] = getegid();
			finis
		
			testing("Call to chown() sets the group to the value of argument");
			testing("group when group is the calling processes effective");
			testing("group ID or one of its supplementary group IDs.");
			start
				for(i=0; i<n_grps; i++, s_gids++) {
					Zchown(path, UID, *s_gids);
					Zstat(path, &st_buf);
					if(st_buf.st_gid != *s_gids)
						error("%s\n\t\"%s\"\n%s%d.\n%s%lu\n%s%lu\n",
				    		"Call to chown() failed to set group ID for path ",
				    		path,
							"on getgroups() array index of value ",
							i,
				    		"Expected:	",
				    		(unsigned long) *s_gids,
				    		"actual:	",
				    		(unsigned long) st_buf.st_gid);
					else
						testing("Tested index %d, GID \"%lu\"\n",
										i, (unsigned long) *s_gids);
				}
			finis
		}
		else
			error("FIPS 151-2, item 'i' not supported.");
	finis

	assertion(38, BASE);
	start
#ifdef	_POSIX_NO_TRUNC
		testing("Call to pathconf(path, _PC_NO_TRUNC) returns a");
		testing("value of {_POSIX_NO_TRUNC}.");
		start
			Zstrcpy(path, temp_path());

			if(_POSIX_NO_TRUNC != (ret_val = pathconf(path, _PC_NO_TRUNC)))
				error("%s[%ld].\n%s%ld.",
					"Expected value for {_POSIX_NO_TRUNC} ",
					(long) _POSIX_NO_TRUNC,
					"Actual return value was ",
					ret_val);
		finis
#else
		not_implemented("_POSIX_NO_TRUNC defined in <unistd.h>");
#endif
	finis

								/*	Since FIPS 151-2 requires {_POSIX_NO_TRUNC}
									be supported, this test is identical to
									the combination of tests for assertions
									#51 and #52 in pathconf().
								*/
	assertion(39, BASE);
	start
		testing("Value returned by a call to pathconf(path, _PC_NO_TRUNC)");
		testing("applies to the filenames within the directory.");
		testing("See test results for pathconf[51] and pathconf[52].");
		start
		finis
	finis

	assertion(40, BASE);
	start
#ifdef	_POSIX_VDISABLE
		if(GTI_NOT_PROVIDED)
			not_implemented("GTI devices");
		else {
			testing("Call to pathconf(path, _PC_VDISABLE) returns the");
			testing("value of _POSIX_VDISABLE in <unistd.h>.");
			start
				if((cc_t) _POSIX_VDISABLE !=
						(cc_t) (ret_val = pathconf(PORT_TTY, _PC_VDISABLE)) )
					error("%s[%lx(Hex)].\n%s%lx(Hex).",
						"Expected value for {_POSIX_VDISABLE} ",
						(long) _POSIX_VDISABLE,
						"Actual return value was ",
						ret_val);
			finis
		}
#else
		not_implemented("_POSIX_VDISABLE defined in <unistd.h>");
#endif
	finis

								/*	Since FIPS 151-2 requires {_POSIX_VDISABLE}
									be supported, this test is identical to
									the combination of tests for assertions
									interface[63] thru interface[65].
								*/
	assertion(41, BASE);
	start
		testing("Value returned by a call to pathconf(path, _PC_VDISABLE)");
		testing("can be used to disable the effects associated with the");
		testing("receipt of a special control character from the terminal.");
		testing("See test results for interface[61] thru interface[63].");
		start
		finis
	finis

	done();
}
