#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zutime.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define	FILENAME	"5.6.6_26"


int main(void)
{
	char	path[_POSIX_PATH_MAX +1], path_dir[_POSIX_PATH_MAX +1];
	time_t	s_time;

	setup(E_UTIME);
	funct(E_UTIME);

	assertion(26, BASE);
	start
		prep("Initialize path name for a file where search permission");
		prep("is denied on a component of the path prefix.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_srch.d");
			Zstrcpy(path_dir, path);
			Zstrcat(path, "/_tryme_");
			Zchmod(path_dir, S_IRUSR | S_IWUSR);
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/

			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
		finis

		testing("Call to utime() when search permission is denied on path.");
		start
			expecting(-1);
			expecting(EACCES);
			Zutime(path, (struct utimbuf *) 0);
		finis

		testing("File times not affected.");
		start
			Zchmod(path_dir, PROT_ALL);
			trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
		finis
	finis

	assertion(27, BASE);
	start
		testing("Call to utime() when effective user ID of the process does");
		testing("not match the owner of the file and write access is denied.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "bin_nowrite");
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			Ztime(&s_time);				/* Obtain minimum reference time. */

			expecting(-1);
			expecting(EACCES);
			Zutime(path, (struct utimbuf *) 0);
		finis

		testing("File times not affected.");
		start
			trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
		finis
	finis

	if(tc_name_max()) {
		assertion(28, BASE);
		start
			e_NO_TRUNC(E_UTIME);
		finis
	}
	else
		assertion(28, NOT_TESTABLE);

	if(!tc_name_max()) {
		assertion(29, BASE);
		start
			e_NO_TRUNC(E_UTIME);
		finis
	}
	else
		assertion(29, NOT_TESTABLE);

	assertion(30, BASE);
	start
		e_ENAMETOOLONG(E_UTIME);
	finis

	assertion(31, BASE);
	start
		testing("Call to utime() for a path that does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_utime_");

			expecting(-1);
			expecting(ENOENT);
			Zutime(path, (struct utimbuf *) 0);
		finis
	finis

	assertion(32, BASE);
	start
		testing("Call to utime() when a component of the path prefix does");
		testing("not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_utime_");
			Zstrcat(path, "/_utime");

			expecting(-1);
			expecting(ENOENT);
			Zutime(path, (struct utimbuf *) 0);
		finis
	finis

	assertion(33, BASE);
	start
		testing("Call to utime() when path points to an empty string.");
		start
			expecting(-1);
			expecting(ENOENT);
			Zutime("", (struct utimbuf *) 0);
		finis
	finis

	assertion(34, BASE);
	start
		testing("Call to utime() when path contains non-directory component.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));
			Zstrcat(path, "/dummy");

			expecting(-1);
			expecting(ENOTDIR);
			Zutime(path, (struct utimbuf *) 0);
		finis
	finis

	assertion(35, BASE);
	start
		struct	stat st_end;
		struct	utimbuf	u_time;

		testing("Call to utime() when calling process's effective user ID");
		testing("has write access to the file but does not match the owner");
		testing("of the file and times argument is not NULL.");
		start
			stat(path, &st_end);
			u_time.actime = st_end.st_atime;;
			u_time.modtime = st_end.st_mtime;

			Zstrcpy(path, data_path());
			Zstrcat(path, "utime_write");

			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			Ztime(&s_time);				/* Obtain minimum reference time. */

			expecting(-1);
			expecting(EPERM);
			Zutime(path, &u_time);
		finis

		testing("File times not updated.");
		start
			trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
		finis
	finis

	done();
}
