#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	String		"Woolly bully, woolly bully\n"
#define	Strlen		(sizeof(String) -1)
#include	<locks.h>

#define		FILENAME	"6.5.2_30"


int main(void)
{
	int		fd;
	pid_t	pid;
	char	tmp_buf[FILE_LENGTH];
	struct	flock flock1, flock2;

	setup(E_FCNTL);
	funct(E_FCNTL);

		/*	Before a successful return from a F_SETLK or F_SETLKW request,
			the previous lock type for each byte in the specified region is
			replaced by the new lock type.  The calling process has only one
			type of lock set for each byte in the file.
		*/
	assertion(30, BASE);
	start
		int	i;

		prep("Create a temporary file and write to it.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, tmp_buf, FILE_LENGTH);
		finis

		for(i=0; i<2; i++) {
			if(i==0) {
				prep("Set an exclusive lock using F_SETLK.");
				start
					SET_LOCK(F_WRLCK, fd, SEEK_SET, TOTAL_FILE, (off_t)1,
																	&flock1);
				finis
			}
			else {
				prep("Set an exclusive lock using F_SETLKW.");
				start
					SET_LOCKW(F_WRLCK, fd, SEEK_SET, TOTAL_FILE, (off_t)1,
																	&flock1);
				finis
			}

			FORK
				testing("Lock type set to F_WRLCK.");
				start
					GET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,
																	&flock2);
					if(flock2.l_type != F_WRLCK)
						error("%s\n%s\n%s%s\n",
							"L_type not correct.",
							"Expected l_type to be F_UNLCK.",
							"Actual l_type is ",
							(flock2.l_type == F_RDLCK) ? "F_RDLCK" :
									(flock2.l_type == F_UNLCK) ? "F_UNLCK" :
									"Unknown LOCK type");
				finis
			KROF

			if(i==0) {
				prep("Set a shared lock using F_SETLK.");
				start
					SET_LOCK(F_RDLCK, fd, SEEK_SET, TOTAL_FILE, (off_t)1,
																	&flock1);
				finis
			}
			else {
				prep("Set a shared lock using F_SETLKW.");
				start
					SET_LOCKW(F_RDLCK, fd, SEEK_SET, TOTAL_FILE, (off_t)1,
																	&flock1);
				finis
			}

			FORK
				testing("New lock type replaces old lock type.");
				start
					GET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,
																	&flock2);
					if(flock2.l_type != F_RDLCK)
						error("%s\n%s\n%s%s\n",
							"L_type not correct.",
							"Expected l_type to be F_RDLCK.",
							"Actual l_type is ",
							(flock2.l_type == F_WRLCK) ? "F_WRLCK" :
									(flock2.l_type == F_UNLCK) ? "F_UNLCK" :
									"Unknown LOCK type");
				finis
			KROF
		}

		CLOSE_LOCKED_FILE(fd);
	finis


		/*	All locks associated with a file for a given process
			shall be removed when a file descriptor for that file
			is closed.
		*/
	assertion(31, BASE);
	start
		break_on(ERROR);
		prep("Set an exclusive lock on test file then close it.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, tmp_buf, FILE_LENGTH);
			expecting(SUCCESS);	/* Report error if lock cannot be set */
			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,&flock1);
			Zclose(fd);
		finis

		FORK
			testing("Set an exclusive lock on test file after parent has");
			testing("closed it.");
			start
	  			fd = temp_open(FILENAME, O_RDWR, S_IRWXU);
				expecting(SUCCESS);	/* Report error if lock cannot be set */
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,&flock1);
			finis
		KROF
	finis


		/*	All locks associated with a file for a given process
			are removed when the process terminates.
		*/
	assertion(32, BASE);
	start
		break_on(ERROR);
		prep("Open a test file.");
		start
			fd = temp_open(FILENAME, O_RDWR, S_IRWXU);
		finis

		FORK
			prep("Exclusive lock entire file then exit.");
			start
				expecting(SUCCESS);	/* Report error if lock cannot be set */
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,&flock1);
			finis
		KROF

		testing("Test file can be locked.");
		start
			expecting(SUCCESS);	/* Report error if lock cannot be set */
			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE, &flock1);
		finis

		CLOSE_LOCKED_FILE(fd);
	finis

		/*	Locks are not inherited by a child process created using
			the fork() function.
		*/
	assertion(33, BASE);
	start
	  break_on(ERROR);
	  ENABLE_PSYNC
		prep("Create a temp file, set an exclusive lock.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, tmp_buf, FILE_LENGTH);
			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE, &flock1);
		finis

		prep("Fork a child process.");
		start
			sv_flush( );
			pid = Zfork();
		finis

		if( pid == (pid_t)0 ) {
							/* Inform parent child is ready */
			SEND_PSYNC
							/* Wait for parent to remove lock */
			SET_TIMEOUT
				WAIT_PSYNCA
			CLEAR_ALARM

			prep("Call fcntl to get lock info on file.");
			start
				GET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,&flock2);
			finis

			testing("No lock is set.");
			start
				if(flock2.l_type != F_UNLCK)
					error("%s\n%s\n%s%s\n",
						"L_type not correct.",
						"Expected l_type to be F_UNLCK.",
						"Actual l_type is ",
						(flock2.l_type == F_RDLCK) ? "F_RDLCK" :
								(flock2.l_type == F_WRLCK) ? "F_WRLCK" :
								"Unknown LOCK type");
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		prep("Have parent remove lock.");
		start
			SET_LOCK(F_UNLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE, &flock1);
			sv_flush();
		finis

		SEND_PSYNCA

		WAIT_FOR_PID(pid);

		CLOSE_LOCKED_FILE(fd);
	  DISABLE_PSYNC
	finis

	done();
}
