#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define	 String		"What goes up must come down."
#define	 Strlen		(sizeof(String) -1)
#define	 String1	"Unless it doesn't."
#define	 Strlen1	(sizeof(String1) -1)

int main(void)
{
	int 	fildes[2];
	char	inbuf[Strlen +1];
	struct	stat st_buf;

	setup(E_PIPE);
	funct(E_PIPE);

	assertion(4, BASE);
	start
		testing("Call to pipe(fildes) creates a pipe and places two file");
		testing("descriptors in array fildes and returns a value of zero.");
		start
			exit_on(ERROR);
			expecting(0);
			Zpipe(fildes);

			Zfstat(fildes[0], &st_buf);
			Zfstat(fildes[1], &st_buf);
		finis
	finis

	assertion(5, BASE);
	start
		testing("Data can be written to fildes[1].");
		start
			expecting(Strlen);
			Zwrite(fildes[1], String, Strlen);
		finis
	finis

	assertion(6, BASE);
	start
		testing("Data can be read from fildes[0].");
		start
			expecting(Strlen);
			Zread(fildes[0], inbuf, Strlen);
			inbuf[Strlen] = '\000';

			if(strncmp(String, inbuf, Strlen))
				error("%s\n%s\"%s\"\n%s\"%s\"\n",
					"Data read does not match data written.",
					"Expected: ",
					String,
					"Actual: ",
					inbuf);
		finis

		cleanup("Close pipe's file descriptors.");
		start
			Zclose(fildes[0]);
			Zclose(fildes[1]);
		finis
	finis

	assertion(7, BASE);
	start
		prep("Close file descriptors 0 and 3.");
		start
			close(0);
			close(3);
		finis

		testing("The two file descriptors placed in the array fildes by");
		testing("pipe() shall be the two lowest available at the time of");
		testing("the pipe() call.");
		start
			Zpipe(fildes);
			if((fildes[0] + fildes[1]) != 3)
				error("%s\n%s\n%s%d and %d\n",
					  "Lowest available file descriptors not returned.",
					  "Expected file descriptors of: 0 and 3.",
					  "Actual file descriptors:	",
					  fildes[0],
					  fildes[1]);
		finis
	finis

	assertion(8, BASE);
	start
		testing("The O_NONBLOCK flag is cleared on both file descriptors.");
		start
			if((Zfcntl(fildes[0], F_GETFD) & O_NONBLOCK))
				error("O_NONBLOCK flag is set for fildes[0].");

			if((Zfcntl(fildes[1], F_GETFD) & O_NONBLOCK))
				error("O_NONBLOCK flag is set for fildes[1].");
		finis
	finis

	assertion(9, BASE);
	start
		testing("The FD_CLOEXEC flag is cleared on both file descriptors.");
		start
			if((Zfcntl(fildes[0], F_GETFD) & FD_CLOEXEC))
				error("FD_CLOEXEC flag is set for fildes[0].");

			if((Zfcntl(fildes[1], F_GETFD) & FD_CLOEXEC))
				error("FD_CLOEXEC flag is set for fildes[1].");
		finis
	finis

	assertion(10, BASE);
	start
		testing("Data written to fildes[1] will be read from fildes[0]");
		testing("on a first-in first-out basis.");
		start
			expecting(Strlen);
			Zwrite(fildes[1], String, Strlen);

			expecting(Strlen1);
			Zwrite(fildes[1], String1, Strlen1);

			expecting(Strlen);
			Zread(fildes[0], inbuf, Strlen);
			inbuf[Strlen] = '\000';

			if(strncmp(String, inbuf, Strlen))
				error("%s\n%s\"%s\"\n%s\"%s\"\n",
					"Data read does not match data written.",
					"Expected: ",
					String,
					"Actual: ",
					inbuf);

			expecting(Strlen);
			Zwrite(fildes[1], String, Strlen);

			expecting(Strlen1);
			Zread(fildes[0], inbuf, Strlen1);
			inbuf[Strlen1] = '\000';

			if(strncmp(String1, inbuf, Strlen1))
				error("%s\n%s\"%s\"\n%s\"%s\"\n",
					"Data read does not match data written.",
					"Expected: ",
					String1,
					"Actual: ",
					inbuf);

			expecting(Strlen1);
			Zwrite(fildes[1], String1, Strlen1);
			expecting(Strlen1);
			Zwrite(fildes[1], String1, Strlen1);

			expecting(Strlen);
			Zread(fildes[0], inbuf, Strlen);
			inbuf[Strlen] = '\000';

			if(strncmp(String, inbuf, Strlen))
				error("%s\n%s\"%s\"\n%s\"%s\"\n",
					"Data read does not match data written.",
					"Expected: ",
					String,
					"Actual: ",
					inbuf);
		finis

		cleanup("Close pipe's file descriptors.");
		start
			Zclose(fildes[0]);
			Zclose(fildes[1]);
		finis
	finis

	assertion(11, BASE);
	start
		time_t	s_time;

		testing("Call to pipe() marks for update \"st_atime\", \"st_ctime\",");
		testing("and \"st_mtime\" time-related fields.");
		start
			Ztime(&s_time);				/* Obtain minimum reference time.	*/
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			Zpipe(fildes);

			trf_ck(CHK_ATIME, TRUE, s_time, fildes[0], NO_PATH);
			trf_ck(CHK_ATIME, TRUE, s_time, fildes[1], NO_PATH);
			trf_ck(CHK_CTIME, TRUE, s_time, fildes[0], NO_PATH);
			trf_ck(CHK_CTIME, TRUE, s_time, fildes[1], NO_PATH);
			trf_ck(CHK_MTIME, TRUE, s_time, fildes[0], NO_PATH);
			trf_ck(CHK_MTIME, TRUE, s_time, fildes[1], NO_PATH);
		finis
	finis

	done();
}
