#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FIFONAME "6.4.2_26_fifo"


int main(void)
{
	char	fifop[_POSIX_PATH_MAX +1];
	int		pipfd[2];
	char	*buf_in, *outbuf;
	int 	fifo_rd, fifo_wr;
	int		i;
	size_t	pipe_buf_sz, fifo_buf_sz, max_buf_sz, pipe_max, fifo_max;
	int		flags;
	pid_t	pid;

	setup(E_WRITE);
	funct(E_WRITE);

	exit_on(ERROR);

			/*
				When the file's type is a pipe or a FIFO, the O_NONBLOCK
				flag is clear, and nbyte is <= {PIPE_BUF}, and insufficient
				space exists in the pipe or FIFO to satisfy the write, then
				a call to write(fildes, buf, nbyte) blocks until space is
				available to complete the write() or the write is interrupted
				by a signal.
			*/
	assertion(26, BASE);
	start
		prep("Establish path to FIFO and create it.");
		start
			Zstrcpy(fifop, temp_path());
			Zstrcat(fifop, FIFONAME);
			Zmkfifo(fifop, PROT_ALL);
		finis

							/* Maximum bytes written in pipe and FIFO */
		max_buf_sz=g_max_pipe(&pipe_buf_sz, &fifo_buf_sz, &pipe_max, &fifo_max);

		prep("Allocate and initialize I/O buffers.");
		start
			buf_in = (char *) Zmalloc(max_buf_sz);
			outbuf = (char *) Zmalloc(max_buf_sz);
			memset(buf_in, '\0', max_buf_sz);

			for(i = 0; i < max_buf_sz; ++i)
				outbuf[i] = i % CHAR_MAX;
		finis

		ENABLE_PSYNC
		FORK
										/*	Part A: pipe
											Test when space becomes available.
										*/
			prep("Create pipe.");
			start
				Zpipe(pipfd);
			finis
	
			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Fill up pipe's write channel.");
				start
					size_t	n_bytes;
					size_t	nb_wr = 0;

					SET_TIMEOUT
						for(n_bytes = pipe_buf_sz; n_bytes; n_bytes /= 2)
							while(pipe_max -nb_wr >= n_bytes) {
								expecting(n_bytes);
								nb_wr += Zwrite(pipfd[1], outbuf, n_bytes);
							}
					CLEAR_ALARM
				finis

				SEND_PSYNC

				prep("Expect child to block on next write() to pipe.");
				start
					SET_ALARM(TIMEOUT)
						expecting((ssize_t)(pipe_buf_sz/2));
						Zwrite(pipfd[1], outbuf, pipe_buf_sz/2);
					CLEAR_ALARM
				finis
	
				done();
			}
	
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
	
			testing("Child blocked on write() to pipe.");
			start
				int	stat_loc;
	
				SET_ALARM(2)
					wait(&stat_loc);
				EXPECT_ALARM
			finis
	
			testing("A read() of data from the pipe allows child's write()");
			testing("to the pipe to complete.");
			start
				SET_ALARM(WAIT_FOR_TIME);
					expecting((ssize_t)(pipe_buf_sz/2));
					Zread(pipfd[0], outbuf, pipe_buf_sz/2);
				CLEAR_ALARM
	
				wait_for(pid, WAIT_FOR_TIME);
			finis
		KROF
										/*	Part A: FIFO
											Test when space becomes available.
										*/
		FORK
			prep("Open FIFO with O_NONBLOCK clear on write channel.");
			start
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY);
			finis
	
			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Fill up FIFO's write channel.");
				start
					size_t	n_bytes;
					size_t	nb_wr = 0;

					SET_TIMEOUT
						for(n_bytes = fifo_buf_sz; n_bytes; n_bytes /= 2)
							while(fifo_max -nb_wr >= n_bytes) {
								expecting(n_bytes);
								nb_wr += Zwrite(fifo_wr, outbuf, n_bytes);
							}
					CLEAR_ALARM
				finis

				SEND_PSYNC

				prep("Expect child to block on next write() to FIFO.");
				start
					SET_ALARM(TIMEOUT)
						expecting((int)fifo_buf_sz/2);
						Zwrite(fifo_wr, outbuf, fifo_buf_sz/2);
					CLEAR_ALARM
				finis
	
				done();
			}
	
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
	
			testing("Child blocked on write() to FIFO.");
			start
				int	stat_loc;
	
				SET_ALARM(WAIT_FOR_TIME)
					wait(&stat_loc);
				EXPECT_ALARM
			finis
	
			SLEEP(WAIT_FOR_TIME);
	
			testing("A read() of data from the FIFO allows child's write()");
			testing("to the FIFO to complete.");
			start
				SET_ALARM(WAIT_FOR_TIME);
					expecting((int)(fifo_buf_sz/2));
					Zread(fifo_rd, outbuf, fifo_buf_sz/2);
								/*  Read all the data written to the pipe. */
					while(read(fifo_rd, outbuf, fifo_buf_sz/2) > 0)
						;
				CLEAR_ALARM
	
				wait_for(pid, WAIT_FOR_TIME);
			finis
		KROF
										/*	Part B: pipe
											Test when interrupt occurs.
										*/
		FORK
			prep("Create pipe.");
			start
				Zpipe(pipfd);
			finis
	
			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Fill up pipe's write channel.");
				start
					size_t	n_bytes;
					size_t	nb_wr = 0;

					SIGNAL(SIGINT, sig_int);
					SET_TIMEOUT
						for(n_bytes = pipe_buf_sz; n_bytes; n_bytes /= 2)
							while(pipe_max -nb_wr >= n_bytes) {
								expecting(n_bytes);
								nb_wr += Zwrite(pipfd[1], outbuf, n_bytes);
							}
					CLEAR_ALARM
				finis

				SEND_PSYNC

				prep("Expect child to block on next write() to pipe.");
				start
					SET_ALARM(TIMEOUT)
						expecting(EINTR);
						Zwrite(pipfd[1], outbuf, pipe_buf_sz/2);
					CLEAR_ALARM
				finis
	
				done();
			}
	
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
	
			testing("Child blocked on write() to pipe.");
			start
				int	stat_loc;
	
				SET_ALARM(2)
					wait(&stat_loc);
				EXPECT_ALARM
			finis
	
			testing("A signal to child terminates the child's write().");
			start
				Zkill(pid, SIGINT);
				wait_for(pid, WAIT_FOR_TIME);
			finis
		KROF
										/*	Part B: FIFO
											Test when interrupt occurs.
										*/
		FORK
			prep("Open FIFO with O_NONBLOCK clear on write channel.");
			start
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY);
			finis
	
			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Fill up FIFO's write channel.");
				start
					size_t	n_bytes;
					size_t	nb_wr = 0;

					SIGNAL(SIGINT, sig_int);
					SET_TIMEOUT
						for(n_bytes = fifo_buf_sz; n_bytes; n_bytes /= 2)
							while(fifo_max -nb_wr >= n_bytes) {
								expecting(n_bytes);
								nb_wr += Zwrite(fifo_wr, outbuf, n_bytes);
							}
					CLEAR_ALARM
				finis

				SEND_PSYNC

				prep("Expect child to block on next write() to FIFO.");
				start
					SET_ALARM(TIMEOUT)
						expecting(EINTR);
						Zwrite(fifo_wr, outbuf, fifo_buf_sz/2);
					CLEAR_ALARM
				finis
	
				done();
			}
	
			SET_ALARM(TIMEOUT *4)
				WAIT_PSYNC
			CLEAR_ALARM
	
			testing("Child blocked on write() to FIFO.");
			start
				int	stat_loc;
	
				SET_ALARM(WAIT_FOR_TIME)
					wait(&stat_loc);
				EXPECT_ALARM
			finis
	
			SLEEP(WAIT_FOR_TIME);
	
			testing("A signal to child terminates the child's write().");
			start
				Zkill(pid, SIGINT);
				wait_for(pid, WAIT_FOR_TIME);
			finis
		KROF
		DISABLE_PSYNC
	finis

					/*	Changes are expected in the next Draft with respect
						to these conditionals.  May need a PCTS_PIPE_BUF
						and a PCTS_FIFO_BUF.
					*/
	if((pipe_buf_sz > PCTS_PIPE_BUF) && (fifo_buf_sz > PCTS_PIPE_BUF)) {
		assertion(27, NOT_TESTABLE);
		assertion(28, NOT_TESTABLE);
		assertion(29, NOT_TESTABLE);
		done();
	}
				/*	When the file's type is a pipe or a FIFO, the O_NONBLOCK
					flag is set, nbyte is greater than {PIPE_BUF} bytes, and
					at least one byte can be written, then a call to write()
					transfers what it can and returns the number of bytes
					written.
				*/
	assertion(27, BASE);
	start
		FORK
			if(pipe_buf_sz <= PCTS_PIPE_BUF) {				/* pipe test */
				prep("Create a pipe and set O_NONBLOCK flag on write channel.");
				start
					Zpipe(pipfd);
					flags = Zfcntl(pipfd[1], F_GETFL);
					Zfcntl(pipfd[1], F_SETFL, flags | O_NONBLOCK);
				finis
	
				prep("Fill up pipe's write channel.");
				start
					size_t	n_bytes;
					size_t	nb_wr = 0;
	
					SET_TIMEOUT
						for(n_bytes = pipe_buf_sz; n_bytes; n_bytes /= 2)
							while(pipe_max -nb_wr >= n_bytes) {
								expecting(n_bytes);
								nb_wr += Zwrite(pipfd[1], outbuf, n_bytes);
							}
					CLEAR_ALARM
				finis
		
				testing("When nbyte exceeds {PIPE_BUF}, a call to write()");
				testing("transfers what it can.");
				start
					ssize_t	nb_wr;

					SET_ALARM(2)
						expecting(1);
						Zread(pipfd[0], buf_in, 1);
						nb_wr = write(pipfd[1], outbuf, pipe_buf_sz +1);
						if(	(nb_wr != (ssize_t)1) &&
							(nb_wr != (ssize_t)(pipe_buf_sz +1)) ) {
							/*	Test for STREAMS based implementation.
								For pipes, after reaching the high water mark
								when writing to a pipe a read() of a single
								byte lowers the STREAM below its high water
								mark and you are free to write another
								PIPE_BUF bytes.
							*/
							if(nb_wr == (ssize_t)(pipe_buf_sz)) {
								expecting(1);
								Zread(pipfd[0], buf_in, 1);
								expecting(pipe_buf_sz);
								nb_wr = write(pipfd[1], outbuf, pipe_buf_sz +1);
							}
							else
								error("%s\n%s%ld.\n%s%ld.",
									"Unexpected value returned from write().",
									"Expected value of 1 or ",
									(long) pipe_buf_sz,
									"Actual value was ",
									(long) nb_wr);
						}
					CLEAR_ALARM
				finis
		
			}

			if(fifo_buf_sz <= PCTS_PIPE_BUF) {				/* FIFO test */
				prep("Opening FIFO and set O_NONBLOCK flag.");
				start
					fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
					fifo_wr = Zopen(fifop, O_WRONLY | O_NONBLOCK);
				finis
	
				prep("Fill up FIFO's write channel.");
				start
					size_t	n_bytes;
					size_t	nb_wr = 0;
	
					SET_TIMEOUT
						for(n_bytes = fifo_buf_sz; n_bytes; n_bytes /= 2)
							while(fifo_max -nb_wr >= n_bytes) {
								expecting(n_bytes);
								nb_wr += Zwrite(fifo_wr, outbuf, n_bytes);
							}
					CLEAR_ALARM
				finis
	
				testing("When nbyte exceeds {PIPE_BUF}, a call to write()");
				testing("transfers what it can.");
				start
					ssize_t	nb_wr;

					SET_ALARM(2)
						expecting(1);
						Zread(fifo_rd, buf_in, 1);
						nb_wr = write(fifo_wr, outbuf, fifo_buf_sz +1);
						if(	(nb_wr != (ssize_t)1) &&
							(nb_wr != (ssize_t)(fifo_buf_sz +1)) ) {
							/*	Test for STREAMS based implementation.
								For pipes, after reaching the high water mark
								when writing to a pipe a read() of a single
								byte lowers the STREAM below its high water
								mark and you are free to write another
								PIPE_BUF bytes.
							*/
							if(nb_wr == (ssize_t)(fifo_buf_sz)) {
								expecting(1);
								Zread(fifo_rd, buf_in, 1);
								expecting(fifo_buf_sz);
								nb_wr = write(fifo_rd, outbuf, fifo_buf_sz +1);
							}
							else
								error("%s\n%s%ld.\n%s%ld.",
									"Unexpected value returned from write().",
									"Expected value of 1 or ",
									(long) fifo_buf_sz,
									"Actual value was ",
									(long) nb_wr);
						}
					CLEAR_ALARM
				finis
			}
		KROF
	finis
				/*	When the file's type is a pipe or a FIFO, the O_NONBLOCK
					flag is set, nbyte is greater than {PIPE_BUF} bytes, and
					no data can be written, then a call to write() returns
					a value of [(ssize_t) -1], sets errno to [EAGAIN] and
					transfers no data.
				*/
	assertion(28, BASE);
	start
		FORK
			if(pipe_buf_sz <= PCTS_PIPE_BUF) {				/* pipe test */
				prep("Create a pipe and set O_NONBLOCK flag on write channel.");
				start
					Zpipe(pipfd);
					flags = Zfcntl(pipfd[1], F_GETFL);
					Zfcntl(pipfd[1], F_SETFL, flags | O_NONBLOCK);
				finis
	
				prep("Fill up pipe's write channel.");
				start
					size_t	n_bytes;
					size_t	nb_wr = 0;
	
					SET_TIMEOUT
						for(n_bytes = pipe_buf_sz; n_bytes; n_bytes /= 2)
							while(pipe_max -nb_wr >= n_bytes) {
								expecting(n_bytes);
								nb_wr += Zwrite(pipfd[1], outbuf, n_bytes);
							}
					CLEAR_ALARM
				finis
		
				testing("Call to write() returns [EAGAIN].");
				start
					SET_ALARM(2)
						expecting(-1);
						expecting(EAGAIN);
						Zwrite(pipfd[1], outbuf, 1);
					CLEAR_ALARM
				finis
			}

			if(fifo_buf_sz <= PCTS_PIPE_BUF) {				/* FIFO test */
				prep("Opening FIFO and set O_NONBLOCK flag.");
				start
					fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
					fifo_wr = Zopen(fifop, O_WRONLY | O_NONBLOCK);
				finis
	
				prep("Fill up FIFO's write channel.");
				start
					size_t	n_bytes;
					size_t	nb_wr = 0;
	
					SET_TIMEOUT
						for(n_bytes = fifo_buf_sz; n_bytes; n_bytes /= 2)
							while(fifo_max -nb_wr >= n_bytes) {
								expecting(n_bytes);
								nb_wr += Zwrite(fifo_wr, outbuf, n_bytes);
							}
					CLEAR_ALARM
				finis
	
				testing("Call to write() returns [EAGAIN].");
				start
					SET_ALARM(2)
						expecting(-1);
						expecting(EAGAIN);
						Zwrite(fifo_wr, outbuf, 1);
					CLEAR_ALARM
				finis
			}
		KROF
	finis
	
				/*	When the file's type is a pipe or a FIFO, the O_NONBLOCK
					flag is set, nbyte is greater than {PIPE_BUF} bytes, and
					all data written to the pipe or FIFO has been read, then
					a call to write() transfers at least {PIPE_BUF} bytes.
				*/
	assertion(29, BASE);
	start
		unsigned long	cnt;
		FORK
			if(pipe_buf_sz <= PCTS_PIPE_BUF) {				/* pipe test */
				prep("Create a pipe and set O_NONBLOCK flag on write channel.");
				start
					Zpipe(pipfd);
					flags = Zfcntl(pipfd[1], F_GETFL);
					Zfcntl(pipfd[1], F_SETFL, flags | O_NONBLOCK);
				finis
	
				testing("A write() to a pipe when nbyte greater than {PIPE_BUF}");
				testing("transfers at least {PIPE_BUF} bytes.");
				start
					SET_ALARM(WAIT_FOR_TIME)
						cnt = Zwrite(pipfd[1], outbuf, pipe_buf_sz +1);
					CLEAR_ALARM
					if(cnt < pipe_buf_sz)
						error("%s\n%s(%lu).\n%s%lu.",
							"write() transferred less than {PIPE_BUF} bytes.",
							"Expected {PIPE_BUF} ", 
							(unsigned long)pipe_buf_sz,
							"Actual bytes transfered was ",
							cnt);
				finis
			}

			if(fifo_buf_sz <= PCTS_PIPE_BUF) {				/* FIFO test */
				prep("Opening FIFO and set O_NONBLOCK flag.");
				start
					fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
					fifo_wr = Zopen(fifop, O_WRONLY | O_NONBLOCK);
				finis
	
				testing("A write() to FIFO when nbyte greater than {PIPE_BUF}");
				testing("transfers at least {PIPE_BUF} bytes.");
				start
					SET_ALARM(WAIT_FOR_TIME)
						cnt = Zwrite(fifo_wr, outbuf, fifo_buf_sz +1);
					CLEAR_ALARM
					if(cnt < fifo_buf_sz)
						error("%s\n%s(%lu).\n%s%lu.",
							"write() transferred less than {PIPE_BUF} bytes.",
							"Expected {PIPE_BUF} ", 
							(unsigned long)fifo_buf_sz,
							"Actual bytes transfered was ",
							cnt);
				finis
			}
		KROF
	finis
	
	done();
}
