#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>
#include	<locks.h>

#define FILENAME "8.1.49_05"
char	String[] = "Hello, world!";
#define	Strlen	(sizeof(String) -1)
#define	Status	0xFF


int	main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	int 	fd;
	struct	flock flock;
	pid_t	pid;
	int		stat_val;
	int		i;
	sigset_t	set;

	setup(E_ABORT);
	funct(E_ABORT);

	SIGNAL(SIGABRT, sig_int);		/* Insurance */

	assertion(5, BASE);
	start
		for(i=0; i<3; i++) {
			FORK
				if(i==0) {
					SIGNAL(SIGABRT, SIG_IGN);
				} else {
					if(i==1) {
						SIGBLOCK(set, SIGABRT);
					} else {
						SIGNAL(SIGABRT, sig_int);
					}
				}

				testing("Call to abort() ceases the execution of subsequent");
				testing("program statements when SIGABRT signal is %s.",
								(i==0) ? "being ignored" :
								(i==1) ? "being blocked" :
								"caught and returns");
				start
					sv_flush();
					abort();

					error("Process continued after abort() was called.");
				finis
			KROF
		}
	finis


				/*	File descriptors in calling process are closed.	*/
	assertion(6, BASE);
	start
		int	pipfd[2];

		prep("Create a regular file for testing.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zopen(path, O_CREAT | O_RDWR, PROT_ALL);
		finis

		ENABLE_PSYNC
		for(i=0; i<3; i++) {
			flock.l_type = F_WRLCK;
			flock.l_whence = SEEK_SET;
			flock.l_start = (off_t)0;
			flock.l_len = (off_t)0;

			prep(" Create a pipe and fork a child process.");
			start
				Zpipe(pipfd);
				sv_flush();
				pid = Zfork();
			finis
	
		   	if(pid == (pid_t)0) {
				if(i==0) {
					SIGNAL(SIGABRT, SIG_IGN);
				} else {
					if(i==1) {
						SIGBLOCK(set, SIGABRT);
					} else {
						SIGNAL(SIGABRT, sig_int);
					}
				}

				prep("Child process locks regular file.");
				start
					Zfcntl(fd, F_SETLK, &flock);
					SEND_PSYNC
				finis
	
				prep("Child process delays then calls abort().");
				start
					SET_TIMEOUT
						WAIT_PSYNCA
					CLEAR_ALARM

					abort();
					error("Process continued after abort() was called.");
				finis
	
				done();
			}
	
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
	
			prep("Regular file not accessible to parent for locking.");
			start
				Zfcntl(fd, F_GETLK, &flock);
	
				if(flock.l_type != F_WRLCK)
					error("Regular file accesible for locking.");
			finis
	
			prep("Parent closes pipe for reading, writes to child");
			prep("then waits for child to exit.");
			start
				Zclose(pipfd[0]);
	
				SET_TIMEOUT
					Zwrite(pipfd[1], String, Strlen);
					SEND_PSYNCA
					Zwaitpid(pid, &stat_val, 0);
				CLEAR_ALARM
			finis
	
			testing("On a call to abort() open file descriptors of child");
			testing("process are closed when SIGABRT signal is %s.",
								(i==0) ? "being ignored" :
								(i==1) ? "being blocked" :
								"caught and returns");
			start
				SIGNAL(SIGPIPE, SIG_IGN);		/* Ignore expected SIGPIPE */
				expecting(EPIPE);	/*  No process has pipe open for reading. */
				expecting(FAILURE);
				Zwrite(pipfd[1], String, Strlen);
	
				expecting(SUCCESS);
				Zfcntl(fd, F_GETLK, &flock);		/* File can be locked. */
	
				if(flock.l_type != F_UNLCK)
					error("Child's regular file descriptor was not closed.");

				Zclose(pipfd[1]);
			finis
		}
	
		cleanup("Close and unlink open files.");
		start
			Zclose(fd);
			Zunlink(path);
		finis
		DISABLE_PSYNC
	finis
	
	assertion(7, EXTENDED);

	assertion(8, BASE);
	start
		for(i=0; i<3; i++) {
			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				if(i==0) {
					SIGNAL(SIGABRT, SIG_IGN);
				} else {
					if(i==1) {
						SIGBLOCK(set, SIGABRT);
					} else {
						SIGNAL(SIGABRT, sig_int);
					}
				}
				abort();
		
				error("Process continued after abort() was called.");
				done();
			}
	
	    	testing("When parent process is waiting, the parent is notified");
			testing("of the child's termination when SIGABRT signal is %s.",
								(i==0) ? "being ignored" :
								(i==1) ? "being blocked" :
								"caught and returns");
			start
				if(i==1)
					Zwaitpid(pid, &stat_val, 0);		/* Test via waitpid() */
				else
					Zwait(&stat_val);					/* Test via wait() */
			finis

	    	testing("Status returned is that of a process terminated by");
			testing("the SIGABRT signal (Assertion #15).");
			start
				if( ! WIFSIGNALED(stat_val))
		   			error("Status returned says no signal terminated process.");
				if(WTERMSIG(stat_val) != SIGABRT)
		   			error("Process terminated with %s.",
										sig_lookup(WTERMSIG(stat_val)));
			finis
		}
	finis

	assertion(15, BASE);
	start
	    testing("Tested as part of the previous test.");
		start
		finis
	finis

	assertion(9, BASE);
	start
		for(i=0; i<3; i++) {
			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				if(i==0) {
					SIGNAL(SIGABRT, SIG_IGN);
				} else {
					if(i==1) {
						SIGBLOCK(set, SIGABRT);
					} else {
						SIGNAL(SIGABRT, sig_int);
					}
				}
				abort();
		
				error("Process continued after abort() was called.");
				done();
			}

			prep("Parent delays to ensure child has exited.");
			start
				SLEEP(WAIT_FOR_TIME +1);
			finis
	
	    	testing("When parent process is not waiting, parent is notified");
			testing("of the child's termination when SIGABRT signal is %s.",
								(i==0) ? "being ignored" :
								(i==1) ? "being blocked" :
								"caught and returns");
			start
				if(i==1)
					Zwaitpid(pid, &stat_val, 0);		/* Test via waitpid() */
				else
					Zwait(&stat_val);					/* Test via wait() */
			finis

	    	testing("Status returned is that of a process terminated by");
			testing("the SIGABRT signal (Assertion #15).");
			start
				if( ! WIFSIGNALED(stat_val))
		   			error("Status returned says no signal terminated process.");
				if(WTERMSIG(stat_val) != SIGABRT)
		   			error("Process terminated with %s.",
										sig_lookup(WTERMSIG(stat_val)));
			finis
		}
	finis

	assertion(10, BASE);
	start
		ENABLE_PSYNC
		for(i=0; i<3; i++) {
			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				if(i==0) {
					SIGNAL(SIGABRT, SIG_IGN);
				} else {
					if(i==1) {
						SIGBLOCK(set, SIGABRT);
					} else {
						SIGNAL(SIGABRT, sig_int);
					}
				}
				prep("Fork a child process.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {
					pid_t	i_pid;
	
					prep("Save current parent process ID of child process.");
					start
						i_pid = getppid();
						SEND_PSYNC
					finis
	
					testing("Exit of parent process does not terminate child");
					testing("process and child process is assigned a new");
					testing("parent process ID.");
					start
						SET_TIMEOUT
							while(i_pid == getppid())
								;
						CLEAR_ALARM
					finis
	
					done();
				}
	
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM

				abort();
				error("Process continued after abort() was called.");
				done();
			}
			WAIT_FOR_PID(pid);
			SLEEP(WAIT_FOR_TIME);	/* Allow time for test to complete. */
		}
		DISABLE_PSYNC
	finis
	
	assertion(11, BASE);
	start
		ENABLE_PSYNC
		for(i=0; i<3; i++) {
			FORK
				prep("Fork a grandchild.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {
					if(i==0) {
						SIGNAL(SIGABRT, SIG_IGN);
					} else {
						if(i==1) {
							SIGBLOCK(set, SIGABRT);
						} else {
							SIGNAL(SIGABRT, sig_int);
						}
					}

					prep("Have child delay, then call abort().");
					start
						sv_flush();
						SET_TIMEOUT
							WAIT_PSYNC
						CLEAR_ALARM
					finis

					abort();
					error("Process continued after abort() was called.");
					done();
				}
	
				prep("Have parent catch SIGCHLD signal");
				start
					sig_rcvd = 0;
					SIGNAL(SIGCHLD, sig_int);
					SEND_PSYNC
				finis

				sleep(WAIT_FOR_TIME);				/* SIGHUP expected. */
	
				testing("SIGCHLD signal received by parent process.");
				start
					if( ! sig_rcvd)
						error("SIGCHLD not generated when child exited.");
				finis
	
				WAIT_FOR_PID(pid);
			KROF
		}
		DISABLE_PSYNC
	finis
	
	done();
}
