#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>
#include	<locks.h>

#define	FILENAME	"8.1.18_18"
#define	FILENAME_	"8.1.18_18_"
#define FIFONAME 	"8.1.18_18_fifo"
char	String[] = "Hello, world!\n";
#define	Strlen	(sizeof(String) -1)


int	main(void)
{
	FILE	*sfd, *sfd_ref;
	int		fd;
	struct	stat st_buf;
	char	path[_POSIX_PATH_MAX +1];
	char	buffer[_POSIX_PIPE_BUF];
	char	inbuf[Strlen +1];
	struct	flock flock;
	pid_t	pid;

	setup(E_FREOPEN);
	funct(E_FREOPEN);

	assertion(18, BASE);
	start
		prep("Create a file to test with.");
		start
			exit_on(ERROR);
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			sfd_ref = Zfopen(path, "a");
			fd = Zfileno(sfd_ref);			/* Used in following assertion. */
		finis

		testing("Call to freopen() allocates a file descriptor.");
		start
			sfd = Zfreopen(path, "r+", sfd_ref);
			if(Zfileno(sfd) == -1)
				error("Call to freopen() did not allocate a file descriptor.");
			Zfclose(sfd);
		finis
	finis

	assertion(19, BASE);
	start
		testing("The freopen() call deallocated the file descriptor");
		testing("underlying the stream indicated by its argument.");
		start
			if(fd != Zopen(path, O_RDWR, PROT_ALL))
				error("File descriptor was not deallocated.");
			close(fd);
		finis
	finis

	assertion(20, BASE);
	start
		short	lock_typ;
		int		i;
		pid_t	pid;

		ENABLE_PSYNC
		for(lock_typ=F_RDLCK, i=0; i<2; lock_typ=F_WRLCK, i++) {
			prep("Initialize lock structure, open file, then fork() a child.");
			start
				INIT_LOCK_STRUCT(lock_typ, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,
												flock);
				fd = Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
				sfd = Zfdopen(fd, "r+");
				sv_flush ();
				pid = Zfork();
			finis
		
			if(pid == (pid_t)0) {
				prep("Have child set a lock, then delay a while.");
				start
					Zfcntl(fd, F_SETLK, &flock);
					sv_flush();
		
					SEND_PSYNC
					SET_TIMEOUT
						WAIT_PSYNCA
					CLEAR_ALARM
				finis
		
				prep("Call freopen() on stream pointer, then delay while");
				prep("parent checks file's locks.");
				start
					Zfreopen(path, "r+", sfd);
					sv_flush();
		
					SEND_PSYNC
					SET_TIMEOUT
						WAIT_PSYNCA
					CLEAR_ALARM
				finis

				done();
			}
		
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
			prep("Parent ensures lock is currently set on test file.");
			start
				flock.l_type = F_WRLCK;
				Zfcntl(fd, F_GETLK, &flock);
				if(flock.l_pid != pid)
					error("Child does not have lock as expected.");
				sv_flush();
				SEND_PSYNCA
			finis
		
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
			testing("Parent test lock is removed when call to freopen() is");
			testing("performed on the stream by the child process.");
			start
				Zfcntl(fd, F_GETLK, &flock);
				if(flock.l_type != F_UNLCK)
					error("A lock is still set on the closed file.");
				SEND_PSYNCA
			finis
		
			WAIT_FOR_PID(pid);

			cleanup("Close used stream and underlying file descriptor.");
			start
				Zfclose(sfd);
			finis
		}
		DISABLE_PSYNC
	finis
		
	assertion(21, EXTENDED);

	assertion(22, BASE);
	start
		FILE	*sfd_rd, *sfd_wr;

		break_on(ERROR);
		ENABLE_PSYNC

		prep("Establish FIFO and write data to it.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FIFONAME);
			Zmkfifo(path, PROT_ALL);

			SET_ALARM(1)
				fd = Zopen(path, O_RDONLY | O_NONBLOCK);
				sfd_rd = Zfdopen(fd, "r");
				sfd_wr = Zfopen(path, "w");

				Zfputs(String, sfd_wr);
			CLEAR_ALARM
		finis

		prep("Close FIFO output channel.");
		start
			SET_ALARM(1)
				fclose(sfd_wr);
			CLEAR_ALARM
		finis

		prep("Fork a child process.");
		start
			sv_flush ();
			pid = Zfork();
		finis
		
		if(pid == (pid_t)0) {
			prep("Close FIFO input channel via freopen().");
			start
				WAIT_PSYNC
				SET_ALARM(WAIT_FOR_TIME +1)
					sfd_rd = Zfreopen(path, "r", sfd_rd);
				CLEAR_ALARM
			finis

			testing("When all file descriptors associated with a FIFO");
			testing("have been closed by freopen(), data remaining in");
			testing("the pipe is discarded.");
			start
				SET_ALARM(1)
					fd = Zfileno(sfd_rd);
					Zfcntl(fd, F_SETFL, O_NONBLOCK);
					expecting(EAGAIN);
					expecting(-1);
					Zread(fd, inbuf, Strlen);
				CLEAR_ALARM
			finis

			done();
		}

		prep("Parent closes FIFO input channel.");
		start
			SET_ALARM(1)
				fclose(sfd_rd);
			CLEAR_ALARM
		finis

		prep("Delay to ensure close of input channel by child is completed,");
		prep("then open FIFO output channel to allow freopen() to complete.");
		start
			sv_flush();
			SEND_PSYNC
			SLEEP(1);
			sfd_wr = Zfopen(path, "w");
		finis

		WAIT_FOR_PID(pid);
		DISABLE_PSYNC

		cleanup("Remove FIFO file.");
		start
			fclose(sfd_rd);
			fclose(sfd_wr);
			Zunlink(path);
		finis
	finis

	assertion(23, EXTENDED);
	assertion(24, EXTENDED);
	assertion(25, EXTENDED);

	assertion(26, BASE);
	start
		time_t	s_time, cur_time;
		struct	stat st_buf;

		prep("Establish test file for testing time-related fields when an");
		prep("freopen() is performed.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zopen(path, O_RDWR | O_CREAT | O_TRUNC, PROT_ALL);
			expecting(Strlen);
			Zwrite(fd, String, Strlen);
			Zclose(fd);

			sfd = Zfopen(path, "r");
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval. */
			expecting(Strlen);
			Zfread(inbuf, sizeof(char), Strlen, sfd);
			Zfreopen(path, "r", sfd);
			Ztime(&cur_time);
			dt_secs(TRF_DELAY);
			Zstat(path, &st_buf);
		finis

		testing("When a file is closed, via freopen() by the last process");
		testing("that had it open, all time-related fields still marked");
		testing("for update are updated.");
		start
					/* Avoid calling ctime() twice in single function call. */
			char	s1[26];
											/* st_atime updated */
			if(st_buf.st_atime <= s_time) {
				strcpy(s1, ctime(&s_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_atime\" in error?.",
					"Expected minimum time is ",
					s1,
					"Actual time recorded is ",
					ctime(&(st_buf.st_atime)) +4);
			}
	
			if(st_buf.st_atime > cur_time) {
				strcpy(s1, ctime(&cur_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_atime\" larger than current time.",
					"Current time is ",
					s1,
					"Time-related field \"st_atime\" is ",
					ctime(&(st_buf.st_atime)) +4);
			}
		finis

		prep("Establish test file for testing time-related fields when an");
		prep("freopen() is performed.");
		start
				/* Close used stream pointer and underlying file descriptor. */
			Zfclose(sfd);

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval. */
			fd = Zopen(path, O_RDWR | O_CREAT | O_TRUNC, PROT_ALL);
			sfd = Zfdopen(fd, "r+");
			Zfputs(String, sfd);
			Zfreopen(path, "r+", sfd);
			Ztime(&cur_time);
			dt_secs(TRF_DELAY);
			Zstat(path, &st_buf);
		finis

		testing("Call to fclose() updated all time related fields marked");
		testing("for update.");
		start
					/* Avoid calling ctime() twice in single function call. */
			char	s1[26];
											/* st_ctime updated */
			if(st_buf.st_ctime <= s_time) {
				strcpy(s1, ctime(&s_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_ctime\" in error?.",
					"Expected minimum time is ",
					s1,
					"Actual time recorded is ",
					ctime(&(st_buf.st_ctime)) +4);
			}
	
			if(st_buf.st_ctime > cur_time) {
				strcpy(s1, ctime(&cur_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_ctime\" larger than current time.",
					"Current time is ",
					s1,
					"Time-related field \"st_ctime\" is ",
					ctime(&(st_buf.st_ctime)) +4);
			}
											/* st_mtime updated */
			if(st_buf.st_mtime <= s_time) {
				strcpy(s1, ctime(&s_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_mtime\" in error?.",
					"Expected minimum time is ",
					s1,
					"Actual time recorded is ",
					ctime(&(st_buf.st_mtime)) +4);
			}
	
			if(st_buf.st_mtime > cur_time) {
				strcpy(s1, ctime(&cur_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_mtime\" larger than current time.",
					"Current time is ",
					s1,
					"Time-related field \"st_mtime\" is ",
					ctime(&(st_buf.st_mtime)) +4);
			}
		finis

		cleanup("Close used stream and underlying file descriptor.");
		start
			Zfclose(sfd);
		finis
	finis

	assertion(27, BASE);
	start
		time_t	s_time;

		prep("Establish test file with buffered data.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			sfd = Zfopen(path, "w");
			setvbuf(sfd, buffer, _IOFBF, _POSIX_PIPE_BUF);
			expecting(Strlen);
			Zfwrite(String, sizeof(char), Strlen, sfd);
		finis

		testing("When the referenced stream is writable and buffered data");
		testing("has not yet been written, then a call to freopen() marks");
		testing("for update the \"st_ctime\" and \"st_mtime\" time-related");
		testing("fields of the underlying file.");
		start
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME_);
			Zfreopen(path, "w", sfd);

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, TRUE, s_time, NO_FILDES, path);
		finis
	finis

	assertion(28, BASE);
	start
		prep("Establish path to non-existing file.");
		start
			umask(0);
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			unlink(path);

		finis

		testing("When a call to freopen() creates a file, then the file");
		testing("permission bits are set appropriately.");
		start
			Zfreopen(path, "w", sfd);

			Zstat(path, &st_buf);
			if( ! (
				((st_buf.st_mode&(S_IRUSR | S_IWUSR)) == (S_IRUSR | S_IWUSR)) &&
				((st_buf.st_mode&(S_IRGRP | S_IWGRP)) == (S_IRGRP | S_IWGRP)) &&
				((st_buf.st_mode&(S_IROTH | S_IWOTH)) == (S_IROTH | S_IWOTH)) &&
				((st_buf.st_mode&(S_IXUSR | S_IXGRP | S_IXOTH)) == 0)
				  ))
				error("%s\n%s%lo (octal)\n%s%lo (octal)\n",
				  		"Unexpected mode.",
					  	"Expected:	",
					  	(unsigned long) (S_IRUSR | S_IWUSR | S_IRUSR | S_IWUSR
											| S_IROTH | S_IWOTH),
					  	"Actual:	",
					  	(unsigned long) (st_buf.st_mode & PROT_ALL));
		finis
	finis

	done();
}
