#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	FILENAME	"8.1.18_29"
#define	FILENAME_T	"8.1.18_29t"

int	main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	char	path_t[_POSIX_PATH_MAX +1];
	FILE	*sfd;

	setup(E_FREOPEN);
	funct(E_FREOPEN);

	assertion(29, BASE);
	start
		prep("Establish a test file for obtaining an open file.");
		start
			Zstrcpy(path_t, temp_path());
			Zstrcat(path_t, FILENAME_T);
			Zclose(Zcreat(path_t, PROT_ALL));
		finis

		prep("Initialize path name for a file where search permission");
		prep("is denied on a component of the path prefix.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_perm.d");
			Zstrcat(path, "/freopen_tryme");
			sfd = Zfopen(path_t, "r");
		finis

		testing("Expecting [EACCES] result.");
		start
			expecting(NULL);
			expecting(EACCES);
			Zfreopen(path, "r", sfd);
		finis
	finis

	assertion(30, BASE);
	start
		prep("Initialize a path name.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_perms.d");
			Zstrcat(path, "/no_read");
			sfd = Zfopen(path_t, "r");
		finis

		testing("Read permission along the path is denied.");
		start
			expecting(NULL);
			expecting(EACCES);
			Zfreopen(path, "r", sfd);
		finis

		prep("Initialize a path name.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_perms.d");
			Zstrcat(path, "/no_write");
		finis

		testing("Write permission along the path is denied.");
		start
			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EACCES);
			Zfreopen(path, "w", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EACCES);
			Zfreopen(path, "w+", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EACCES);
			Zfreopen(path, "wb", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EACCES);
			Zfreopen(path, "wb+", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EACCES);
			Zfreopen(path, "w+b", sfd);
		finis
	finis


	assertion(31, BASE);
	start
		prep("Initialize a path name.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_write.d");
			Zstrcat(path, "/new_file");
		finis

		testing("When file does not exist and write permission is denied");
		testing("for the parent directory of the file to be created, then a");
		testing("call to freopen() returns a NULL pointer and sets errno to");
		testing("[EACCES].");
		start
			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EACCES);
			Zfreopen(path, "w", sfd);
		finis

		testing("File is not created.");
		start
			expecting(FAILURE);
			Zaccess(path, F_OK);
		finis
	finis

	assertion(32, BASE);
	start
		pid_t	pid;
		time_t	s_time;
		struct	stat st_buf;

		prep("Create a FIFO file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "er_eintr");
			Zmkfifo(path, PROT_ALL);

			Zstat(path, &st_buf);			/*	Set base-line times for FIFO */
			Zstrcpy(path, temp_path());
			Zstat(path, &st_buf);			/*	and parent directory.		*/

			Zstrcpy(path, temp_path());
			Zstrcat(path, "er_eintr");

			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
		finis


		ENABLE_PSYNC
		prep("Fork a child.");
		start
			sv_flush( );
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {		/* Child */
			exit_on(ERROR);
			SIGNAL(SIGINT, sig_int);

			testing("On freopen() interrupted, errno set to EINTR.");
			start
				sfd = Zfopen(path_t, "r");
				SEND_PSYNC
				expecting(NULL);
				expecting(EINTR);
				Zfreopen(path, "w", sfd);
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		prep("Send a signal to the child.");
		start
			SLEEP(WAIT_FOR_TIME);
			Zkill(pid, SIGINT);
		finis

		WAIT_FOR_PID(pid);
		DISABLE_PSYNC

		testing("Time related fields \"st_ctime\", \"st_mtime\", and");
		testing("\"st_atime\" of the file and \"st_ctime\", \"st_mtime\" of");
		testing("of the parent directory are not updated.");
		start
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);

			Zstrcpy(path, temp_path());
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
		finis
	finis

	assertion(33, BASE);
	start
		time_t	s_time;
		struct	stat st_buf;

		prep("Initialize a path name.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "open_dir");
			Zmkdir(path, PROT_ALL);
			Zstat(path, &st_buf);   /*  Set base-line time for created file. */
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
		finis

		testing("Named file is a directory and type is set for writing.");
		start
			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "w", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "a", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "r+", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "w+", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "a+", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "wb", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "ab", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "r+b", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "rb+", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "w+b", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "wb+", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "a+b", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(EISDIR);
			Zfreopen(path, "ab+", sfd);
		finis

		testing("Time related fields \"st_ctime\", \"st_mtime\", and");
		testing("\"st_atime\" of the file are not updated.");
		start
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);
		finis
	finis

	{
		int		open_max, topen_max, i;

		break_on(ERROR);

		Zstrcpy(path, temp_path());
		Zstrcat(path, "fill_up");
		open_max = g_open_max();	/* Get value for {OPEN_MAX} */
									/* Get testable value for {OPEN_MAX} */
		topen_max = (open_max <= PCTS_OPEN_MAX) ? open_max : PCTS_OPEN_MAX;

		if(open_max <= PCTS_OPEN_MAX)
			assertion(34, EXTENDED);
		else {
			assertion(34, BASE);
			start
				prep("Close all file descriptors except 0, 1 and 2.");
				start
					for(i=3; i<topen_max; i++)
						(void) close(i);
				finis

				prep("Testing for {PCTS_OPEN_MAX} < {OPEN_MAX}.");
				prep("Consume {PCTS_OPEN_MAX} file descriptors."); 
				start
					for(i=3; i<(topen_max -1); i++)
						Zcreat(path, PROT_ALL);
					sfd = Zfopen(path_t, "r");
				finis

				prep("Initialize a path name.");
				start
					Zstrcpy(path, temp_path());
					Zstrcat(path, "emfile_tst");
					Zclose(Zcreat(path, PROT_ALL));
				finis
	
				testing("Call freopen() on {PCTS_OPEN_MAX} file descriptor.");
				start
					Zfreopen(path, "r+", sfd);
				finis

				cleanup("Close all file descriptors except 0, 1 and 2.");
				start
					for(i=3; i<topen_max; i++)
						close(i);
				finis
			finis
		}
	}

	if(tc_name_max()) {
		assertion(35, BASE);
		start
			e_NO_TRUNC(E_FREOPEN);
		finis
	}
	else
		assertion(35, NOT_TESTABLE);

	if(!tc_name_max()) {
		assertion(36, BASE);
		start
			e_NO_TRUNC(E_FREOPEN);
		finis
	}
	else
		assertion(36, NOT_TESTABLE);

	if(tc_path_max()) {
		assertion(37, BASE);
		start
			e_ENAMETOOLONG(E_FREOPEN);
		finis
	}
	else
		assertion(37, NOT_TESTABLE);

	assertion(38, EXTENDED);

	assertion(39, BASE);
	start
		prep("Setup a pathname for a file which does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_freopen_");
		finis

		testing("Named file does not exist.");
		start
			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(ENOENT);
			Zfreopen(path, "r", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(ENOENT);
			Zfreopen(path, "r+", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(ENOENT);
			Zfreopen(path, "rb", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(ENOENT);
			Zfreopen(path, "r+b", sfd);

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(ENOENT);
			Zfreopen(path, "rb+", sfd);
		finis
	finis

	assertion(40, BASE);
	start
		testing("Call to freopen() when a path prefix does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_freopen_/test");

			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(ENOENT);
			Zfreopen(path, "w", sfd);
		finis

		testing("Call to freopen() when path points to an empty string.");
		start
			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(ENOENT);
			Zfreopen("", "w", sfd);
		finis
	finis

	assertion(41, EXTENDED);

	assertion(42, BASE);
	start
		testing("Call to freopen() when path contains a non-directory");
		testing("component.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));
			Zstrcat(path, "/test");


			sfd = Zfopen(path_t, "r");
			expecting(NULL);
			expecting(ENOTDIR);
			Zfreopen(path, "r", sfd);
		finis
	finis

	done();

}
