#include	<pcts.h>

#include	<time.h>

#include	<sv_env.h>

extern	char **environ;
extern	char *tzname[2];

static char	*good_tz[] = {
							"ABC", "DEF", "GHI", "JLK", "MNO",
							"PQR", "STU", "VWX", "YZz",
							"abc", "def", "ghi", "jlk", "mno",
							"pqr", "stu", "vwx", "yzZ",
							"",
						   };

static char	*bad_tz[] = {
							":AB", ",AB", "-AB", "+AB",
							"0AB", "1AB", "2AB", "3AB", "4AB",
							"5AB", "6AB", "7AB", "8AB", "9AB",
							"",
						   };

static	char	*tztest[] = { "A", "B", };

#define	EST	"AAA"
#define	DST	"BBB"

int	main(void)
{
	char	tz_env[100];				/* Contains TZ environment */
	char	*tzname_save[2];
	char	*env[2];

	setup(E_TZSET);
	funct(E_TZSET);

	assertion(3, BASE);
	start
		testing("Assign values to tzname elements.");
		start
			tzname_save[0] = tzname[0];
			tzname_save[1] = tzname[1];
			tzname[0] = tztest[0];
			tzname[1] = tztest[1];
		finis

		testing("Compare tzname to variable of type \"char *x[].\"");
		start
			if(tzname[0] == tztest[0])
				;
			if(tzname[1] == tztest[1])
				;
		finis

		testing("Size of \"tzname[]\"");
		start
			if(sizeof(tzname) != sizeof(tztest))
				error("%s\n%s%d\n%s%d\n",
					"Length of \"tzname\" is incorrect.",
					"Expecting : ",
					sizeof(tztest),
					"Actual : ",
					sizeof(tzname));
		finis

		cleanup("Restore tzname parameters established by IUT.");
		start
			tzname[0] = tzname_save[0];
			tzname[1] = tzname_save[1];
		finis
	finis

	assertion(4, BASE);
	start
		testing("The tzset() function sets tzname[0] to point to the");
		testing("standard timezone designation and tzname[1] to point to");
		testing("the summer timezone designation.");
		start
			env[0] = tz_env;			/* Establish TZ environment.	*/
			env[1] = NULL;				/* Terminate TZ environment.	*/
			environ = env;
			sprintf(tz_env, "TZ=%s0%s", EST, DST);
			tzset();
			if(strcmp(tzname[0], EST))
				error("%s%s\n%s%s",
						"Expected: ",
						EST,
						"Received: ",
						tzname[0]);
			if(strcmp(tzname[1], DST))
				error("%s%s\n%s%s",
						"Expected: ",
						DST,
						"Received: ",
						tzname[0]);
		finis
	finis


	assertion(5, BASE);
	start
		int		i;

		environ[0] = tz_env;	/* Establish TZ environment.		*/
		for(i=0; *good_tz[i+1]; i++) {
			sprintf(tz_env, "TZ=%.*s0%.*s", strlen(good_tz[i]), good_tz[i],
										strlen(good_tz[i +1]), good_tz[i +1]);

			testing("TZ is acceptable with form: TZ=%s", getenv("TZ"));
			start
				tzset();
				if(strncmp(tzname[0], good_tz[i], strlen(good_tz[i])))
					error("tzname[0]: %s\nStandard designation: %.*s",
									tzname[0], strlen(good_tz[i]), good_tz[i]);
				if(strncmp(tzname[1], good_tz[i+1], strlen(good_tz[i+1])))
					error("tzname[1]: %s\nSummer designation: %.*s",
								tzname[1], strlen(good_tz[i+1]), good_tz[i+1]);
			finis
		}

		for(i=0; *bad_tz[i+1]; i++) {
			sprintf(tz_env, "TZ=%.*s0%.*s", strlen(bad_tz[i]), bad_tz[i],
										strlen(bad_tz[i +1]), bad_tz[i +1]);

			testing("TZ is unacceptable with form: TZ=%s", getenv("TZ"));
			if(*bad_tz[i] == ':') {
				testing("NOTE: Failure to meet the expected result when the");
				testing("first character is a ':' can be resolved by the");
				testing("APTL resolved test code of PCD_DOCUMENTED.");
				testing("See POSIX.1, 8.1.1, lines 60-61 and its required");
				testing("documentation assertion in POSIX.3.1, 8.1.56.1, D01.");
			}
			start
				tzset();
				if(!strncmp(tzname[0], bad_tz[i], strlen(bad_tz[i])))
					error("tzname[0]: %s\nStandard designation: %.*s",
									tzname[0], strlen(bad_tz[i]), bad_tz[i]);
				if(!strncmp(tzname[1], bad_tz[i+1], strlen(bad_tz[i+1])))
					error("tzname[1]: %s\nSummer designation: %.*s",
								tzname[1], strlen(bad_tz[i+1]), bad_tz[i+1]);
			finis
		}
	finis

	done();
}
