#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>

int main(int argc, char *argv[])
{
	int		fd1, fd2, fd3;
	off_t	offset;
	struct	stat st_buf;
	int		e_index;
	ino_t	ino;
	dev_t	dev;
	int		ret_val;

	e_index = atoi(*argv);
	sv_config();					/* Required for next line.				*/
	sv_nproc();						/* Attain proper journal ID sequences.	*/
	ass_nu = 36;

	prep("Pick up passed arguments.");
	start
		fd1 = atoi(*++argv);
		fd2 = atoi(*++argv);
		fd3 = atoi(*++argv);
		offset = atoi(*++argv);
		ino = atoi(*++argv);
		dev = atoi(*++argv);
	finis

	testing("File descriptors open in the calling process remain open");
	testing("in the new process.");
	start
		Zfstat(fd1, &st_buf);
		Zfstat(fd2, &st_buf);
	finis

	testing("File descriptors open in the calling process are closed");
	testing("in the new process when the FD_CLOEXEC flag is set.");
	start
		ret_val = fstat(fd3, &st_buf);
		if(ret_val == 0) {
			if((st_buf.st_ino == ino) && (st_buf.st_dev == dev))
				error("File not closed after an \"exec\".");
		} else {
			expecting(EBADF);
			Zfstat(fd3, &st_buf);
		}
	finis

	assertion(37, BASE);
	start
		int	smode;

		testing("File offset is unchanged from calling process's open file.");
		testing("descriptor.");
		start
			if(Zlseek(fd1, (off_t)0, SEEK_CUR) != offset)
				error("%s%d\n%s%d\n",
						"Expecting an offset of: ",
						offset,
						"Actual offset is: ",
						lseek(fd1, (off_t)0, SEEK_CUR) );
		finis

		testing("File status flags are unchanged from calling process's.");
		testing("open file descriptor.");
		start
			smode = Zfcntl(fd1, F_GETFL, 0);
			if( ((smode & O_ACCMODE) & O_RDWR) != O_RDWR)
				error("%s\n%s%o\n",
						"Expecting open file descriptor status of O_RDWR.",
						"Actual status (in octal) is: ",
						smode & O_ACCMODE);

			smode = Zfcntl(fd2, F_GETFL, 0);
			if( ((smode & O_ACCMODE) & O_WRONLY) != O_WRONLY)
				error("%s\n%s%o\n",
						"Expecting open file descriptor status of O_WRONLY.",
						"Actual status (in octal) is: ",
						smode & O_ACCMODE);
		finis

		testing("File mode is unchanged from calling process's.");
		testing("open file descriptor.");
		start
			smode = Zfcntl(fd1, F_GETFL, 0);
			if((smode & O_APPEND) != O_APPEND)
				error("%s\n%s\n",
					"Expecting open file descriptor mode of O_APPEND.",
					"Actual mode of O_APPEND is not set.");
			if((smode & O_NONBLOCK) == O_NONBLOCK)
				error("%s\n%s\n",
					"Expecting open file descriptor mode O_NONBLOCK not set.",
					"Actual mode of O_NONBLOCK is set.");

		finis

		testing("Lock's are unchanged from calling process's.");
		testing("open file descriptor.");
		start
			struct flock	lock_struct;
			pid_t	pid;

			if((pid = Zfork()) == 0) {
				lock_struct.l_type = F_WRLCK;
				lock_struct.l_whence = SEEK_SET;
				lock_struct.l_start = 0;
				lock_struct.l_len = 0;
				Zfcntl(fd2, F_GETLK, &lock_struct);
				if(lock_struct.l_type == F_UNLCK)
					error("%s\n%s\n",
						"Expected file to be locked.",
						"Actually file is unlocked.");
	
				Zfcntl(fd1, F_GETLK, &lock_struct);
				if(lock_struct.l_type != F_UNLCK)
					error("%s\n%s\n",
						"Expected file to be unlocked.",
						"Actually file is locked.");
				done();
			}
			WAIT_FOR_PID(pid);
		finis
	finis

	done();
}
