#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_sigs.h>
#include	<pcts_timer.h>


int	main(void)
{
	pid_t	pid;

	setup(E_FORK);
	funct(E_FORK);

	assertion(12, BASE);
	start
		prep("Set an alarm.");
		start
		finis

		alarm(2);
			FORK
				testing("Pending alarms are cleared for the child process.");
				start
					unsigned int	tl_alm;

					if((tl_alm = alarm(0)) != (unsigned int) 0)
						error("%s\n%s%u",
							"Expecting return from an alarm() call to be zero.",
							"Actual time left was:	",
							tl_alm);
				finis
			KROF
		alarm(0);
	finis

	assertion(13, BASE);
	start
		sigset_t	set;
		int			i;

		FORK
			prep("Establish pending signals for testing.");
			start
				for(i=0; i<NSIGS; i++) {
					if((sigs[i].sig == SIGKILL) || (sigs[i].sig == SIGSTOP))
						continue;
	
					SIGNAL(sigs[i].sig, sig_int);
					Zsigemptyset(&set);
					SIGBLOCK(set, sigs[i].sig);
					Zkill(getpid(), sigs[i].sig);
	
										/* Ensure signal is pending. */
					Zsigpending(&set);
					if( ! sigismember(&set, sigs[i].sig) )
						error("%s\n%s\"%s\"\n%s\n",
							"Pending signal not reported.",
							"Expected pending signal:	",
							sigs[i].signame,
							"Actual signal number was not pending.");
				}
			finis
	
			FORK
				testing("Signals are not pending in child process.");
				start
					for(i=0; i<NSIGS; i++) {
						if((sigs[i].sig == SIGKILL) || (sigs[i].sig == SIGSTOP))
							continue;
	
						Zsigpending(&set);
						if(sigismember(&set, sigs[i].sig) )
							error("%s\n%s\"%s\"\n%s\n",
								"Expected no pending signals.",
								"Actual signal ",
								sigs[i].signame,
								" is pending.");
					}
				finis
			KROF
		KROF
	finis
	
	assertion(14, BASE);
	start
		char	path[_POSIX_PATH_MAX +1], path2[_POSIX_PATH_MAX +1];

		prep("Get parent's current working directory and root directory.");
		start
			Zgetcwd(path, _POSIX_PATH_MAX);
		finis

		FORK
			prep("Get child's current working directory and root directory.");
			start
				Zgetcwd(path2, _POSIX_PATH_MAX);
			finis

			testing("Current working directory and root directory are same.");
			start
				if(strcmp(path, path2)) 
					error("%s\"%s\"\n%s\"%s\"",
						"Current working directory is ",
						path,
					 	"Root directory is ",
						path2);
			finis
		KROF
	finis

	assertion(15, BASE);
	start
		int			i;

		FORK
			for(i=0; i<NSIGS; i++) {
				if((sigs[i].sig == SIGKILL) || (sigs[i].sig == SIGSTOP))
					continue;
				if(sigs[i].sig == SIGCHLD)
					break;
	
				prep("Setup to catch signal %s.", sigs[i].signame);
				start
					SIGNAL(sigs[i].sig, sig_int);
				finis
	
				FORK
					prep("Send signal %s to child process.", sigs[i].signame);
					start
						sig_rcvd = 0;
						Zkill(getpid(), sigs[i].sig);
					finis
	
					testing("Child caught the signal.");
					start
						if(sig_rcvd == 0)
							error("Signal %s not caught.\n", sigs[i].signame);
					finis
				KROF
			}
		KROF
	finis

	assertion(16, BASE);
	start
		prep("Get real and effective user IDs and group IDs for parent.");
		start
			expecting(UID);
			Zgetuid();

			expecting(UID);
			Zgeteuid();

			expecting(GID);
			Zgetgid();

			expecting(GID);
			Zgetegid();
		finis

		FORK
			testing("Real and effective user IDs and group IDs for child");
			testing("are the same as the parent.");
			start
				expecting(UID);
				Zgetuid();

				expecting(UID);
				Zgeteuid();

				expecting(GID);
				Zgetgid();

				expecting(GID);
				Zgetegid();
			finis
		KROF
	finis

	done();
}
