#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>


int	main(void)
{
    pid_t	pid;
	int		stat_val;

    setup(E_PAUSE);
    funct(E_PAUSE);

			/*	The pause() function suspends the calling process until
				delivery of a signal whose action is either to execute
				a signal catching function or to terminate the process.
			*/
	assertion(4, BASE);
	start
		break_on(ERROR);

		FORK
			prep("Call pause(), and wait for SIGALRM signal to be caught.");
			start
				SET_ALARM(WAIT_FOR_TIME)
					pause();
				EXPECT_ALARM
			finis

			testing("Call to pause() suspended the calling process until");
			testing("delivery of a signal which was being caught.");
			start
				if(unused_time != 0)
					error("The pause() call returned too early.");
			finis
		KROF

		ENABLE_PSYNC
    	prep("Fork a child process.");
    	start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Set signal action for SIGINT to SIG_DFL, call pause().");
			start
				sv_flush();
				SIGNAL(SIGINT, SIG_DFL);
				SEND_PSYNC
				Zpause();
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		testing("Child still active after calling pause().");
		start
			SLEEP(WAIT_FOR_TIME);
			if(Zwaitpid(pid, &stat_val, WNOHANG | WUNTRACED) != (pid_t)0)
					error("Child has exited or has stopped.");
		finis

		prep("Send SIGINT to child.");
		start
			Zkill(pid, SIGINT);
		finis

		testing("Call to pause() suspended calling process until delivery");
		testing("of a signal whose action was to terminate the process.");
		start
			int	stat_val;

			SET_TIMEOUT
				waitpid(pid, &stat_val, 0);

				if(WIFSIGNALED(stat_val)) {
					if(WTERMSIG(stat_val) != SIGINT)
						error("%s\n%s%s",
							"Expected child process terminated by SIGINT.",
							"Actually child terminated by ",
							sig_lookup(WTERMSIG(stat_val)) );
				}
				else
					error("WIFSIGNALED reports no signal terminated child.");
			CLEAR_ALARM
		finis
		DISABLE_PSYNC
	finis


		/*	Call to pause() when a signal is received and the action
			is to terminate the process, pause() does not return.
		*/
	assertion(5, BASE);
	start
		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			testing("Call to pause() interrupted by a signal and action");
			testing("is to terminate the process does not return.");
			start
				SIGNAL(SIGALRM, SIG_DFL);
				alarm(1);
				pause();
				error("Call to pause() returned.");
			finis

			done();
		}

		SET_TIMEOUT
			waitpid(pid, &stat_val, 0);

			if(WIFSIGNALED(stat_val)) {
				if(WTERMSIG(stat_val) != SIGALRM)
					error("%s\n%s%s",
							"Expected child process terminated by SIGALARM.",
							"Actually child terminated by ",
							sig_lookup(WTERMSIG(stat_val)) );
			}
			else
				error("WIFSIGNALED reports no signal terminated child.");
		CLEAR_ALARM
	finis


	assertion(6, BASE);
	start
		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			testing("Call to pause() when interrupted by a caught signal");
			testing("returns after the signal-catching function returns.");
			start
				SIGNAL(SIGALRM, sig_int);
				sig_rcvd = 0;
				alarm(1);
										/* This is the test for Assertion #7. */
				expecting(-1);
				expecting(EINTR);
				Zpause();
				if(sig_rcvd == 0)
					error("Signal catching function was not executed.");
			finis

			done();
		}

		SET_TIMEOUT
			waitpid(pid, &stat_val, 0);

			if(WIFSIGNALED(stat_val)) {
				error("%s\n%s%s",
						"Expected child to exit normally. ",
						"Actually child terminated by ",
						sig_lookup(WTERMSIG(stat_val)) );
			}
		CLEAR_ALARM
	finis

		/*	Call to pause() when a signal is caught by the calling process
			and control is returned from the signal catching function shall
			return -1 and set errno to [EINTR].
		*/
	assertion(7, BASE);
	start
		testing("Tested as part of Assertion #6.");
		start
		finis
	finis

    done();
}
