#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_sigs.h>
#include	<pcts_timer.h>


int main(void)
{
	pid_t 		pid;
	int			stat_val;
	sigset_t	set;
	int			i;

    setup(E_SIGNAL_CON);
    funct(E_SIGNAL_CON);

	assertion(27, BASE);
	start
		FORK
			prep("Set process signal mask to block SIGTTSP, SIGTTIN, and");
			prep("SIGTTOU signals, then send SIGTTSP, SIGTTIN, and");
			prep("SIGTTOU signals to this process.");
			start
				SIGNAL(SIGCONT, SIG_DFL);
				for(i=NSIGS_STOP; i<NSIGS; i++) {
					if(sigs[i].sig == SIGSTOP)
						continue;
					SIGNAL(sigs[i].sig, SIG_DFL);
					SIGBLOCK(set, sigs[i].sig);
					Zkill(getpid(), sigs[i].sig);
					SLEEP(WAIT_FOR_TIME);
				}
			finis

			prep("Check stop signals are pending.");
			start
				Zsigpending(&set);
				for(i=NSIGS_STOP; i<NSIGS; i++) {
					if(sigs[i].sig == SIGSTOP)
						continue;
					if( ! sigismember(&set, sigs[i].sig) )
						error("Signal %s not pending.",sigs[i].signame);
				}
			finis

			prep("Deliver SIGCONT to this process.");
			start
				Zkill(getpid(), SIGCONT);
				SLEEP(WAIT_FOR_TIME);
			finis

			testing("Stop signals are no longer pending.");
			start
				Zsigpending(&set);
				for(i=NSIGS_STOP; i<NSIGS; i++) {
					if(sigs[i].sig == SIGSTOP)
						continue;
					if(sigismember(&set, sigs[i].sig) )
						error("Signal %s pending.",sigs[i].signame);
				}
			finis
		KROF
	finis

	assertion(28, BASE);
	start
							/* Check for SIGCONT blocked. */
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Set process signal mask to block SIGCONT.");
			start
				SIGNAL(SIGCONT, SIG_DFL);
				SIGBLOCK(set, SIGCONT);
			finis
		
			prep("Have child stop.");
			start
				Zkill(getpid(), SIGSTOP);
				sleep(WAIT_FOR_TIME);
			finis

			done();
		}

		prep("Wait for child process to be stopped.");
		start
			SET_ALARM(TIMEOUT)
				Zwaitpid(pid, &stat_val, WUNTRACED);
				if( ! WIFSTOPPED(stat_val) )
					error("%s\n%s",
						"Expected child process to be stopped.",
						"Actually child process was terminated.");
			CLEAR_ALARM
		finis

		prep("Deliver a SIGCONT to the child process.");
		start
			Zkill(pid, SIGCONT);
		finis

		testing("Child Process has continued.");
		start
			SET_ALARM(TIMEOUT)
				Zwaitpid(pid, &stat_val, 0);
			CLEAR_ALARM
			if(sigalrm_rcvd != 0) {			/* waitpid() timed out.		*/
				Zkill(pid, SIGKILL); 		/* Terminate child process.	*/
				error("Child process is still stopped.");
			}
		finis

						/* Check for SIGCONT ignored. */
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Set signal action for SIGCONT to SIG_IGN");
			start
				SIGNAL(SIGCONT, SIG_IGN);
			finis
		
			prep("Have child STOP.");
			start
				Zkill(getpid(), SIGSTOP);
				sleep(WAIT_FOR_TIME);
			finis

			done();
		}

		prep("Wait for child process to be stopped.");
		start
			SET_ALARM(TIMEOUT)
				Zwaitpid(pid, &stat_val, WUNTRACED);
				while( ! WIFSTOPPED(stat_val) )
					Zwaitpid(pid, &stat_val, WUNTRACED | WNOHANG);
			CLEAR_ALARM
		finis

		prep("Deliver a SIGCONT to the child process.");
		start
			Zkill(pid, SIGCONT);
		finis

		testing("Child Process has continued.");
		start
			SET_ALARM(TIMEOUT)
				Zwaitpid(pid, &stat_val, 0);
			CLEAR_ALARM
			if(sigalrm_rcvd != 0) {			/* waitpid() timed out.		*/
				Zkill(pid, SIGKILL); 		/* Terminate child process.	*/
				SLEEP(WAIT_FOR_TIME);
				waitpid(pid, &stat_val, WNOHANG);
				error("Child process is still stopped.");
			}
		finis
	finis


				/*	When SIGCONT is generated for a process that
					is stopped, and the SIGCONT signal is both blocked
					and not ignored, then it remains pending until it is
					either unblocked or a stop signal is generated for
					the process.
				*/
	assertion(29, BASE);
	start
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
								/* Test Stop signal discards pending
									SIGCONT signal.
								*/
			prep("Block and set action to catch signal SIGCONT.");
			start
				SIGBLOCK(set, SIGCONT);
				sig_rcvd = 0;
				SIGNAL(SIGCONT, sig_int);
			finis

			prep("Setup a pending SIGCONT signal.");
			start
				Zkill(getpid(), SIGCONT);
				SLEEP(WAIT_FOR_TIME);
				Zsigpending(&set);
				if( ! sigismember(&set, SIGCONT) )
					error("%s\n%s",
						"Expected signal SIGCONT pending.",
						"Actually signal SIGCONT is not pending.");
			finis
		
			prep("Deliver a stop signal to the child process.");
			start
				Zkill(getpid(), SIGSTOP);
				SLEEP(WAIT_FOR_TIME);
			finis

							/* Stop signal discards pending SIGCONT
							   signal, but parent send another SIGCONT
							   which becomes pending.

							   Discard of pending SIGCONT signals not
							   testable unless implementation queues up
							   multiple pending SIGCONT signals.
							*/
			testing("Process continues with SIGCONT pending.");
			start
				Zsigpending(&set);
				if( ! sigismember(&set, SIGCONT) )
					error("%s\n%s",
						"Expected signal SIGCONT pending.",
						"Actually signal SIGCONT is not pending.");
			finis

			prep("Unblock SIGCONT.");
			start
				Zsigemptyset(&set);
				Zsigaddset(&set, SIGCONT);
				Zsigprocmask(SIG_UNBLOCK, &set, (sigset_t *)0);
			finis
		
								/* Test unblocking SIGCONT removes
									pending SIGCONT signal.
								*/
			testing("SIGCONT signal no longer pending.");
			start
				Zsigpending(&set);
				if(sigismember(&set, SIGCONT) )
					error("%s\n%s",
						"Expected signal SIGCONT not pending.",
						"Actually signal SIGCONT is pending.");
			finis

			testing("Only one SIGCONT signal delivered.");
			start
				if(sig_rcvd != 1)
					error("%s\n%s%d",
						"Expected a single SIGCONT signal to be delivered.",
						"Actually delivery of SIGCONT signals was ",
						sig_rcvd);
			finis

			done();
		}

		prep("Wait for child process to be stopped.");
		start
			SET_ALARM(TIMEOUT)
				Zwaitpid(pid, &stat_val, WUNTRACED);
				while( ! WIFSTOPPED(stat_val) )
					Zwaitpid(pid, &stat_val, WUNTRACED | WNOHANG);
			CLEAR_ALARM
		finis

		prep("Send SIGCONT to continue child process.");
		prep("The SIGCONT signal also becomes pending.");
		start
			sv_flush();
			Zkill(pid, SIGCONT);
		finis

		WAIT_FOR_PID(pid);

		cleanup("Ensure child process exits.");
		start
			kill(pid, SIGKILL);
		finis
	finis

	done();
}
