#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_sigs.h>
#include	<termios.h>
#include	<pcts_timer.h>

	/* Testing:  tcdrain() not affected by ignored signal. */


int main(void)
{
	pid_t	pid;
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	int		i;
	char	*outbuf;
	size_t	max_input;
	size_t	nb_xmit;

	setup(E_SIGNAL_CON);
	funct(E_SIGNAL_CON);

	if(GTI_NOT_PROVIDED) {
		assertion(59, NOT_TESTABLE);
		done();
	}

	assertion(59, BASE);
	start
		max_input = g_max_input(PORT_TTY);
		if(max_input > (unsigned)(PCTS_MAX_INPUT)) /* Limit buffer size */
			max_input = (unsigned)(PCTS_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
		nb_xmit = (baud_value(baud_min())/10)*TIMEOUT/2;
		if(nb_xmit > max_input)
			nb_xmit = max_input;

		prep("Open terminal file and set initial state.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);
	
			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VMIN] = 0;
			tty_termios_p->c_cc[VTIME] = 0;
			Zcfsetospeed(loop_termios_p, baud_min());
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			Zcfsetispeed(tty_termios_p, baud_min());
		
			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);

			outbuf = (char *) Zmalloc(nb_xmit +1);
		finis
		
		ENABLE_PSYNC
		for(i=0; i<NSIGS_JC; i++) {
			if(sigs[i].sig == SIGKILL)
				continue;
			prep("Fork a child process to test tcdrain().");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Set up output buffer.");
				start
					memset(outbuf, (int)'A', sizeof(outbuf));
					Ztcflush(tty_fildes, TCIOFLUSH);
						/* Ensure parent's prep() message is processed
						   before child's testing() message.  Allow for
						   parent to process prep() call on non-timeshared
						   systems.
						*/
					SLEEP(WAIT_FOR_TIME);
				finis
	
				prep("Write data to controlling terminal.");
				start
					write(tty_fildes, outbuf, nb_xmit);

								/* Stop accepting data, allowing signal
								   to be generated while tcdrain() is
								   active.
								*/
					Zwrite(loop_fildes, (char *)&tty_termios_p->c_cc[VSTOP], 
														sizeof(char));
				finis
	
				prep("Set signal action to SIG_IGN for %s.", sigs[i].signame);
				start
					SIGNAL(sigs[i].sig, SIG_IGN);
				finis
	
				testing("tcdrain() not affected by ignored signal.");
				start
					SEND_PSYNC
					Ztcdrain(tty_fildes);
				finis
	
				cleanup("Flush tty queues.");
				start
					Ztcflush(loop_fildes, TCIOFLUSH);
					Ztcflush(tty_fildes, TCIOFLUSH);
				finis

				done();
			}
	
			prep("Wait for child to write out data.");
			prep("Then issue signal %s to child process.", sigs[i].signame);
			start
				sv_flush();
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM
	
				Zkill(pid, sigs[i].sig);
								/* Signal has been generated while tcdrain()
								   was active, now accept data.
								*/
				Zwrite(loop_fildes, (char *)&tty_termios_p->c_cc[VSTART], 
														sizeof(char));
			finis
	
			WAIT_FOR_PID(pid);
		}
		DISABLE_PSYNC
	finis
	
	cleanup("Close terminal files.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		close(tty_fildes);
		close(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
