#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_sigs.h>
#include	<pcts_timer.h>

#define FIFONAME	"3.3.1_60"
#define	String		"Just a little"
#define Strlen		(sizeof(String)-1)

	/*	Testing functions: not affected by blocked signal. */


int main(void)
{
	pid_t		pid, ret_val;
	int			fd, stat_val, i;
	char		inbuf[50];
	struct		flock flock1, flock2;
	sigset_t	mask;
	char    	fifop[_POSIX_PATH_MAX +1];
	int 		fiford, fifowr;
	char    	*in_buf;
	char    	*outbuf;
	size_t		pipe_buf_sz, fifo_buf_sz, max_buf_sz, pipe_max, fifo_max;

	setup(E_SIGNAL_CON);
	funct(E_SIGNAL_CON);

	assertion(60, BASE);
	start
		prep("Establish path and create a FIFO.");
		start
			Zstrcpy(fifop, temp_path());
			Zstrcat(fifop, FIFONAME);
			Zmkfifo(fifop, PROT_ALL);
		finis

							/* Maximum bytes written in pipe and FIFO */
		max_buf_sz=g_max_pipe(&pipe_buf_sz, &fifo_buf_sz, &pipe_max, &fifo_max);

		prep("Allocate and initialize I/O buffers.");
		start
			in_buf = (char *) Zmalloc(max_buf_sz);
			outbuf = (char *) Zmalloc(max_buf_sz);
			memset(outbuf, 'Z', max_buf_sz);
		finis

		for(i=0; i<NSIGS_JC; i++) {
			if(sigs[i].sig == SIGKILL)
				continue;
			prep("Create a temp file, set a shared lock on the entire file.");
			start
				flock1.l_type = F_RDLCK;
				flock1.l_whence = SEEK_SET;
				flock1.l_start = ((off_t)0);
				flock1.l_len = ((off_t)20);

				fd = temp_open("sig_con", O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
				Zwrite(fd, inbuf, 20);
				Zfcntl(fd, F_SETLK, &flock1);
			finis

			 prep("Fork a child process to test fcntl().");
			 start
				sv_flush();
				pid = Zfork();
			finis
			
			if(pid == (pid_t)0) {
				prep("Block signal action for signal %s.", sigs[i].signame);
				start
					Zsigemptyset(&mask);
					SIGBLOCK(mask, sigs[i].sig);
				finis
	
				testing("Child calls fcntl() with F_SETLKW to wait to place");
				testing("a write lock on the file.");
				testing("Signal %s has no effect on the behavior of fcntl().",
										sigs[i].signame);
				start
					flock2.l_type = F_WRLCK;
					flock2.l_whence = SEEK_SET;
					flock2.l_start = ((off_t)0);
					flock2.l_len = ((off_t)5);
		
					Zfcntl(fd, F_SETLKW, &flock2);
				finis
		
				done();
			}
		
			prep("Issue a %s.",sigs[i].signame);
			start
				SLEEP(WAIT_FOR_TIME);	/* Allow child to call fcntl(). */
				Zkill(pid, sigs[i].sig);
				SLEEP(WAIT_FOR_TIME);
			finis

			cleanup("Clear write lock so child can complete successfully.");
			start
				sv_flush();
				Zclose(fd);
			finis

			WAIT_FOR_PID(pid);

										/* Testing:  open() */
			 prep("Fork a child process to test open().");
			 start
				sv_flush();
				pid = Zfork();
			finis
			
			if(pid == (pid_t)0) {
				prep("Block signal action for signal %s.", sigs[i].signame);
				start
					Zsigemptyset(&mask);
					SIGBLOCK(mask, sigs[i].sig);
				finis
	
				testing("Child calls open() and hangs.");
				testing("Signal %s has no effect on open().", sigs[i].signame);
				start
					Zopen(fifop, O_WRONLY);
				finis
		
				done();
			}
		
			prep("Issue a %s.",sigs[i].signame);
			start
				SLEEP(WAIT_FOR_TIME);	/* Allow child to call open(). */
				Zkill(pid, sigs[i].sig);
				SLEEP(WAIT_FOR_TIME);
			finis

			cleanup("Open FIFO's read channel so child can exit.");
			start
				sv_flush();
								/*	Allow for child to process open() before
									closing read channel.
								*/
				fd = Zopen(fifop, O_RDONLY);
			finis

			WAIT_FOR_PID(pid);
			Zclose(fd);

												/* Testing: pause() */
			 prep("Fork a child process to test pause().");
			 start
				sv_flush();
				pid = Zfork();
			finis
			
			if(pid == (pid_t)0) {
				prep("Block signal action for signal %s.", sigs[i].signame);
				start
					Zsigemptyset(&mask);
					SIGBLOCK(mask, sigs[i].sig);
				finis
	
				prep("Child calls pause().");
				start
					sv_flush();
					pause();
				finis
		
				done();
			}
		
			prep("Issue a %s.",sigs[i].signame);
			start
				SLEEP(WAIT_FOR_TIME);	/* Allow child to call pause(). */
				Zkill(pid, sigs[i].sig);
				SLEEP(WAIT_FOR_TIME);	/* Allow kill() to be processed. */
			finis

			testing("Signal %s has no effect on pause().", sigs[i].signame);
			start
										/* Test process is still active. */
				ret_val = Zwaitpid(pid, &stat_val, WNOHANG);
				if(ret_val != (pid_t)0) {
					if(WIFEXITED(stat_val) || WIFSIGNALED(stat_val) ||
														WIFSTOPPED(stat_val))
						error("Child has exited or is stopped.");
					else
						error("%s\n%s\n%s%ld.",
								"Call to waitpid() with WNNOHANG failed.",
								"Expected: return value of zero .",
								"Actual: return value was ",
								(long)ret_val);
				}
			finis

			cleanup("Terminate child process.");
			start
				Zkill(pid, SIGKILL);
			finis

			WAIT_FOR_PID(pid);

													/* Testing: read() */
			 prep("Fork a child process to test read().");
			 start
				sv_flush();
				pid = Zfork();
			finis
			
			if(pid == (pid_t)0) {
				prep("Block signal action for signal %s.", sigs[i].signame);
				start
					Zsigemptyset(&mask);
					SIGBLOCK(mask, sigs[i].sig);
				finis
	
				prep("Open the FIFO read channel.");
				start
					fiford = Zopen(fifop, O_RDONLY);
				finis

				testing("Signal %s has no effect on read().", sigs[i].signame);
				start
					expecting(Strlen);
					Zread(fiford, inbuf, Strlen);
				finis
		
				done();
			}
		
			prep("Open the FIFO write channel.");
			start
				fifowr = Zopen(fifop, O_WRONLY);
			finis

			prep("Issue a %s.",sigs[i].signame);
			start
				sv_flush();
				SLEEP(WAIT_FOR_TIME);	/* Allow child to call read(). */
				Zkill(pid, sigs[i].sig);
				SLEEP(WAIT_FOR_TIME);
			finis

			cleanup("Write data to FIFO so child can exit.");
			start
				expecting(Strlen);
				Zwrite(fifowr, String, Strlen);
				Zclose(fifowr);
			finis

			WAIT_FOR_PID(pid);

												/* Testing:  sleep() */
						/* SIGALRM not tested.
						   Sending SIGALRM to a sleep()ing process may
						   cause the sleep() to return early.
						*/
			if(sigs[i].sig != SIGALRM) {
				prep("Fork a child process to test sleep().");
				start
					sv_flush();
					pid = Zfork();
				finis
			
				if(pid == (pid_t)0) {
					prep("Block signal action for signal %s.", sigs[i].signame);
					start
						Zsigemptyset(&mask);
						SIGBLOCK(mask, sigs[i].sig);
					finis
	
					prep("Child calls sleep().");
					start
						sv_flush();
						sleep(TIMEOUT);
					finis
		
					done();
				}
		
				prep("Issue a %s.",sigs[i].signame);
				start
					SLEEP(WAIT_FOR_TIME);	/* Allow child to call sleep(). */
					Zkill(pid, sigs[i].sig);
					SLEEP(WAIT_FOR_TIME);	/* Allow kill() to be processed. */
				finis

				testing("Signal %s has no effect on sleep().", sigs[i].signame);
				start
									/* Test child process is still active. */
					if(Zwaitpid(pid, &stat_val, WNOHANG|WUNTRACED) != (pid_t)0)
						if(WIFEXITED(stat_val) || WIFSIGNALED(stat_val) ||
														WIFSTOPPED(stat_val))
							error("Child has exited or is stopped.");
				finis

				cleanup("Terminate child process.");
				start
					Zkill(pid, SIGKILL);
				finis

				WAIT_FOR_PID(pid);
			}

											/* Testing:  sigsuspend() */
			prep("Fork a child process to test sigsuspend().");
			start
				sv_flush();
				pid = Zfork();
			finis
			
			if(pid == (pid_t)0) {
				prep("Block signal action for signal %s.", sigs[i].signame);
				start
					Zsigemptyset(&mask);
					SIGBLOCK(mask, sigs[i].sig);
				finis
	
				prep("Child calls sigsuspend()");
				start
					sv_flush();
					sigsuspend(&mask);
				finis
		
				done();
			}
		
			prep("Issue a %s.",sigs[i].signame);
			start
				SLEEP(WAIT_FOR_TIME);	/* Allow child to call sigsuspend(). */
				Zkill(pid, sigs[i].sig);
				SLEEP(WAIT_FOR_TIME);	/* Allow kill() to be processed. */
			finis

			testing("Signal %s has no effect on sigsuspend().",sigs[i].signame);
			start
									/* Test child process is still active. */
				if(Zwaitpid(pid, &stat_val, WNOHANG|WUNTRACED) != (pid_t)0)
					if(WIFEXITED(stat_val) || WIFSIGNALED(stat_val) ||
														WIFSTOPPED(stat_val))
						error("Child has exited or is stopped.");
			finis

			cleanup("Terminate child process.");
			start
				Zkill(pid, SIGKILL);
			finis

			WAIT_FOR_PID(pid);

											/* Testing:  wait() */
			prep("Fork a child process to test wait().");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Fork a grandchild process.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {
					prep("Grandchild sleeps for awhile.");
					start
						SLEEP(4 * WAIT_FOR_TIME);
					finis
	
					done();
				}
			
				prep("Block signal action for signal %s.", sigs[i].signame);
				start
					Zsigemptyset(&mask);
					SIGBLOCK(mask, sigs[i].sig);
				finis
	
				prep("Child calls wait().");
				start
					sv_flush();
					Zwait(&stat_val);
				finis
		
				done();
			}
		
			prep("Issue a %s.",sigs[i].signame);
			start
				SLEEP(WAIT_FOR_TIME);	/* Allow child to call wait(). */
				Zkill(pid, sigs[i].sig);
				SLEEP(WAIT_FOR_TIME);	/* Allow kill() to be processed. */
			finis

			testing("Signal %s has no effect on wait().", sigs[i].signame);
			start
									/* Test child process is still active. */
				if(Zwaitpid(pid, &stat_val, WNOHANG|WUNTRACED) != (pid_t)0)
					if(WIFEXITED(stat_val) || WIFSIGNALED(stat_val) ||
														WIFSTOPPED(stat_val))
						error("Child has exited or is stopped.");
			finis

			WAIT_FOR_PID(pid);

												/* Testing:  waitpid() */
			 prep("Fork a child process to test waitpid().");
			 start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Fork a grandchild process.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {
					prep("Grandchild sleeps for awhile.");
					start
						SLEEP(4 * WAIT_FOR_TIME);
					finis
	
					done();
				}
			
				prep("Block signal action for signal %s.", sigs[i].signame);
				start
					Zsigemptyset(&mask);
					SIGBLOCK(mask, sigs[i].sig);
				finis
	
				prep("Child calls waitpid().");
				start
					sv_flush();
					Zwaitpid(pid, &stat_val, 0);
				finis
		
				done();
			}
		
			prep("Issue a %s.",sigs[i].signame);
			start
				SLEEP(WAIT_FOR_TIME);	/* Allow child to call waitpid(). */
				Zkill(pid, sigs[i].sig);
				SLEEP(WAIT_FOR_TIME);	/* Allow kill() to be processed. */
			finis

			testing("Signal %s has no effect on waitpid().", sigs[i].signame);
			start
									/* Test child process is still active. */
				if(Zwaitpid(pid, &stat_val, WNOHANG|WUNTRACED) != (pid_t)0)
					if(WIFEXITED(stat_val) || WIFSIGNALED(stat_val) ||
														WIFSTOPPED(stat_val))
						error("Child has exited or is stopped.");
			finis

			WAIT_FOR_PID(pid);

												/* Testing:  write() */
			if(sigs[i].sig == SIGALRM)
				continue;

			 ENABLE_PSYNC
			 prep("Fork a child process to test write().");
			 start
				sv_flush();
				pid = Zfork();
			finis
			
			if(pid == (pid_t)0) {
				prep("Block signal action for signal %s.", sigs[i].signame);
				start
					Zsigemptyset(&mask);
					SIGBLOCK(mask, sigs[i].sig);
				finis
	
				prep("Open the FIFO write channel.");
				start
					fifowr = Zopen(fifop, O_WRONLY);
				finis
	
				prep("Fill up FIFO's write channel.");
				start
					size_t	n_bytes;
					size_t	nb_wr = 0;

					SET_TIMEOUT
						for(n_bytes = fifo_buf_sz; n_bytes; n_bytes /= 2)
							while(fifo_max -nb_wr >= n_bytes) {
								expecting(n_bytes);
								nb_wr += Zwrite(fifowr, outbuf, n_bytes);
							}
					CLEAR_ALARM
					SEND_PSYNC
				finis

				testing("Signal %s has no effect on write().", sigs[i].signame);
				start
					expecting((int)fifo_buf_sz);
					Zwrite(fifowr, outbuf, fifo_buf_sz);
				finis
		
				done();
			}
		
			prep("Open the FIFO read channel.");
			prep("Wait for child to block on write().");
			start
				sv_flush();
				fiford = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				WAIT_PSYNC
				SLEEP(WAIT_FOR_TIME);	/* Allow child to call write(). */
			finis

			testing("Issue a %s.",sigs[i].signame);
			start
				Zkill(pid, sigs[i].sig);
				SLEEP(WAIT_FOR_TIME);	/* Allow kill() to get processed. */
			finis

			cleanup("Read data from FIFO so child can exit.");
			start
				expecting((int)fifo_buf_sz);
				Zread(fiford, in_buf, fifo_buf_sz);
								/* Read it all. */
				while(read(fiford, in_buf, fifo_buf_sz) > 0)
					;
								/* Have child complete before closing FIFO. */
				WAIT_FOR_PID(pid);
				Zclose(fiford);
			finis

			DISABLE_PSYNC
		}
	finis
	
	done();
}
