#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>


int	main(void)
{
	pid_t		pid;
	sigset_t	mask;
	int			stat_val;

	setup(E_SIGSUSPEND);
	funct(E_SIGSUSPEND);

#define	TIME_TO_DELAY	(4 * WAIT_FOR_TIME)
	assertion(4, BASE);
	start
		prep("Clear signal process mask.");
		start
			Zsigemptyset(&mask);
			Zsigprocmask(SIG_SETMASK, &mask, (sigset_t *) 0);
		finis

		ENABLE_PSYNC
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Establish initial process's signal mask with set of");
			prep("signals set to SIGUSR1.");
			start
				Zsigemptyset(&mask);
				Zsigaddset(&mask, SIGUSR1);
				Zsigprocmask(SIG_SETMASK, &mask, (sigset_t *) 0);
			finis

			prep("Establish a signal mask to be used as argument to");
			prep("sigsuspend with process's signal mask set to SIGHUP.");
			start
				Zsigemptyset(&mask);
				Zsigaddset(&mask, SIGHUP);
			finis

			prep("Catch SIGINT and SIGHUP and call sigsuspend().");
			start
				SIGNAL(SIGINT, sig_int);
				SIGNAL(SIGHUP, sig_int);

				SET_ALARM(TIMEOUT)
					SEND_PSYNC
					sigsuspend(&mask);
				CLEAR_ALARM
			finis

			testing("Process was suspended until delivery of a");
			testing("signal-catching function.");
			start
				if(unused_time > (TIMEOUT -TIME_TO_DELAY +1))
					error("The process was not suspended.");
			finis


										/* Test for Assertion #6. */
			testing("Signal mask is restored to the set that existed");
			testing("prior to the sigsuspend() call.");
			start
				Zsigprocmask(SIG_SETMASK, (sigset_t *)0, &mask);
				if(!sigismember(&mask, SIGUSR1))
					error("Signal mask was not restored.");
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		testing("A sigsuspend() call replaces the process's signal mask.");
		start
			Zkill(pid, SIGHUP);		/* Process should be blocking this signal */
			SLEEP(WAIT_FOR_TIME);
									/* Test process is still active. */
			if(Zwaitpid(pid, &stat_val, WNOHANG | WUNTRACED) != (pid_t)0)
				error("Child has exited or is stopped.");

			SLEEP(TIME_TO_DELAY -WAIT_FOR_TIME)

			Zkill(pid, SIGINT);		/* Should cause child process to continue.*/
		finis

		WAIT_FOR_PID(pid);
		DISABLE_PSYNC
	finis

	assertion(5, BASE);
	start
		ENABLE_PSYNC
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Set action for SIGINT to SIG_DFL.");
			start
				SIGNAL(SIGINT, SIG_DFL);
			finis

			testing("No return from sigsuspend() call.");
			start
				Zsigemptyset(&mask);
				SEND_PSYNC
				expecting(EINTR);
				Zsigsuspend(&mask);
				error("%s\n%s",
						"Sigsuspend returned.",
						"Expected no return.");
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		Zkill(pid, SIGINT);				/* Terminate child process. */
		WAIT_FOR_PID(pid);
		DISABLE_PSYNC
	finis

	assertion(6, BASE);
	start
		testing("Tested as part of the test for Assertion #4.");
		start
		finis
	finis

	assertion(7, BASE);
	start
		ENABLE_PSYNC
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == 0) {
			prep("Establish a signal mask with SIGKILL signal set.");
			prep("Call sigsuspend().");
			start
				sv_flush();
				Zsigemptyset(&mask);
				Zsigaddset(&mask, SIGKILL);

				SEND_PSYNC
				sigsuspend(&mask);
				error("Call to sigsuspend() returned.");
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
			SLEEP(WAIT_FOR_TIME)
		CLEAR_ALARM

		testing("Child process active (sigsuspend() call was successful).");
		start
			if(Zwaitpid(pid, &stat_val, WNOHANG|WUNTRACED) != (pid_t)0)
				if(WIFEXITED(stat_val) || WIFSIGNALED(stat_val) ||
														WIFSTOPPED(stat_val))
					error("Child has exited or is stopped.");
		finis

		testing("Sending a SIGKILL signal to child terminates child.");
		start
			Zkill(pid, SIGKILL);
			SET_TIMEOUT
				Zwaitpid(pid, &stat_val, 0);
				if(WIFSIGNALED(stat_val)) {
					if(WTERMSIG(stat_val) != SIGKILL)
						error("%s\n%s%s",
							"Expected child to be terminated by SIGKILL.",
							"Actually child terminated by ",
							sig_lookup(WTERMSIG(stat_val)) );
				}
				else
					error("Child not terminated by SIGKILL signal.");
			CLEAR_ALARM
		finis

		DISABLE_PSYNC
	finis

	assertion(9, BASE);
	start
		ENABLE_PSYNC
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			int	ret_val, s_errno;

			prep("Catch signal SIGINT and call sigsuspend() with an empty");
			prep("signal mask argument.");
			start
				Zsigemptyset(&mask);		/* Clear process signal mask. */
				Zsigprocmask(SIG_SETMASK, &mask, (sigset_t *) 0);

				SIGNAL(SIGINT, sig_int);
				Zsigemptyset(&mask);
				errno = 0;

				SET_TIMEOUT
					SEND_PSYNC
					ret_val = sigsuspend(&mask);
					s_errno = errno;
				CLEAR_ALARM
			finis

			testing("Return from call to sigsuspend() sets errno to [EINTR]");
			testing("and return value is -1.");
			start
				if(ret_val != (int) -1)
					error("%s\n%s\n%s%d",
						"Incorrect return.",
						"Expected: -1.",
						"Actual: ",
						ret_val);
				if(s_errno != EINTR)
					error("The value of errno returned was %s.",
							err_lookup(s_errno));
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		SLEEP(1);
		Zkill(pid, SIGINT);
		SET_TIMEOUT
			Zwaitpid(pid, &stat_val, 0);
			if(WIFSIGNALED(stat_val))
				error("%s\n%s%s",
					"Expected normal exit by child process.",
					"Actually child terminated by ",
					sig_lookup(WTERMSIG(stat_val)) );
		CLEAR_ALARM
		DISABLE_PSYNC
	finis

	assertion(8, BASE);
	start
		ENABLE_PSYNC
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == 0) {
			prep("Call sigsuspend() with signal mask set to SIGSTOP.");
			start
				sv_flush();
				Zsigemptyset(&mask);
				Zsigaddset(&mask, SIGSTOP);

				SIGNAL(SIGINT, SIG_DFL);

				SEND_PSYNC
				sigsuspend(&mask);
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
			SLEEP(WAIT_FOR_TIME)
		CLEAR_ALARM

		testing("Child process active (sigsuspend() call was successful).");
		start
			if(waitpid(pid, &stat_val, WNOHANG) != (pid_t)0) {
				if(WIFSTOPPED(stat_val))
					error("Child is stopped.");
				else
					error("Child has exited.");
			}
		finis

		testing("Sending a SIGSTOP signal to child stops child.");
		start
			Zkill(pid, SIGSTOP);
			SLEEP(WAIT_FOR_TIME);
			SET_TIMEOUT
				Zwaitpid(pid, &stat_val, WUNTRACED);
				if(WIFSTOPPED(stat_val)) {
					if(WSTOPSIG(stat_val) != SIGSTOP)
						error("%s\n%s%s",
							"Expected child to be stopped by SIGSTOP.",
							"Actually child stopped by ",
							sig_lookup(WSTOPSIG(stat_val)) );
				}
				else
					error("Child not stopped by SIGSTOP signal.");
				
				sv_flush();
			CLEAR_ALARM
		finis

										/* POSIX.1, page 53, lines 487-490. */
		testing("SIGINT signal sent to stopped child is not delivered");
		testing("until child is continued.");
		start
			Zkill(pid, SIGINT);
			SLEEP(WAIT_FOR_TIME)

							/* Check that SIGINT was not delivered. */
			if(waitpid(pid, &stat_val, WNOHANG) != (pid_t)0)
				error("SIGINT was delivered while child was stopped.");
		finis

		testing("SIGCONT signal sent to stopped child activates stopped");
		testing("process and SIGINT terminates sigsuspend().");
		start
			Zkill(pid, SIGCONT);

			SET_TIMEOUT
				stat_val = 0;
				Zwaitpid(pid, &stat_val, 0);
				if( ! WIFSIGNALED(stat_val))
					error("SIGCONT did not cause stopped child to continue.");
				else
					if(WTERMSIG(stat_val) != SIGINT) {
					  error("SIGCONT did not cause stopped child to continue.");
				}
			CLEAR_ALARM
		finis

		DISABLE_PSYNC
	finis

	done();
}
