#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<pcts_timer.h>


int	main(void)
{
	int		stat_val;
	pid_t	pid, ret;

	setup(E_WAITPID);
	funct(E_WAITPID);

	assertion(16, BASE);
	start
		struct	sigaction	act1, act2;

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Establish an action to terminate on SIGINT.");
			start
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = SIG_DFL;
				act1.sa_flags = 0;
				Zsigaction(SIGINT, &act1, &act2);
			finis

			prep("Have child send itself a signal.");
			start
				sv_flush();
				kill(getpid(), SIGINT);
			finis

			done();
		}

		ret = waitpid(pid, &stat_val, 0);

		testing("Macro WIFSIGNALED() evaluates to non-zero.");
		testing("Macros WIFEXITED() and WIFSTOPPED() evaluate to zero.");
		start
			if(WIFSIGNALED(stat_val) == 0)
				error("WIFSIGNALED() evaluated to zero.");

			if(WIFEXITED(stat_val) != 0)
				error("WIFEXITED() evaluated to non-zero.");

			if(WIFSTOPPED(stat_val) != 0)
				error("WIFSTOPPED() evaluated to non-zero.");
		finis
	finis

	assertion(17, BASE);
	start
		testing("When WIFSIGNALED() is non-zero, WTERMSIG() evaluates to the");
		testing("number of the signal that caused the termination of the");
		testing("child process.");
		start
			if(WIFSIGNALED(stat_val)) {
				if(WTERMSIG(stat_val) != SIGINT)
					error("%s\n%s%d",
						"Expected: child process to be terminated by SIGINT. ",
						"Actual: value returned from macro WTERMSIG is ",
						WTERMSIG(stat_val) );
			}
			else
				error("WIFSIGNALED reports no signal terminated the child.");
		finis
	finis

	assertion(18, BASE);
	start
		if(sysconf(_SC_JOB_CONTROL) != (long)-1) {
			ENABLE_PSYNC
			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Have process wait awhile before continuing.");
				start
					sv_flush();
					SEND_PSYNC
					WAIT_PSYNCA
				finis

				done();
			}

			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			prep("Send SIGSTOP to child.");
			start
				Zkill(pid, SIGSTOP);
			finis

			waitpid(pid, &stat_val, WUNTRACED);

			testing("Macro WIFSTOPPED() evaluates to non-zero.");
			testing("Macros WIFEXITED() and WIFSIGNALED() evaluate to zero.");
			start
				if(WIFSTOPPED(stat_val) == 0)
					error("WIFSTOPPED() evaluated to zero.");

				if(WIFEXITED(stat_val) != 0)
					error("WIFEXITED() evaluated to non-zero.");

				if(WIFSIGNALED(stat_val) != 0)
					error("WIFSIGNALED() evaluated to non-zero.");
			finis

			cleanup("Send child SIGCONT so it can exit.");
			start
				Zkill(pid, SIGCONT);
				SEND_PSYNCA
			finis

			WAIT_FOR_PID(pid);
			DISABLE_PSYNC
		}
		else
			error("FIPS required option {_POSIX_JOB_CONTROL} not supported.");
	finis

	assertion(19, BASE);
	start
		if(sysconf(_SC_JOB_CONTROL) != (long)-1) {
			ENABLE_PSYNC
			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Have process wait awhile before continuing.");
				start
					sv_flush();
					SEND_PSYNC
					WAIT_PSYNCA
				finis

				done();
			}

			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			prep("Send SIGSTOP to child.");
			start
				Zkill(pid, SIGSTOP);
			finis

			testing("Call to wait() does not return when child is stopped.");
			start
				SET_TIMEOUT
					waitpid(pid, &stat_val, 0);
				EXPECT_ALARM
			finis

			cleanup("Send child SIGCONT so it can exit.");
			start
				Zkill(pid, SIGCONT);
				SEND_PSYNCA
			finis

			WAIT_FOR_PID(pid);
			DISABLE_PSYNC
		}
		else
			error("FIPS required option {_POSIX_JOB_CONTROL} not supported.");
	finis

	assertion(20, BASE);
	start
		if(sysconf(_SC_JOB_CONTROL) != (long)-1) {
			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Stop child process.");
				start
					Zkill(getpid(), SIGSTOP);
					SLEEP(WAIT_FOR_TIME);
				finis
	
				done();
			}
	
			SLEEP(WAIT_FOR_TIME);
	
			testing("When macro WIFSTOPPED is non-zero the macro WSTOPSIG");
			testing("evaluates to the number of the signal that caused the");
			testing("process to stop.");
			start
				SET_ALARM(TIMEOUT)
					Zwaitpid(pid, &stat_val, WUNTRACED);
				CLEAR_ALARM
	
				if( ! WIFSTOPPED(stat_val)) 
					error("WIFSTOPPED evaluated to zero.");
				else
					if(WSTOPSIG(stat_val) != SIGSTOP) 
						error("%s\n%s\n%s%s\n",
							"WSTOPSIG() not equal to expected signal.",
							"Expected: SIGSTOP",
							"Actual: ",
							sig_lookup(WSTOPSIG(stat_val)) );
			finis
	
			cleanup("Terminate stopped child.");
			start
				Zkill(pid, SIGKILL);
				WAIT_FOR_PID(pid);
			finis
		}
		else
			error("FIPS required option {_POSIX_JOB_CONTROL} not supported.");
	finis

	done();
}
