#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	FILENAME	"3.2.1b_21"


int	main(void)
{
	int		stat_val;
	pid_t	pid;
	struct	sigaction	act1, act2;

	setup(E_WAITPID);
	funct(E_WAITPID);

	assertion(21, BASE);
	start
		FORK
			prep("Have child exit.");
			start
				sv_flush();
			finis
		KROF

		testing("Call to waitpid() when process does not exist returns -1");
		testing("and sets errno to [ECHILD].");
		start
			expecting(ECHILD);
			expecting(-1);
			Zwaitpid(pid, &stat_val, 0);
		finis

		testing("Call to waitpid() when process group does not exist");
		testing("returns -1 and sets errno to [ECHILD].");
		start
			expecting(ECHILD);
			expecting(-1);
			Zwaitpid((pid_t)(-1*pid), &stat_val, 0);
		finis
	finis

	assertion(22, BASE);
	start
		testing("Call to waitpid(), when pid is -1 and the calling process");
		testing("does not have any child processes,");
		testing("returns -1 and sets errno to [ECHILD].");
		start
			expecting(ECHILD);
			expecting(-1);
			Zwaitpid(-1, &stat_val, 0);
		finis
	finis

	assertion(23, BASE);
	start
		pid_t	c_pid, s_pgrp;

		FORK
			prep("Create a child process.");
			start
				sv_flush();
				c_pid = Zfork();
			finis

			if(c_pid == (pid_t)0) {		/* Generate process status report. */
				done();
			}

			SLEEP(WAIT_FOR_TIME);		/* Allow for child to complete. */
			prep("Establish a new process group for parent process.");
			start
				s_pgrp = Zgetpgrp();
				pid = Zgetpid();
				Zsetpgid((pid_t)0, pid);
			finis

			testing("When pid is equal to 0, call to waitpid() when calling");
			testing("process does not have any child processes in its");
			testing("process group.");
			start
				expecting(-1);
				expecting(ECHILD);
				Zwaitpid(0, &stat_val, 0);
			finis

			prep("Revert back to previous process group.");
			start
				Zsetpgid((pid_t)0, s_pgrp);
			finis

			testing("The wait status of the child process remains");
			testing("available for subsequent call by parent.");
			start
				expecting(SUCCESS);
				Zwaitpid(c_pid, &stat_val, 0);
			finis
		KROF
	finis

	assertion(24, BASE);
	start
		pid_t	c_pid, s_pgrp;

		FORK
			prep("Create a child process.");
			start
				sv_flush();
				c_pid = Zfork();
			finis

			if(c_pid == (pid_t)0) {		/* Generate process status report. */
				done();
			}

			SLEEP(WAIT_FOR_TIME);		/* Allow for child to complete. */
			prep("Establish a new process group for parent process.");
			start
				s_pgrp = Zgetpgrp();
				pid = Zgetpid();
				Zsetpgid((pid_t)0, pid);
			finis

			testing("When pid is less than -1, call to waitpid() when calling");
			testing("process does not have any child processes in its");
			testing("process group.");
			start
				expecting(-1);
				expecting(ECHILD);
				Zwaitpid((pid_t)(-1*pid), &stat_val, 0);
			finis

			prep("Revert back to previous process group.");
			start
				Zsetpgid((pid_t)0, s_pgrp);
			finis

			testing("The wait status of the child process remains");
			testing("available for subsequent call by parent.");
			start
				expecting(SUCCESS);
				Zwaitpid(c_pid, &stat_val, 0);
			finis
		KROF
	finis

	assertion(25, BASE);
	start
		int		fd;
		pid_t	gc_pid;
		char	path[_POSIX_PATH_MAX +1];

		ENABLE_PSYNC
			prep("Establish path for data file.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, FILENAME);
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Fork a grandchild process.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {
					prep("Have grandchild pass process ID to parent process.");
					start
						sv_flush();
						fd = Zcreat(path, PROT_ALL);
						pid = getpid();
						Zwrite(fd, (char *)&pid, sizeof(pid_t));
						Zclose(fd);
	
						SEND_PSYNC

						SET_TIMEOUT
							WAIT_PSYNCA
						CLEAR_ALARM
					finis
	
					done();
				}
	
				testing("The wait status of the grandchild process remains");
				testing("available for subsequent waitpid() call by parent.");
				start
					expecting(SUCCESS);
					Zwaitpid(pid, &stat_val, 0);
				finis

				done();
			}
					/* Wait for grandchild to write out its pid. */
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			testing("A call to waitpid() when the process ID is not a child");
			testing("of the calling process returns -1 and sets errno to");
			testing("[ECHILD].");
			start
				fd = Zopen(path, O_RDONLY, PROT_ALL);
				expecting(sizeof(pid_t));
				Zread(fd, (char *)&gc_pid, sizeof(pid_t));
				Zclose(fd);

				expecting(ECHILD);
				expecting(-1);
				Zwaitpid(gc_pid, &stat_val, WNOHANG);
			finis
	
			cleanup("Wait for child process to exit.");
			start
				SEND_PSYNCA
				SET_ALARM(TIMEOUT);
					Zwaitpid(pid, &stat_val, 0);
				CLEAR_ALARM
			finis
		DISABLE_PSYNC
	finis

	assertion(26, BASE);
	start
		ENABLE_PSYNC

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Fork a grandchild process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Have grandchild send parent a signal.");
				start
					sv_flush();
					SET_TIMEOUT
						WAIT_PSYNC
					CLEAR_ALARM

					SLEEP(1)
					Zkill(getppid(), SIGINT);
				finis

				done();
			}

			prep("Establish SIG_DFL for SIGINT signal.");
			start
				sv_flush();
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = SIG_DFL;
				act1.sa_flags = 0;
				Zsigaction(SIGINT, &act1, &act2);
			finis

			SLEEP(1);	/* Allow previous "prep" messages to be outputted. */

			testing("Call to waitpid() interrupted by a signal, returns -1");
			testing("and sets errno to [EINTR].");
			start
				SEND_PSYNC
				expecting(EINTR);
				expecting(-1);
				Zwaitpid(pid, &stat_val, 0);
			finis

			testing("The wait status of the grandchild process remains");
			testing("available for subsequent call, by parent, of");
			testing("waitpid().");
			start
				Zwaitpid(pid, &stat_val, 0);
			finis

			done();
		}
		WAIT_FOR_PID(pid);
		DISABLE_PSYNC
	finis

	assertion(27, BASE);
	start
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Have child delay then exit.");
			start
				sv_flush();
				SLEEP(2 * WAIT_FOR_TIME);
			finis

			done();
		}

		SLEEP(WAIT_FOR_TIME);

		testing("Call to waitpid(), when the options argument is invalid,");
		testing("returns -1 and sets errno to [EINVAL].");
		start
			expecting(EINVAL);
			expecting(-1);
			Zwaitpid(pid, &stat_val, EINVAL_WAITPID);
		finis
	
		testing("The wait status of the child process remains");
		testing("available for subsequent call, by parent, of");
		testing("waitpid().");
		start
			expecting(SUCCESS);
			Zwaitpid(pid, &stat_val, 0);
		finis
	finis

	done();
}
