/* sv_env.h - macro definitions for test control.  */

#include	<stdlib.h>
#include	<setjmp.h>
#include	<stddef.h>
#include	<sys/types.h>
#include	<termios.h>

#define	temp_path	sv_temp_path
#define	data_path	sv_data_path
extern	char	*temp_path(void);
extern	char	*data_path(void);

#define	SUBTMP		"subtmp"

#ifndef TRUE
#define	TRUE		1
#endif

#ifndef FALSE
#define FALSE		0
#endif

/*
 *	typedefs of pointer types, useful for casts
 */
typedef	char *pointer;
typedef	unsigned int sv_ugid;	/* typedef for comparison of [ug]id's */
typedef	int (*ifunctptr)();
typedef	pointer (*pfunctptr)();
/*
 *	copy struct sv_env's and jmpbuf's
 */
#define	sv_ecpy(a,b) (memcpy((pointer)a,(pointer)b,sizeof(struct sv_env)))
#define	sv_jcpy(a,b) (memcpy((pointer)a,(pointer)b,sizeof(sigjmp_buf)))

/*
 *	system detected error conditions
 *
 *	use as arguments to warn_on(), error_on(), severr_on().
 *
 *	failure	- funct returned failure when success was expected
 *  success	- funct returned success when failure was expected
 *  success	- funct returned unexpected value when explicit val expected
 *	errno	- funct returned expected failure, but with incorrect errno
 *	except	- caught signal, also use for other asynch problems
 *	undef	- funct returned value outside range of success/failure
 */
#define E_FAILURE		(1<<8)
#define E_SUCCESS		(2<<8)
#define E_UNEXP			(4<<8)
#define E_ERRNO			(8<<8)
#define	E_EXCEPT		(16<<8)
#define E_UNDEF			(32<<8)
#define	sv_fsev			sv_sev[0]	/* severity for unexp failure error	*/
#define sv_ssev			sv_sev[1]	/* severity for unexp success error	*/
#define sv_esev			sv_sev[2]	/* severity for unexp errno error	*/
#define sv_xsev			sv_sev[3]	/* severity for unexp signal error	*/
#define sv_vsev			sv_sev[4]	/* severity for unexp ret val error	*/
#define	sv_usev			sv_sev[5]	/* severity for undefined ret val	*/
#ifdef SV_ENV
int sv_err_table[] = {
	E_FAILURE,	E_SUCCESS,	E_UNEXP,	E_ERRNO,	E_EXCEPT,	E_UNDEF
};
#endif
#define	SV_NERRS	6	/*	( sizeof( sv_err_table )/sizeof( int ) )	*/

/*
 *	error types
 *
 *	these error flags are passed to the error reporting function
 *		to indicate the type of error, and also are used for
 *		tracking actions vs. error types
 *
 *	message - information only, used for suppressed tests
 *	warning	- will not prevent verification of system under test
 *	error	- prevents verification due to problem
 *	severr	- same as error, but more serious problem detected
 */
#define	SV_MID			0
#define	SV_WID			1
#define	SV_EID			2
#define	SV_SID			3

#define MESSAGE			(1<<SV_MID)
#define WARNING			(1<<SV_WID)
#define	ERROR			(1<<SV_EID)
#define SEVERR			(1<<SV_SID)

#define	sv_mact			sv_act[SV_MID]	/* action for messages			*/
#define	sv_wact			sv_act[SV_WID]	/* action for warnings			*/
#define	sv_eact			sv_act[SV_EID]	/* action for errors 			*/
#define sv_sact			sv_act[SV_SID]	/* action for severrs			*/

#define	sv_mjmp			sv_jmpb[SV_MID]	/* jmpbuf for break on message	*/
#define	sv_wjmp			sv_jmpb[SV_WID]	/* jmpbuf for break on warning	*/
#define	sv_ejmp			sv_jmpb[SV_EID]	/* jmpbuf for break on error	*/
#define	sv_sjmp			sv_jmpb[SV_SID]	/* jmpbuf for break on severr	*/
#define	sv_jmp			sv_jmpb[SV_SID+1]/* jmpbuf for current block	*/

#ifdef SV_ENV
int sv_sev_table[] = {
	MESSAGE,  WARNING,	ERROR,	SEVERR
};
char *sv_snam_table[] = {
	"MESSAGE",  "WARNING",	"ERROR",	"SEVERR"
};
#endif
#define	SV_NSEVS		4	/*	( sizeof( sv_sev_table )/sizeof( int ) )*/
/*
 *	error actions
 *
 *	these flags are used in conjunction with the error flags to determine
 *		what action to take when an error is encountered.
 *
 *	abort	- terminate entire test
 *	exit	- terminate current process
 *	break	- terminate current test block
 *	continue- no termination
 */
#define	ABORT			1
#define	EXIT			2
#define	BREAK			4
#define	CONTINUE		8
/*
 *	internal set of possible outcomes which are expected,
 *	these macros are not intended for use by the test writer.
 *
 *	possible outcomes which can be expected are:
 *
 *		SV_VALU:	success with a particular value
 *		SV_SUCC:	success
 *		SV_FAIL:	failure
 *		SV_ERRN:	failure with a particular errno
 *		SV_VOID:	don't check return value
 *		SV_LOCK:	set when in the expecting() macro
 *
 *	these are used when the expecting() macro is invoked, see the
 *	following section.
 */
#define SV_VALU			1
#define SV_SUCC			2
#define SV_FAIL			4
#define SV_ERRN			8
#define SV_VOID			16
#define SV_LOCK			32

/*
 *	expected returns from functions
*/
extern	int	sv_sets(void);
extern	int	sv_setf(void);
extern	int	sv_setv(void);
/*
 *
 *	These values are used as arguments to expecting() to indicate the
 *		expected result from a function call.
 *
 *		failure	- expecting function to fail
 *		success - expecting function to succeed
 *		void	- disable automatic function checking
 *
 *	In addition, the test writer may specify either an errno value, or
 *	an expected value, in which case the evaluation function will check
 *	the return/errno against the values passed to expecting.  In order 
 *	to implement this, a second variable, sv_expecting, is set with the
 *	macros SV_SUCC, SV_FAIL, SV_VOID, SV_VALU, and SV_ERRN.
 */
#define	SUCCESS			sv_sets()
#define FAILURE			sv_setf()
#define VOID			sv_setv()
/*
 *	some useful macros for setting, testing, and clearing sv_expecting.
 */
#define sv_svalu( xnv )	(xnv->sv_expecting |= SV_VALU)
#define sv_serrn( xnv )	(xnv->sv_expecting |= SV_ERRN)
#define sv_ssucc( xnv )	(xnv->sv_expecting = SV_SUCC)
#define sv_sfail( xnv )	(xnv->sv_expecting = SV_FAIL)
#define sv_svoid( xnv )	(xnv->sv_expecting = SV_VOID)
#define sv_slock( xnv )	(xnv->sv_expecting |= SV_LOCK)
#define sv_ltest( xnv )	(xnv->sv_expecting & SV_LOCK)
#define sv_ulock( xnv )	(xnv->sv_expecting &= ~SV_LOCK)

/*
 *	the expecting macro, passes the expected return status to sv_exp(),
 *	which saves it for later use by the evaluation function.
 */
#define	expecting(foo) 	(sv_slock(sv_xenv),sv_exp((char *)0, foo))

/*
 *	possible success/failure existences for a function
 *
 *		NULL:	undefined/untestable, requires expecting( VOID )
 *		VALUE:	defined, but requires expecting( "expression" )
 *		GONE:	function will not return for this state
 *		RET:	if function returns, it is in this state
 */
extern	int	sv_gone(void);
extern	int	sv_value(void);
extern	int	sv_return(void);
#define GONE			(sv_gone)
#define	VALUE			(sv_value)
#define	RET				(sv_return)

struct sv_env {
	struct sv_env *sv_oenv;		/* ptr to prev frame					*/
	int	sv_expecting;			/* expect: succ, fail, void, errno, val	*/
	int sv_cnt;					/* cnt of test cases at current frame	*/
	int	sv_fcnt;				/* cnt of children at current frame		*/
	int	sv_levflg;				/* need current level "1.2.3" printed?	*/
	int sv_sev[SV_NERRS];		/* severities for errors				*/
	int sv_act[SV_NSEVS];		/* actions for severities				*/
	sigjmp_buf	sv_jmpb[SV_NSEVS+1];	/* jmpbufs for breaks			*/
	int	sig_smask;				/* signal mask to restore on a break	*/
};

/*
 *	ftable is a structure describing the behavior of a function 
 *	for both success and failure.  ftable is used by the evaluation 
 *	function to determine whether an unexpected event has occurred.
 *
 *	There is at least one struct ftable for each function in the SVID.
 *	Some functions which have different return ranges for different 
 *	inputs may have more than one struct. These use a hard coded switch to
 *	determine which particular structure applies in a given case.
 *
 *	Test writers may define their own struct ftable's if they desire
 *	automatic return value checking for their own functions.  For examples
 *	see the supplied replacement functions.
 */
struct ftable {
	char	*sv_name;		/* function name							*/
	char	*sv_sstr;		/* string describing success				*/
	char	*sv_fstr;		/* string describing failure				*/
	char	*((*sv_fmt)());	/* formatting function						*/
	int		(*sv_test)();	/* funct which returns 1/0 for act == exp	*/
	int		(*sv_succ)();	/* funct which returns 1/0 for success		*/
	int		(*sv_fail)();	/* funct which returns 1/0 for failure		*/
	pointer	(*sv_xval)();	/* funct which returns ptr to exp ret val	*/
	int		sv_errnodef;	/* 1: errno defined on failure */
};

extern	void	sv_einit(struct sv_env *);
				/*	Allocate memory for struct sv_env on the heap rather
					than the stack. (Changed 5/24/94).
				*/
#define sv_begblk	{ 			struct  sv_env *sv_eblk_ptr; 			sv_eblk_ptr = (struct  sv_env *)malloc(sizeof(struct  sv_env)); 			sv_einit(sv_eblk_ptr); 			if(!sigsetjmp(sv_xenv->sv_jmp, sv_xenv->sig_smask)) {

#define sv_endblk	} 					sv_xenv = sv_eblk_ptr->sv_oenv; 					free(sv_eblk_ptr); 					} 					sv_xenv->sv_levflg=1; 					sv_xenv->sv_cnt++;

extern	void	sv_config(void);
extern	void	sv_nproc(void);
extern	void	sv_message(char *, ...);
extern	void	sv_warning(char *, ...);
extern	void	sv_error(char *, ...);
extern	void	sv_severr(char *, ...);
extern	void	sv_prerr(int, char *, ...);
extern	void	sv_exp(char *, ...);
extern	void	sv_printf(char *, ...);
extern	void	sv_abstract(char *, ...);
extern	void	sv_preparation(char *, ...);
extern	void	sv_testing(char *, ...);
extern	void	sv_cleanup(char *, ...);
extern	void 	unresolved(char *, char *);
extern	int		sv_cmd(char *, ...);
extern	char	*err_lookup(int);
extern	char	*sv_newlev(struct sv_env *, char *);
extern	char	*sig_lookup(int);
extern	void	evaluate(struct ftable *, pointer);
extern	void	sv_actset(int, int);
extern	void	sv_done(void);
extern	void	done_e(void);
extern	void	sv_flush(void);
extern	void	sv_sevset(int, int);
extern	void	subtest(char *, int);
extern	void	sv_setup(int);
extern	void	sv_funct(int);
extern	int		wait_for(pid_t, unsigned);
extern	void	termios_prep(int *, struct termios *, int *, struct termios *);
extern	void	termios_nct(int *, struct termios *, int *, struct termios *);
extern	void	not_implemented(char *);
#define	no_sig	sv_no_sig
extern	void	sv_no_sig(int);
extern	void	sv_set_exit_val(int);

/*	Functions located in  "src/tools/sv_tools"	*/
extern void	assertion(int, char);
extern void not_implemented(char *);
extern void yes_implemented(char *);
extern void un_implemented(char *);
extern void trf_ck(int, int, time_t, int, char *);
extern int  del_secs(int);
extern int  dt_secs(int);
extern int  data_open(char *, int, mode_t);
extern int  DIR_impl(void);
extern void not_implemented(char *);
extern char *g_lname(void);
extern int	tc_name_max(void);
extern int	tc_path_max(void);
extern long gi_path_max(void);
extern char *g_lpath(void);
extern char *g_path_max(void);
extern void mnt_fs(char *, int);
extern void umnt_fs(char *);
extern char *fs_mount(int, int);
extern void fs_umount(int);
extern void fsp_umount(int);
extern void mnt_fs(char *, int);
extern void umnt_fs(char *);
extern int	g_open_max(void);
extern int	get_open_max(void);
extern gid_t	gsup_gid(void);
extern long g_pipe_buf(char *, size_t *, size_t *);
extern long g_max_pipe(size_t *, size_t *, size_t *, size_t *);
extern void	put_env(char *);
extern int  temp_creat(char *, mode_t);
extern int  temp_open(char *, int, mode_t);
extern void baud_sleep(int, speed_t);
extern int	open_ctty(char *, int, mode_t);
extern void	tty_prep(int *, struct termios *);
extern void	loop_prep(int *, struct termios *);
extern void tty_init(int, struct termios *);
extern void loop_init(int, struct termios *);
extern char *get_nth_token(char *, char *, int);
extern char *get_token(char *, char *);
extern char *get_token(char *, char *);
extern size_t  g_max_input(char *);
extern speed_t baud_mask(const char *);
extern int baud_value(speed_t);
extern speed_t baud_min(void);
extern speed_t baud_max(void);
extern int baud_rate(speed_t);
extern void	chk_mcro(int, char *);

#define abstract		sv_abstract
#define preparation		sv_preparation
#define prep			sv_preparation
#define testing			sv_testing
#define cleanup			sv_cleanup
#define	setup(function)	sv_setup(function)
#define funct			sv_funct
#define done			sv_done
#define start			sv_begblk
#define finis			sv_endblk

#define warning_on		warn_on
#define warn_on(cond)	sv_sevset( cond, WARNING )
#define error_on(cond)	sv_sevset( cond, ERROR )
#define severr_on(cond)	sv_sevset( cond, SEVERR )
#define cont_on(sev)	sv_actset( sev, CONTINUE )
#define continue_on(sev)	cont_on( sev )
#define break_on(sev)	sv_actset( sev, BREAK )
#define exit_on(sev)	sv_actset( sev, EXIT )
#define abort_on(sev)	sv_actset( sev, ABORT )
#define reset(params)	sv_reset( params )

#define message			sv_message
#define warning			sv_warning
#define error			sv_error
#define severr			sv_severr


#define STRVAL	64			/* static area for each sv_fmt funct		*/

extern struct sv_env sv_deflt;	/* initial environment in sv_env.c		*/
extern struct sv_env *sv_env;	/* initial environment ptr, set in sv_env.c*/
extern struct sv_env *sv_xenv;	/* extern used for passing current env	*/

/*
 *	convenient macros to pass to expecting
 */
#define	NON_NULL	((pointer)01)
#define	ZERO		(sv_setz(),0.0)
