/*	This include file provides information that will only be used
 *	by the NIST-PCTS support modules.
*/

#define	PATH_CMD	""	/* If specified, absolute path for all commands below.
						   If null, uses environment variable PATH to
						   locate commands unless command begins with a '/'.
						*/

/*	Command and required flags for utility that executes the macro
	preprocessor for C routines and sends the results to standard output.
*/
#define	MACRO_PREPROCESSOR_COMMAND "cc -E"

/*	Command to perform a global recognition pattern search as performed by
	"grep" in POSIX.2.
*/
#define	GREP_COMMAND "grep"

/*	The following commands define the syntax for performing the
 *	required operation on the current POSIX implementation.
 *
 *	The "cmd" entry is mapped using the corresponding "svcomp" entry
 *	into the implementation specific command required to perform the
 *	required functionality.
 *
 *	The notation %@% maps the remaining entries of the svcomp file into
 *	the "cmd" string.
 *	The notation FLAGS%@% maps the remaining entries of the svcomp file
 *  into the "cmd" string with the "FLAGS" syntax preceding each entry.
 *
 *	The notation %object% maps the next argument from the "svcomp"
 *	entry into the "cmd" string.
 *
 *	Modify the command syntax as required for your implementation.
*/


struct	cmd_imp	{
		char	*symbol;
		char	*cmd;
};
static	struct	cmd_imp impcmd[] = {
	{
		/* Compile command for producing an object module */
		"CC1",
		"cc -c -I%include% %input%",
	},
	{
		/* Compile command for producing an executable module 
		   from an object module.
		*/
		"CC2",
		"cc -o %output% %input% %@%",
	},
	{
		/* Compile command for producing an executable module
		   from source code.
		*/
		"CC3",
		"cc -o %output% -I%include% %input% %@%",
	},
	{
		/* Compile command for producing an executable module from an
		   object module for routines
				./STD/LS/src/acos
				./STD/LS/src/asin
				./STD/LS/src/atan
				./STD/LS/src/ceil
				./STD/LS/src/cos
				./STD/LS/src/cosh
				./STD/LS/src/exp
				./STD/LS/src/fabs
				./STD/LS/src/floor
				./STD/LS/src/frexp
				./STD/LS/src/ldexp
				./STD/LS/src/log
				./STD/LS/src/log10
				./STD/LS/src/modf
				./STD/LS/src/pow
				./STD/LS/src/sin
				./STD/LS/src/sinh
				./STD/LS/src/sqrt
				./STD/LS/src/tan
				./STD/LS/src/tanh
		*/
		"CC4",
		"cc -o %output% %input% %@% -lm",
	},
	{
		/* Compile command for producing an executable module from an
		   object module for routine
				./STD/LS/src/setlocale
		*/
		"CC5",
		"cc -o %output% %input% %@%",
	},
	{
		/* Compile command for producing an executable module from an
		   object module for routine
				./STD/LS/src/fmod
		*/
		"CC6",
		"cc -o %output% %input% %@% -lm",
	},
	{
		/* Compile command for producing an executable module from an
		   object module for routine
				./STD/LS/src/strrchr
		*/
		"CC7",
		"cc -o %output% %input% %@%",
	},
	{
		/* Remove files command */
		"RM",
		"rm -f %@%",
	},
	{
		/* Archive files command */
		"AR",
		"ar -rc %archive_file% %@%",
	},
	{
		/* Copy files command */
		"CP",
		"cp %in_file% %out_file%",
	},
	{
		/* Move command */
		"MV",
		"mv %in_file% %location%",
	},
	{
		/* Remove files command for "verify" */
		"RMV",
		"rm -rf %directory%",
	},
	{
		/* Make directory command for "verify" */
		"MDV",
		"mkdir %directory%",
	},
	{
		/* Random library */
		"RL",
		/*"ranlib %library%", */
		"",
		/* If the RL command is not required because the loader
		   provides this functionality replace the
		   ["ranlib %library%",] line with ["",].
		*/
	},
	{
		/* Provide proper command for echoing a comment */
		"ECHO",
		"echo %@%",
	},
	{
		/* End of Symbols */
		(char *) 0,
		"END",
	},
};
