#ifndef	_POSIX_SOURCE
#define	_POSIX_SOURCE
#endif

#include <install.h>
#include <version.h>
#include <chk_mcro.h>

#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/wait.h>

#ifdef	BETA_PCTS
#include "../driver/validate.h"
#endif

#define	F_MODE	((mode_t)(S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH))
#undef	getc	/* Avoid lint complaints. */


void	parse(void);
int		fgetln(int, char *);
int		nonblank(char *str);
int		whitespace(char c);
void	f_source(char *str);
void	f_dest(char *str);
void	f_comp(char *str);
void	f_options(char *str);
void	f_libs(char *str);
void	f_sections(char *str);
void	f_print(char *);
void	f_badinput(char *str);
void	install(void);
void	build(char *);
char	*nxtok(char *, char *);
void	wr_all(char *);
void	clean(char *, char *);
char	*getdef(char *);
void	chunk(char *, char *);
int		execute(char *);
void	date_stamp(void);
void	getlibs(int, char *);
void	i_cmd(char *, char *);
char	*getcmd(char *);
FILE	*p_popen(char *);
int		p_pclose(FILE *);

struct	pstruct	{
	char	*value;
	void	(*funct)(char *str);
};

static struct	pstruct	keys[] = {
	{	"source",f_source	},
	{	"destination",f_dest,	},
	{	"compiler",f_comp	},
	{	"options",f_options	},
	{	"libraries",f_libs	},
	{	"sections",f_sections	},
	{	"print",f_print		},
	{	"Bad Input",f_badinput	},
};

#define	NUMKEYS	7
#define	DEFAULT	""


#define	EOLN	'\n'

/*	This program reads in a file which is made up of lines with
**	keywords on the beginning of each line and processes that file
**	according to the rules defined above. The constant NUMKEYS
**	is the number of keywords recognized by the parser.  The
**	array "keys" contains NUMKEYS + 1 entries, the last of which
**	handles unrecognized keywords. Each function will take as
**	its argument a string containing the data up to but not
**	including the next newline read from the input file
**	specified either by argv[1] or the  default input file 
**	defined by the constant DEFAULT.  If DEFAULT is a
**	null string and no file is specified, stdin is used for
**	input.
*/


static	int	journal = -1;

static	char	temp[2048];
static	char	curlin[2048];
static	char	cmd[4096];
static	char	icmd[4096];
static	char	buildfile[100]="build";

struct	settings{
	char	source[512];
	char	dest[512];
	char	compiler[512];
	char	options[512];
	char	libs[512];
	char	sections[512];
} setup = {
	"",
	"",
	"",
	"",
	"",
	"",
};

struct	param {
	char	*name;
	char	*value;
};


static	char	svlibs[512]="";

static struct	param parms[] ={
	{	"SOURCE",&setup.source[0]	},
	{	"DEST",&setup.dest[0]		},
	{	"COMP",&setup.compiler[0]	},
	{	"OPTS",&setup.options[0]	},
	{	"LIBS",&setup.libs[0]		},
	{	"SECT",&setup.sections[0]	},
	{	"SVLIBS",&svlibs[0]		},
	{	"Bad Input",""			},
};

#define	NUMPARMS ((sizeof(parms)/sizeof(struct param))-1)

#define	MAXCMD	80	/*Max size of a line in a system() buffer*/


int	main(int argc, char *argv[])
{
	int	fd, i, opened=0;

#ifdef	BETA_PCTS
	v_date_of_run();	/* Validate date of run for BETA_NIST-PCTS:151-2. */
#endif

	for(i = 1; i < argc; ++i){
		if(!opened &&(*argv[i] != '-')){
			if((fd = open(argv[i],O_RDONLY)) == -1){
				sprintf(temp,"%s: Cannot open - \"%s\"\n", argv[0], argv[i]);
				wr_all(temp);
				_exit(1);
			}
			opened = 1;
		}
		else {
			if(*argv[i] == '-'){
				sprintf(temp,"usage: %s [file]\n", argv[0]);
				wr_all(temp);
				_exit(-1);
			}
		}
	}
	if(!opened){
		fd = STDIN_FILENO;		/*If no files specified, then stdin*/
	}
	while(fgetln(fd,curlin)){
		parse();
	}
	if((fd > 2) && close(fd)){
		wr_all("Call to close() failed on input file.\n");
		_exit(-1);
	}
	install();
	_exit(0);
}

void	parse(void)
{
	int	i;
	char	key[30];
	char	*p_curlin;

	if(nonblank(curlin)){
		p_curlin = nxtok(key, curlin);
		for(i = 0;(i < NUMKEYS) && strcmp(key,keys[i].value); ++i);
		(*keys[i].funct)((i < NUMKEYS) ? p_curlin : key);
	}
}

int	fgetln(int fd, char *str)
{
	int	rval;
	char	*p;

	p = str;
	while((rval = (int) read(fd,str,1)) != 0 &&(*str != EOLN))
		++str;
	*str = 0;
	return(rval || !(p == str));
}

int	nonblank(char *str)
{
	while(*str){
		if(!whitespace(*str))
			return(1);
		++str;
	}
	return(0);
}

int	whitespace(char c)
{
	return((c == ' ') ||(c == '	') || (c == '\n'));
}

void	f_source(char *str)
{
	(void)nxtok(setup.source,str);
}

void	f_dest(char *str)
{
	(void)nxtok(setup.dest,str);
}

void	f_comp(char *str)
{
	(void)nxtok(setup.compiler,str);
}

void	f_options(char *str)
{
	static	char	first = 1;

	if(first){
		*setup.options = 0;
		first = 0;
	}
	while(*str){
		str = nxtok(temp,str);
		strcat(setup.options,temp);
		strcat(setup.options," ");
	}
}

void	f_libs(char *str)
{
	static	char	first = 1;

	if(first){
		*setup.libs = 0;
		first = 0;
	}
	while(*str){
		str = nxtok(temp,str);
		strcat(setup.libs,temp);
		strcat(setup.libs," ");
	}
}

void	f_sections(char *str)
{
	static	char	first = 1;

	if(first){
		*setup.sections = 0;
		first = 0;
	}
	while(*str){
		str = nxtok(temp,str);
		strcat(setup.sections,temp);
		strcat(setup.sections," ");
	}
}

void	f_print(char *dummmy)
{
	sprintf(temp, "Source directory is: %s\n", setup.source);
	wr_all(temp);
	sprintf(temp, "Destination directory is: %s\n", setup.dest);
	wr_all(temp);
	sprintf(temp, "Compiler is: %s\n", setup.compiler);
	wr_all(temp);
	sprintf(temp, "Options are: %s\n", setup.options);
	wr_all(temp);
	sprintf(temp, "Libs are: %s\n", setup.libs);
	wr_all(temp);
	sprintf(temp, "Sections are: %s\n\n", setup.sections);
	wr_all(temp);
}

void	f_badinput(char *str)
{
	sprintf(temp,"Bad Input: %s\n",str);
	wr_all(temp);
}

static	char	cursect[512];
static	char	jfile[512];

void	install(void)
{
	int	fd;
	char	*p,*q;
	static	char	curfile[512]; /*Conserve stack*/
	static	char	opath[512],path[512];

	/* Ensure paths are specified as absolute paths. */
	getcwd(path, 512);
	p = strstr(path, "NIST-PCTS");
	if(p == NULL) {
		sprintf(temp, "Improper path obtained from getcwd() \"%s\"\n", path);
		wr_all(temp);
		_exit(-1);
	}
									/* Get last occurrence of "NIST-PCTS". */
	while((q = strstr(p +1, "NIST-PCTS")) != NULL)
		p = q;
	*(p+strlen("NIST-PCTS")) = '\000';
		
	if(!(*setup.source))
		strcpy(setup.source, path);
	if(!(*setup.dest))
		strcpy(setup.dest, path);
	if((*setup.libs) &&(*setup.libs != '/')) {
		char	tmp_buf[512];
		strcpy(tmp_buf, setup.libs);
		strcpy(setup.libs, path);
		strcat(setup.libs, "/");
		strcat(setup.libs, tmp_buf);
	}

	if(!((*setup.source) ||(*setup.dest))){
		sprintf(temp, "Source or Destination not specified, cannot build.\n");
		wr_all(temp);
		_exit(-1);
	}
	if(access(setup.source, R_OK | X_OK) != 0 || 
    				access(setup.dest, R_OK | W_OK | X_OK) != 0) {
		sprintf(temp, "Source or Destination not accessable.  Cannot build.\n");
		wr_all(temp);
		f_print("");
		_exit(-1);
	}
	(void)sprintf(jfile,"%s/journal",setup.dest);
	if((journal = open(jfile,O_WRONLY|O_TRUNC|O_CREAT,F_MODE))<0){
		sprintf(temp, "Cannot create journal file \"%s\"", jfile);
		wr_all(temp);
		_exit(-1);
	}
								/* Data to journal. */
	sprintf(temp,"\n\t\t%s, %s\n\n\n", NIST_PCTS_NAME, NIST_PCTS_VERSION);
	wr_all(temp);
	p = &setup.sections[0];
	while((p = nxtok(cursect,p)),*cursect) {
												/* Processing Section STD/?? */
		wr_all(cursect);
		date_stamp();							/* Assign date to section. */
		(void)sprintf(opath,"%s/%s", setup.source, cursect);
		if(chdir(opath)){
			wr_all("\tERROR\n,\tchdir() failed unexpectedly\n");
			_exit(-1);
		}
		fd = open(buildfile,O_RDONLY);
		if(fd < 0){
			wr_all("\tERROR\n");
			sprintf(temp,"\tCan not open build file: %s/%s\n",opath,buildfile);
			wr_all(temp);
			_exit(-1);
		}
		getlibs(fd, svlibs);
		while(fgetln(fd, curfile)){
			q = nxtok(path, curfile);
												/* Processing element src/xxx */
			sprintf(temp, "  ./%s\n", path);
			wr_all(temp);
			if(chdir(path)){
				wr_all("\tERROR\n");
				sprintf(temp,"\tCan not chdir to %s/%s\n", opath, path);
				wr_all(temp);
				_exit(-1);
			}
			build(q);
			if(chdir(opath)){
				wr_all("\tERROR\n");
				sprintf(temp,"\tCan not chdir to %s\n",opath);
				wr_all(temp);
				_exit(-1);
			}
			wr_all("\n");
		}
		(void)close(fd);
	}
	if(journal > -1)
		(void)close(journal);
}

/*
** build(file)
** This function reads the file named in the string "file" and
** performs the operations specified in that file using the
** user defined parameters to substitute in the place of keywords
** delimited by % % in the file.
*/

void build(char *file)
{
	int		infile;
	int		p_macro = 0;		/* Element is not defined as a macro. */
	char	*l_str, *o_str, *p_str;
	int		stat_val;

	if((infile = open(file, O_RDONLY)) == -1){
		sprintf(temp, "\tERROR\n\topen(): Failed to open file: \"%s\"\n", file);
		wr_all(temp);
		_exit(-1);
	}
	while(fgetln(infile, curlin)){
		i_cmd(icmd, curlin);
		clean(cmd, icmd);
		sprintf(temp, "\t%s\n", cmd);
		wr_all(temp);
		if(execute(cmd)){
			wr_all("\tERROR\n\tCommand exited with an error.\n");
		}

		if(p_macro) {
					/* Only execute commands that create the load module. */
			if(strstr(cmd, " -o ") != NULL) {
					/*	Modify command to create load module appended with 
						MACRO_TEST_DESIGNATOR (See <chk_mcro.h) using
						library "sv_base_M.a".
					*/
				strcpy(icmd, cmd);
				if( ! ( ((o_str = strstr(icmd, "../../bin/")) == NULL) ||
						((l_str = strstr(icmd, "sv_base.a")) == NULL) ) ) {
					wr_all("\tCreate load module using sv_base_M.a library.\n");
					if((o_str = strchr(o_str, (int)' ')) == NULL) {
						wr_all("\tERROR\n\tIncorrectly specified command.");
						_exit(1);
					}
					p_str = cmd;
					p_str += o_str -icmd;
					strcpy(p_str, MACRO_TEST_DESIGNATOR),
					p_str += strlen(MACRO_TEST_DESIGNATOR);
					strcpy(p_str, o_str);

					p_str = strstr(cmd, "sv_base.a");
					p_str += strlen("sv_base");
					l_str += strlen("sv_base");
					strcpy(p_str, MACRO_TEST_DESIGNATOR),
					p_str += strlen(MACRO_TEST_DESIGNATOR);
					strcpy(p_str, l_str);

					sprintf(temp,"\t%s\n",cmd);
					wr_all(temp);
					if(execute(cmd)){
						wr_all("\tERROR\n\tCommand exited with an error.\n");
					}
				}
				else {
					wr_all("\tERROR\n\tLib sv_base.a or ../../bin/ not used.");
					_exit(-1);
				}
			}

		} else {
			/*	Test if element is also defined as a macro.
				When a module [^_]*_syn.c exists and executed by install(),
				exits with a status code of MAC_ELEMENT_RET, then a macro
				is defined for the element.
			*/
			if((p_str = strstr(cmd, "_syn ")) != NULL) {
				*(p_str +strlen("_syn")) = '\0';
				while( ! ((*(p_str -1) == ' ') || (*(p_str -1) == '\t')) )
					p_str--;
				if(fork() == (pid_t)0) {
					execl(p_str, p_str, INSTALL_CALL, (char *)0);
					sprintf(temp,
						"\tERROR\n\tCall to execl(%s, %s %s,(char *)0) failed.",
						p_str, p_str, INSTALL_CALL);
					wr_all(temp);
					_exit(1);
				}
				wait(&stat_val);
				if(WEXITSTATUS(stat_val) == MAC_ELEMENT_RET)
					p_macro++;
			}
		}

		*cmd = 0;
	}
	close(infile);
}

char	*nxtok(char *dst, char *src)
{
	while(*src &&(whitespace(*(src)))) ++src;
	while(*src &&(!whitespace(*(dst) = *(src++)))) ++dst;
	while(*src &&(whitespace(*(src)))) ++src;
	*dst = 0;
	return(src);
}

void wr_all(char *data)
{
	if(journal != -1) {
		if(write(journal, data, strlen(data)) != (ssize_t) strlen(data)) {
			sprintf(temp, "ERROR: writing to journal\n");
			write(STDERR_FILENO, temp, strlen(temp));
			_exit(-1);
		}
	}
	if(write(STDERR_FILENO, data, strlen(data)) != (ssize_t) strlen(data)) {
		_exit(-1);
	}
}


/*
** clean(out,in)
** This function takes the input string specified in "in" and
** performs the substitutions indicated by calls to getdef().
** The result is placed in the output string "out".
*/

void	clean(char *out, char *in)
{
	char	tmp[20], *t;
	char	flags[20], *p_ptr;

	while(*in){
		t = &tmp[0];
		while(*in && *in != '%'){
			strncat(out,in++,1);
		}
		if(*in){
			++in;
			while((*in) && *(in) != '%'){
				*(t++) = *(in++);
			}
			*t = 0;
		}
		if(*in){
			if(*(strlen(out) +out -1) == ' ')
				strcat(out,getdef(tmp));
			else {
				p_ptr = strrchr(out, ' ');
				strcpy(flags, p_ptr);
				p_ptr = getdef(tmp);
				for( ; *p_ptr; p_ptr++) {
					if((*p_ptr == ' ') && (*(p_ptr +1)) )
						strcat(out, flags);
					else
						strncat(out, p_ptr, 1);
				}
			}
			++in;
		}
	}
}

/*
** getdef(ident)
** This function determines the substitution to use for the
** keyword given in the string "ident".  It returns a pointer
** to the string to be substituted.
*/

char	*getdef(char *ident)
{
	int	i;
	static	char	tmp[512];

	for(i=0; i<NUMPARMS; ++i){
		if(!strcmp(ident,parms[i].name)){
			break;
		}
	}
	chunk(tmp,parms[i].value);
	return(tmp);
}

void	chunk(char *out, char *in)
{
	char	*p,*e,ch;
	char	tmp[128];
	int		got=0;

	p = in;
	while(*p){
		while(*p &&(*p != ':')){
			p++;
		}
		if(*p){
			p = nxtok(tmp,p+1);
			if(!strcmp(tmp,cursect)){
				got = 1;
				break;
			}
		}
	}
	if(!got){
		p = in;
	}
	e = p;
	while(*e &&(*e != ':'))
		++e;
	ch = *e;
	*e = '\0';
	strcpy(out,p);
	*e = ch;
}


int	execute(char *c)
{
	FILE	*cmdout;
	int	ch;
	char	*p;

	p = &temp[0];
	*(p++) = '\t';
	if((cmdout = p_popen(c)) == NULL)
		return(-1);
	if((ch = getc(cmdout)) != EOF) {
		wr_all("\tMESSAGE\n");
		while(ch != EOF) {
			*(p++) =(char)ch;
			if(ch == '\n'){
				*p = 0;
				wr_all(temp);
				p = &temp[0];
			}
			ch = getc(cmdout);
			
		}
		*(p++) = '\n';
		*p = 0;
		wr_all(temp);
	}
	return(p_pclose(cmdout));
}



void	date_stamp(void)
{
	time_t	tm;
	char	*yr,*dt	;

	tm = time((time_t *)0);
	dt = ctime(&tm);
	*(dt + 24) = 0;
	yr = dt + 20;
	*(dt + 19) = 0;
	dt += 4;
	sprintf(temp,"\t\t%s %s\n", dt, yr);
	wr_all(temp);
}

void	getlibs(int fd, char *str)
{
	char	*p;

	*str = 0;
	(void)fgetln(fd,curlin);
	p = nxtok(temp,curlin);
	while(*temp){
		strcat(str,setup.dest);
		strcat(str,"/lib/");
		strcat(str,temp);
		strcat(str," ");
		p = nxtok(temp,p);
	}
}



/*
** i_cmd(out,in)
** This function takes the input string specified in "in" and
** if the first character is a "!", converts the string to a
** syntax consistent with that specified by the corresponding syntax
** in <install/install.h>.
** Rules for conversion are:
**	Copy the corresonding string entry in "struct impcmd" into "out".
**
**	If the string %@% is found copy the remaining arguments in "in"
**	to "out".
**
**	If the string %...% is found, copy the next argument in "in" to
**	"out".
*/

void	i_cmd(char *out, char *in)
{
	char	*h_ptr;					/* Command syntax from header file */
	char	*i_ptr;					/* Svcomp entry			   */
	char	*o_ptr;					/* Revised command syntax	   */
	char	*c_o_ptr;
	char	flags[10], *p_flags;	/* Support common flags for '@' */

	if(*in != '!') {		/* Command is not to be mapped */
		*out = '\000';
		strcat(out, in);
		return;
	}

	o_ptr = out;
	i_ptr = in;
	h_ptr = getcmd(in +1);
	while(*i_ptr++ != ' ')		/*Skip cmd symbol and subsequent space*/
		;
	while(*h_ptr){
		while(*h_ptr &&(*h_ptr != '%'))
			*o_ptr++ = *h_ptr++;
		if(*h_ptr){
			*flags = '\0';		/* Initialize optional library flags array. */
			c_o_ptr = o_ptr;
			++h_ptr;		/* Bump past '%' */

			/* Special case '@', picks up all remaining arguments */
			if(*h_ptr == '@') {
										/* Process common flags if they exist */
				if(*(o_ptr -1) == ' ') {
					*flags  = ' ';
					*(flags +sizeof(char)) = '\000';
				}
				else {
					*o_ptr = '\000';
					p_flags = strrchr(out, ' ');  /*Pick up any common flags*/
					strcpy(flags, p_flags);
					o_ptr = p_flags;
				}

				while(*i_ptr) {
					if(*i_ptr == ' ') {
						while(*i_ptr == ' ') /* Remove any consecutive spaces */
							i_ptr++;
						p_flags = flags;
						while(*p_flags)				/* Add in common flag */
							*o_ptr++ = *p_flags++;
					}
					else
						*o_ptr++ = *i_ptr++;
				}

				while(*h_ptr &&(*h_ptr != '%'))
					h_ptr++;
				++h_ptr;	/* Bump past '%' */
			}

			else {
				/* Skip name of argument substituted for */
				while(*h_ptr &&(*h_ptr != '%')) 
					h_ptr++;
				h_ptr++;	/* Bump past '%' */
						/* Skip space characters */
				while(*i_ptr &&(*i_ptr == ' '))
					i_ptr++;
				while(*i_ptr &&(*i_ptr != ' ')) 
					/* Get next argument form svcomp entry*/
					*o_ptr++ = *i_ptr++;
			}
					/* Delete common flag if associated with no libs. */
			if(o_ptr == strlen(flags) +c_o_ptr)
				o_ptr = c_o_ptr;
		}
	}
	*o_ptr = '\000';
	return;
}


/*
** getcmd(in)
** This function determines the command to use for the
** command asymbol given in the "svcomp" entry.  It returns a pointer
** to the command syntax to be used.
*/

char	*getcmd(char *in)
{
	struct	cmd_imp	*ptr;
	char	*in_ptr;
	char	cmd1[10], *c_ptr;

	for(in_ptr = in, c_ptr = cmd1; *in_ptr != ' ';)
		*c_ptr++ = *in_ptr++;
	*c_ptr = '\000';
	for(ptr = impcmd; *(ptr->symbol); ptr++) {
		if( ! strcmp(cmd1, ptr->symbol))
			break;
	}
	if( ! *(ptr->symbol)) {
		wr_all("INSTALLATION CONFIGURATION ERROR:\n");
		sprintf(temp,"\t Cannot find mapping command \"%s\".\n", cmd1);
		wr_all(temp);
		_exit(-1);
	}
	return(ptr->cmd);
}

static	int	p_pid[20];

FILE	*p_popen(char *e_cmd)
{
	int		p[2];
	pid_t	pid;

	if(pipe(p) < 0)
		return NULL;
	if((pid = fork()) == 0) {
		char	*args[4096], a_cmd[512];
		char	*p_mem, *p_cmd;
		int	i;

		close(p[0]);
		/* Set stdout to p[1] */
		close(STDOUT_FILENO);
		(void) fcntl(p[1], F_DUPFD, STDOUT_FILENO);

		/* Redirect stderr to p[1] */
		close(STDERR_FILENO);
		(void) fcntl(p[1], F_DUPFD, STDERR_FILENO);
		close(p[1]);

		/* Convert command string to format suitable for execv() */
		p_mem = malloc( strlen(e_cmd) +1 );
		p_cmd = e_cmd;
		for( i = 0; *p_cmd; i++ ) {
			for(args[i] = p_mem; *p_cmd &&(*p_cmd != ' '); )
				*p_mem++ = *p_cmd++;

			*p_mem++ = '\000';
			while( *p_cmd == ' ' )	/*Skip multiple spaces*/
				p_cmd++;
		}

		/* Check if command has been negated */
		if(p_cmd == e_cmd)
			_exit(0);		/* Exit child gracefully */

		args[i] = 0;		/* Signify end-of-arguments */

		/* Setup absolute path name */
		strcpy(a_cmd, PATH_CMD);
		if(a_cmd[0] != '\0') {
			strcat(a_cmd, "/");
			strcat(a_cmd, args[0]);
		}
		else
			strcpy(a_cmd, args[0]);

		execvp(a_cmd, &args[0]);
		wr_all("\tMESSAGE\n");
		sprintf(temp,"\tExecvp() failed on \"%s\"\n", args[0]);
		wr_all(temp);
		_exit(1);
	}
	if(pid == -1) {
		close(p[0]);
		close(p[1]);
		return NULL;
	}
	p_pid[p[0]] = pid;
	close(p[1]);
	return(fdopen(p[0], "r"));
}

int	p_pclose(FILE *ptr)
{
	int	f, r;
	int	status;

	f = fileno(ptr);
	fclose(ptr);
	while((r = wait(&status)) != p_pid[f] && r != -1)
		;
	if(r == -1)
		status = -1;
	return(status);
}
