/*	sv_expect.c - expected return value catching */
#ifndef	_POSIX_SOURCE
#define	_POSIX_SOURCE
#endif

#include <stdarg.h>
#include <sys/types.h>
#include <time.h>
#include <stddef.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <termios.h>

#include <sv_env.h>

static	char			c_saved;
static	int				i_saved;
static	unsigned		u_saved;
static	long			l_saved;
static	double			d_saved;
static	pointer			p_saved;
static	unsigned short	us_saved;
static	int				err_saved;
static	pid_t			pid_t_saved;
static	uid_t			uid_t_saved;
static	gid_t			gid_t_saved;
static	clock_t			clk_t_saved;
static	off_t			off_t_saved;
static	speed_t			speed_t_saved;
static	size_t			size_t_saved;
static	ssize_t			ssize_t_saved;
static	time_t			time_t_saved;
static	mode_t			mode_t_saved;

/* the "format" argument is used as a dummy variable for stdarg.h */
void	sv_exp(char *format, ...)
{
	va_list	ap;

	if(sv_ltest(sv_xenv)) {	/* if sv_lock is set, this is exp value	*/
								/* else exp_errno, SUCC, FAIL, or VOID	*/
		sv_ulock(sv_xenv);	/* unlock								*/
		sv_svalu(sv_xenv);	/* set valu flag						*/
		va_start(ap, format);
		c_saved = va_arg(ap, char);
		va_end(ap);
		va_start(ap, format);
/*Lint will complain of ptr. alignment problems*/
		i_saved = va_arg(ap, int);
		va_end(ap);
		va_start(ap, format);
		u_saved = va_arg(ap, unsigned);
		va_end(ap);
		va_start(ap, format);
		l_saved = va_arg(ap, long);
		va_end(ap);
		va_start(ap, format);
		d_saved = va_arg(ap, double);
		va_end(ap);
		va_start(ap, format);
		p_saved = va_arg(ap, pointer);
		va_end(ap);
		va_start(ap, format);
		us_saved = va_arg(ap, unsigned short);
/*Lint will not complain of ptr. alignment problems after this point*/
		va_end(ap);
								/* C converts char and short arguments to
								   int before passing to a function.
								*/
		va_start(ap, format);
		if(sizeof(pid_t) < sizeof(int))
			pid_t_saved = (pid_t)va_arg(ap, int);
		else
			pid_t_saved = va_arg(ap, pid_t);
		va_end(ap);
		va_start(ap, format);
		if(sizeof(uid_t) < sizeof(int))
			uid_t_saved = (uid_t)va_arg(ap, int);
		else
			uid_t_saved = va_arg(ap, uid_t);
		va_end(ap);
		va_start(ap, format);
		if(sizeof(gid_t) < sizeof(int))
			gid_t_saved = (gid_t)va_arg(ap, int);
		else
			gid_t_saved = va_arg(ap, gid_t);
		va_end(ap);
		va_start(ap, format);
		clk_t_saved = va_arg(ap, clock_t);
		va_end(ap);
		va_start(ap, format);
		off_t_saved = va_arg(ap, off_t);
		va_end(ap);
		va_start(ap, format);
		if(sizeof(speed_t) < sizeof(int))
			speed_t_saved = (speed_t)va_arg(ap, int);
		else
			speed_t_saved = va_arg(ap, speed_t);
		va_end(ap);
		va_start(ap, format);
		if(sizeof(size_t) < sizeof(int))
			size_t_saved = (size_t)va_arg(ap, int);
		else
			size_t_saved = va_arg(ap, size_t);
		va_end(ap);
		va_start(ap, format);
		if(sizeof(ssize_t) < sizeof(int))
			ssize_t_saved = (ssize_t)va_arg(ap, int);
		else
			ssize_t_saved = va_arg(ap, ssize_t);
		va_end(ap);
		va_start(ap, format);
		time_t_saved = va_arg(ap, time_t);
		va_end(ap);
		va_start(ap, format);
		if(sizeof(mode_t) < sizeof(int))
			mode_t_saved = (mode_t)va_arg(ap, int);
		else
			mode_t_saved = va_arg(ap, mode_t);
		va_end(ap);
	}
}

int sv_sets(void)
{
	if(sv_ltest(sv_xenv))
		sv_ssucc(sv_xenv);
	return(0);
}

int sv_setf(void)
{
	if(sv_ltest(sv_xenv))
		sv_sfail(sv_xenv);
	return(0);
}


int sv_setv(void)
{
	if(sv_ltest(sv_xenv))
		sv_svoid(sv_xenv);
	return(0);
}

static int sv_errset(int exp_err, int *sav_loc)
{
	if(sv_ltest(sv_xenv)) {	/* true if within expecting() macro	*/
		sv_ulock(sv_xenv);		/* tell sv_exp to ignore this value */
		sv_serrn(sv_xenv);		/* tell sv_eval to expect an errno	*/
		*sav_loc = exp_err;
	}
	if(!exp_err) *sav_loc = 0;
	return(exp_err);
}	


int sv_sete(int exp_errno)
{
	return(sv_errset(exp_errno, &err_saved));
}


char *sv_getc(void)
{
	return(&c_saved); 
}


int *sv_geti(void)
{
	return(&i_saved); 
}


unsigned *sv_getu(void)
{
	return(&u_saved); 
}


long *sv_getl(void)
{
	return(&l_saved); 
}

double *sv_getd(void)
{ 
	return(&d_saved); 
}


pointer *sv_getp(void)
{
	return(&p_saved); 
}


unsigned short *sv_getus(void)
{
	return(&us_saved); 
}


int sv_gete(void)
{
	return(err_saved);
}


void sv_getv(void)
{
}


pid_t *sv_gpid_t(void)
{
	return(&pid_t_saved); 
}


uid_t *sv_guid_t(void)
{
	return(&uid_t_saved); 
}


gid_t *sv_ggid_t(void)
{
	return(&gid_t_saved); 
}


clock_t *sv_gclk_t(void)
{
	return(&clk_t_saved); 
}


off_t *sv_goff_t(void)
{
	return(&off_t_saved); 
}


speed_t *sv_speed_t(void)
{
	return(&speed_t_saved); 
}


size_t *sv_gsize_t(void)
{
	return(&size_t_saved); 
}


ssize_t *sv_gssize_t(void)
{
	return(&ssize_t_saved); 
}


time_t *sv_gtime_t(void)
{
	return(&time_t_saved); 
}


mode_t *sv_gmode_t(void)
{
	return(&mode_t_saved); 
}
