#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Zstat.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#include	<tssf.h>

static char	path[_POSIX_PATH_MAX +1];
static char	path1[_POSIX_PATH_MAX +1];

void mnt_fs(char *, int);
void umnt_fs(char *);

char *fs_mount(int type, int file_typ)
{
	prep("Establish a pathname for mounting the additional FS.");
	start
		if(Zstrlen(PCTS_FS) == 0) {
			error("Config parameter \"PCTS_FS\" specified wrong.");
		}

		Zgetcwd(path, _POSIX_PATH_MAX);
		Zstrcat(path, "/");
		Zstrcat(path, data_path());
		Zstrcat(path, "mnt_rofs.d");
	finis

	if(file_typ != NO_FILE) {	/* Create file on file system */
		mnt_fs(path, MOUNT_RW);

		if(file_typ == ROFS_DIR) {
			prep("Create test directory.");
			start
				Zstrcpy(path1, path);
				Zstrcat(path1, "/");
				Zstrcat(path1, ROFS_TESTDIR);
				Zmkdir(path1, PROT_ALL);
			finis
		}

		if(file_typ == ROFS_REG_FILE) {
			prep("Create test file.");
			start
				int	fd;

				Zstrcpy(path1, path);
				Zstrcat(path1, "/");
				Zstrcat(path1, ROFS_TESTFILE);
				fd = Zcreat(path1, PROT_ALL);
				Zwrite(fd, "DaTa\n", sizeof("DaTa\n"));
				Zclose(fd);
			finis
		}

		if(file_typ == ENOSPC_FILE) {
			prep("As non-privileged user, use up all available file space");
			prep("less {_POSIX_PATH_MAX} bytes or less.");
			start
				ssize_t	nbytes;
				int		fd, stat;
				pid_t	pid;

				sv_flush();
				pid = Zfork();

				if(pid == (pid_t)0) {
					Zsetuid(UID);			/* Run as normal user */
					Zstrcpy(path1, path);
					Zstrcat(path1, "/");
					Zstrcat(path1, ENOSPC_TESTFILE);
					fd = Zcreat(path1, PROT_ALL);
					while((nbytes = write(fd, path, _POSIX_PATH_MAX))
							==  _POSIX_PATH_MAX)
							;
					Zclose(fd);
					done();
				}
				Zwait(&stat);
			finis
		}

		if(type == MOUNT_RW)
			return(path);

		prep("Unmount the RW file system.");
			umnt_fs(path);
	}

	mnt_fs(path, type);
	return(path);
}


void fs_umount(int file_typ)
{

	cleanup("Unmount the mounted file system.");
		umnt_fs(path);

	if(file_typ != NO_FILE) {	/* Remove file on file system */
		cleanup("Mount file system read-write.");
			mnt_fs(path, MOUNT_RW);

		if(file_typ == ROFS_DIR) {
			cleanup("Remove test directory.");
			start
				expecting(VOID);
				strcpy(path1, path);
				strcat(path1, "/");
				strcat(path1, ROFS_TESTDIR);
				rmdir(path1);
			finis
		}

		if(file_typ == ROFS_REG_FILE) {
			cleanup("Remove test file.");
			start
				expecting(VOID);
				strcpy(path1, path);
				strcat(path1, "/");
				strcat(path1, ROFS_TESTFILE);
				unlink(path1);
			finis
		}

		if(file_typ == ENOSPC_FILE) {
			cleanup("Remove test file.");
			start
				expecting(VOID);
				strcpy(path1, path);
				strcat(path1, "/");
				strcat(path1, ENOSPC_TESTFILE);
				unlink(path1);
			finis
		}

		cleanup("Unmount the RW file system.");
			umnt_fs(path);
	}
}

void fsp_umount(int file_typ)
{

	prep("Unmount the mounted file system.");
		umnt_fs(path);

	if(file_typ != NO_FILE) {	/* Remove file on file system */
		prep("Mount file system read-write.");
			mnt_fs(path, MOUNT_RW);

		if(file_typ == ROFS_DIR) {
			prep("Remove test directory.");
			start
				Zstrcpy(path1, path);
				Zstrcat(path1, "/");
				Zstrcat(path1, ROFS_TESTDIR);
				Zrmdir(path1);
			finis
		}

		if(file_typ == ROFS_REG_FILE) {
			prep("Remove test file.");
			start
				Zstrcpy(path1, path);
				Zstrcat(path1, "/");
				Zstrcat(path1, ROFS_TESTFILE);
				Zunlink(path1);
			finis
		}

		if(file_typ == ENOSPC_FILE) {
			prep("Remove test file.");
			start
				Zstrcpy(path1, path);
				Zstrcat(path1, "/");
				Zstrcat(path1, ENOSPC_TESTFILE);
				Zunlink(path1);
			finis
		}

		prep("Unmount the RW file system.");
			umnt_fs(path);
	}
}


/*	Mount file system read-write	*/
void mnt_fs(char *path, int type)
{
	prep("Mount a file system with %s permission.",
		(type == MOUNT_RO) ? "read-only" : "read/write");
	prep("Ensure mount point has write permission for all.");
	start
		int ret;

		if(type == MOUNT_RO)
			ret = mtrofs(path, PCTS_FS);
		else
			ret = mtrwfs(path, PCTS_FS);
		if(ret) {
			error("Mounting file system \"%s\" on path \"%s\"  failed\n%s%s\n",
				PCTS_FS,
				path,
				"errno returned: ",
				err_lookup(errno));
			done();
		}
		chmod(path, PROT_ALL);		/* Allow all to write to mounted system. */ 
	finis
}


/*	Unmount file system 	*/
void umnt_fs(char *path)
{
	start
		int	ret;

		if( (ret = unmtfs(path, PCTS_FS)) ) {
			error("Unmounting file system \"%s\" on path \"%s\" failed\n%s%s\n",
				PCTS_FS,
				path,
				"errno returned: ",
				err_lookup(errno));
		}
	finis
}
